
/*
 cc -g -c ../as/asprog.c
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "../as/asinput.h"
#include "../as/asprog.h"


/* ------------------------------- ImvstatemenT --------------------------- */


int Imvstatement_new(statement,code,reply,flag)
struct ImvstatemenT **statement;
char *code;
char *reply;
int flag;
{
 struct ImvstatemenT *s;

 *statement= s= TSOB_FELD(struct ImvstatemenT,1);
 if(s==NULL)
   return(-1);
 s->code= NULL;
 s->reply= NULL;
 s->reply_count= 0;
 s->code= TSOB_FELD(char,strlen(code)+1);
 if(s->code==NULL) {
failed:;
   Imvstatement_destroy(statement,0);
   return(-1);
 }
 strcpy(s->code,code);
 s->reply= TSOB_FELD(char *,1);
 if(s->reply==NULL)
   goto failed;
 s->reply[0]= TSOB_FELD(char,strlen(reply)+1);
 if(s->reply[0]==NULL)
   goto failed;
 strcpy(s->reply[0],reply);
 s->reply_count= 1;
 return(1);
}


int Imvstatement_destroy(statement,flag)
struct ImvstatemenT **statement;
int flag;
{
 int i;

 if(*statement==NULL)
   return(0);
 if((*statement)->code!=NULL)
   Smem_freE((char *) (*statement)->code);
 if((*statement)->reply!=NULL) {
   for(i=0;i<(*statement)->reply_count;i++)
     if((*statement)->reply[i]!=NULL)
       Smem_freE((char *) (*statement)->reply[i]);
   Smem_freE((char *) (*statement)->reply);
 }
 Smem_freE((char *) (*statement));
 *statement= NULL;
 return(1);
}  


int Imvstatement_set_code(statement,code,flag)
struct ImvstatemenT *statement;
char *code;
int flag;
{
 char *cpt;
 
 if(code!=NULL) {
   cpt= TSOB_FELD(char,strlen(code)+1);
   if(cpt==NULL)
     return(-1);
 }
 if(statement->code!=NULL)
   Smem_freE((char *) statement->code);
 if(code!=NULL) {
   statement->code= cpt;
   strcpy(cpt,code);
 } else
   statement->code= NULL;
 return(1);
}


int Imvstatement_set_reply(statement,count,reply,add,flag)
/* this function is safe for reply==statement->reply */
struct ImvstatemenT *statement;
int count;
char **reply;
int add;
int flag;
{
 char **hreply= NULL,**old_reply;
 int ret=0,i,old_count;
 
 if(count+add>0){
   hreply= TSOB_FELD(char *,count+add);
   if(hreply==NULL)
     goto ex;
   for(i=0;i<count+add;i++)
     hreply[i]= NULL;
 }
 old_reply= statement->reply;
 statement->reply= NULL;
 old_count= statement->reply_count;
 statement->reply_count= 0;
 for(i=0;i<count;i++) {
   hreply[i]= TSOB_FELD(char,strlen(reply[i])+1);
   if(hreply[i]==NULL)
     {ret= -1; goto ex;}
   strcpy(hreply[i],reply[i]);
 }
 statement->reply_count= count+add;
 statement->reply= hreply;
 hreply= NULL;
 ret= 1;
ex:;
 if(old_reply!=NULL) {
   for(i=0;i<old_count;i++)
     if(old_reply[i]!=NULL)
       Smem_freE((char *) old_reply[i]);
   Smem_freE((char *) old_reply);
 }   
 if(hreply!=NULL){
   for(i=0;i<count;i++)
     if(hreply[i]!=NULL)
       Smem_freE((char *) hreply[i]);
   Smem_freE((char *) hreply);
 }
 return(ret);
}


int Imvstatement_add_replyline(statement,line,flag)
struct ImvstatemenT *statement;
char *line;
int flag;
{
 int ret;

 ret= Imvstatement_set_reply(statement,statement->reply_count,
                             statement->reply,1,0);
 if(ret<=0)
   return(ret);
 ret= Sregex_string(&(statement->reply[statement->reply_count-1]),line,0);
 if(ret<=0)
   return(ret);
 return(1);
}


int Imvstatement_copy(source,target,flag)
struct ImvstatemenT *source;
struct ImvstatemenT *target;
int flag;
{
 int ret;

 ret= Imvstatement_set_code(target,source->code,0);
 if(ret<=0)
   return(ret);
 ret= Imvstatement_set_reply(target,source->reply_count,source->reply,0,0);
 if(ret<=0)
   return(ret);
 return(1); 
}


/* ------------------------------- AsproG ------------------------------ */


int Asprog_new(prog,input,size,flag)
struct AsproG **prog;
struct AsinpuT *input;
int size;
int flag;
{
 struct AsproG *p;
 char *cpt= NULL;
 int i,ret;

 *prog= p= TSOB_FELD(struct AsproG,1);
 if(p==NULL)
   return(-1);
 p->input= input;
 p->agent= input->agent;
 p->helptext= NULL;
 p->maindir= NULL;
 p->trashdir= NULL;
 p->statements= NULL;
 if(size<=0)
   size= 16;
 p->size= size;
 p->max_step= 16*size;
 p->count= 0;
 p->reply= NULL;
 p->statements= TSOB_FELD(struct ImvstatemenT *,size);
 if(p->statements==NULL) 
   goto ex;
 if(Sregex_string(&(p->helptext),"",0)<=0)
   goto ex;
 if(Asagent_get_maindir(p->agent,&cpt,0)<=0)
   goto ex;
 ret= Asprog_set_maindir(p,cpt,0);
 Sregex_string(&cpt,NULL,0);
 if(ret<=0)
   goto ex;
 if(Asagent_get_trashdir(p->agent,&cpt,0)<=0)
   goto ex;
 ret= Asprog_set_trashdir(p,cpt,0);
 Sregex_string(&cpt,NULL,0);
 if(ret<=0)
   goto ex;
 for(i=0;i<size;i++)
   p->statements[i]= NULL;
 return(1);
ex:;
 Asprog_destroy(prog,0);
 return(-1);
}


int Asprog_destroy(prog,flag)
struct AsproG **prog;
int flag;
{
 int i;
 struct ImvstatemenT **list;

 if(*prog==NULL)
   return(0);
 Asprog_clear(*prog,0);
 Sregex_string(&((*prog)->helptext),NULL,0);
 Sregex_string(&((*prog)->maindir),NULL,0);
 Sregex_string(&((*prog)->trashdir),NULL,0);
 if((*prog)->statements!=NULL) 
   Smem_freE((char *) (*prog)->statements);
 Smem_freE((char *) (*prog));
 *prog= NULL;
 return(1);
}


int Asprog_clear(prog,flag)
struct AsproG *prog;
int flag;
{
 struct ImvstatemenT **list;
 int i,ret= -1;

 if(Sregex_string(&(prog->helptext),"",0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&(prog->maindir),"",0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&(prog->trashdir),"trashdir",0)<=0)
   {ret= -1; goto ex;}
 if(prog->statements!=NULL) {
   list= prog->statements;
   for(i=0;i<prog->count;i++)
     if(list[i]!=NULL)
       Imvstatement_destroy(&(list[i]),0);
 }
 Imvstatement_destroy(&(prog->reply),0);
 prog->count= 0;
 ret= 1;
ex:;
 return(ret);
}


int Asprog_set_helptext(prog,text,flag)
struct AsproG *prog;
char *text;
int flag;
/*
 bit0= append to existing text
*/
{
 char *cpt;
 int l=0;

 if(flag&1) if(prog->helptext!=NULL) l= strlen(prog->helptext);
 cpt= TSOB_FELD(char,l+strlen(text)+1);
 if(cpt==NULL)
   return(-1);
 if(l>0) 
   strcpy(cpt,prog->helptext);
 else
   *cpt= 0;
 if(prog->helptext!=NULL)
   Smem_freE((char *) prog->helptext);
 prog->helptext= cpt;
 strcat(cpt,text);
 return(1);
}


int Asprog_set_maindir(prog,dir,flag)
struct AsproG *prog;
char *dir;
int flag;
/*
 bit0= append
*/
{
 int ret;

 ret= Sregex_string(&(prog->maindir),dir,flag&1);
 if(ret<=0)
   return(ret);
 Asagent_set_maindir(prog->agent,dir,flag&1);
 return(ret);
}


int Asprog_set_trashdir(prog,dir,flag)
struct AsproG *prog;
char *dir;
int flag;
/*
 bit0= append
*/
{
 int ret;

 ret= Sregex_string(&(prog->trashdir),dir,flag&1);
 if(ret<=0)
   return(ret);
 ret= Asagent_set_trashdir(prog->agent,dir,flag&1);
 return(ret);
}


int Asprog_enlarge(prog,step,flag)
struct AsproG *prog;
int step;
int flag;
{
 struct ImvstatemenT **list;
 int i,size;

 if(step<=0){
   step= prog->size;
   if(step>prog->max_step)
     step= prog->max_step;
 }
 size= prog->size+step;
 list= TSOB_FELD(struct ImvstatemenT *,size);
 if(list==NULL)
   return(-1);
 for(i=0;i<prog->count;i++)
   list[i]= prog->statements[i];
 for(i=prog->count;i<size;i++)
   list[i]= NULL;
 Smem_freE((char *) prog->statements);
 prog->statements= list;
 return(1);
}


int Asprog_find_code(prog,code,flag)
struct AsproG *prog;
char *code;
int flag;
/*
 bit0= return first NULL entry
*/
{
 int ret,i;

 for(i=0;i<prog->count;i++) {
   if(prog->statements[i]==NULL) {
     if(flag&1)
       return(i);
 continue;
   }
   if(strcmp(prog->statements[i]->code,code)==0) 
     return(i);
 }
 return(-1);
}


int Asprog_add(prog,code,reply,flag)
struct AsproG *prog;
char *code;
char *reply;
int flag;
{
 int ret,i;

 i= Asprog_find_code(prog,code,0);
 if(i<0){
   i= Asprog_find_code(prog,code,1);
   if(i<0){
     if(prog->count>=prog->size) {
       ret= Asprog_enlarge(prog,0,0);
       if(ret<=0)
         return(ret);
     }
     i= prog->count;
   }
   ret= Imvstatement_new(&(prog->statements[i]),code,reply,0);
   if(ret<=0)
     return(ret);
   if(i==prog->count)
     prog->count++;
 } else {
   ret= Imvstatement_add_replyline(prog->statements[i],reply,0);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Asprog_clear_code(prog,code,flag)
struct AsproG *prog;
char *code;
int flag;
{
 int ret,i;

 i= Asprog_find_code(prog,code,0);
 if(i<0)
   return(2);
 Imvstatement_destroy(&(prog->statements[i]),0);
 return(1);
}


int Asprog_default_kbd(prog,flag)
struct AsproG *prog;
int flag;
{
 int i,ret;
 char line[SfileadrL];
 static char maindir[]= {"."};
 /* a very small keyset . Usually keysets are loaded from files */
 static char code_reply[][40]= {
   "-",   "-moveto",     "trashdir",
   "=",   "-moveto",     "trashdir",
   "?",   "-help_short", "",
   "\177","-back",       "",
   "\\",  "-back",       "",
   "!",   "-jpgb",       "",
   "<",   "-jpgb",       "",
   " ",   "-skip",       "",
   ">",   "-skip",       "",
   "U",   "-undo",       "",
   "u",   "-undo",       "",
   "~",   "-doubleto",   "trashdir",
   "@HOPEFULLY_UNIQUE_BOUNDARY@","",""
 };

 ret= Asprog_set_helptext(prog,
      " ?help >forward <backward =isknown -delete ~deldouble &undo !reform",0);
 if(ret!=1)
   return(-1); 
 ret= Asprog_set_maindir(prog,".",0);
 if(ret<=0)
   return(ret);
 ret= Asprog_set_trashdir(prog,"trashdir",0);
 if(ret<=0)
   return(ret);

 for(i=0;strcmp(code_reply[i],"@HOPEFULLY_UNIQUE_BOUNDARY@")!=0;i+=3) {
   strcpy(line,code_reply[i+1]);
   if(code_reply[i+1]!=0 && code_reply[i+2][0]!=0)
     strcat(line," ");
   if(code_reply[i+2][0]!='/' && code_reply[i+2][0]!=0 ) {
     strcat(line,maindir);
     strcat(line,"/");
   }
   strcat(line,code_reply[i+2]);
   ret= Asprog_add(prog,code_reply[i],line,0);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Asprog_add_md(prog,code,cmd,flag)
struct AsproG *prog;
char *code;
char *cmd;
int flag;
/* return:
 <=0 : error
 3   : incomplete statement
 4   : wrong arguments in statement
*/
{
 int ret,prefix_l;
 char *cpt,*dpt,reply[SfileadrL],*result_pt= NULL; 

 reply[0]= 0;
 for(cpt= cmd; *cpt!=0 && *cpt!='-'; cpt++);
 if(*cpt==0)
   return(3);
 for(dpt= cpt+1; *dpt!=0 && (isalnum(*dpt) || *dpt=='_'); dpt++);
 prefix_l= dpt-cpt;
 strncpy(reply,cpt,prefix_l);
 reply[prefix_l]= 0;
 if(*dpt!=0){
   reply[prefix_l]= ':';
   reply[prefix_l+1]= 0;
   if(*(dpt+1)==0)
     return(4);
   strcat(reply,dpt+1);
 }
 ret= Asprog_add(prog,code,reply,0);
 return(ret);
}


int Asprog_keyset_mapclear(prog,line,flag)
struct AsproG *prog;
char *line;
int flag;
/*
 bit0= clear (else map)
*/
/* return:
 <=0 : error
 1   : ok
 3   : incomplete statement
 4   : wrong arguments in statement
*/
{
 char code[SfileadrL],*eqpt;
 unsigned char code_char[2];
 int ret= 0,code_no= -1,limit,i;
 
 eqpt= strchr(line+1,'@');
 if(eqpt==NULL) {
   if(flag&1)
     eqpt= line+1+strlen(line+1);
   else
     goto incomplete;
 }
 if(*(eqpt+1)==0 && !(flag&1))
   goto incomplete;
 strncpy(code,line,eqpt-line);
 code[eqpt-line]= 0;
 code_char[1]= 0;
 if(code[0]=='@') { /* specially formed single codes */
   code_no= -1;
   if(code[1]=='#') { /* single decimal keynumber */
     sscanf(code+2,"%ld",&code_no);
     if(code_no<0 || code_no>255)
       goto wrong_args;
     code_char[0]= code_no;
     if(flag&1)
       ret= Asprog_clear_code(prog,code_char,0);
     else
       ret= Asprog_add_md(prog,code_char,eqpt+1,0);
   } else if(code[1]==':') {
                      /* multi character code (not transported by input yet) */
     if(flag&1)
       ret= Asprog_clear_code(prog,code+1,0);
     else
       ret= Asprog_add(prog,code+1,eqpt+1,0);
   }
   if(ret!=1)
     goto ex;
 } else { /* multiple single characters */
   limit= eqpt-line;
   for(i=0;i<limit;i++) {
     code_char[0]= code[i];
     if(flag&1)
       ret= Asprog_clear_code(prog,code_char,0);
     else
       ret= Asprog_add_md(prog,code_char,eqpt+1,0);
     if(ret!=1)
       goto ex;
   }
 }
 ret= 1;
ex:
 return(ret);
incomplete:;
 ret= 3;
 goto ex;
wrong_args:;
 ret= 4;
 goto ex;
}


int Asprog_keyset_map(prog,line,flag)
struct AsproG *prog;
char *line;
int flag;
/* return:
 <=0 : error
 1   : ok
 3   : incomplete statement
 4   : wrong arguments in statement
*/
{
 int ret;

 ret= Asprog_keyset_mapclear(prog,line,0);
 return(ret);
}


int Asprog_keyset_clear(prog,line,flag)
struct AsproG *prog;
char *line;
int flag;
/* return:
 <=0 : error
 1   : ok
*/
{
 int ret;

 ret= Asprog_keyset_mapclear(prog,line,1);
 return(ret);
}


int Asprog_define(prog,line,flag)
struct AsproG *prog;
char *line;
int flag;
/* return:
 <=0 : error
 1   : ok
 2   : unrecognized statement
 3   : incomplete statement
 4   : wrong arguments in statement
*/
{
 int ret= 0;
 char *stmt,*cpt= NULL;
 
 for(stmt=line;*stmt!=0 && !isalnum(*stmt) && *stmt!='!' && *stmt!='#'; stmt++);

 if(stmt[0]==0||stmt[0]=='#'||stmt[0]=='!') {
   /* Remark stmt */
   goto ok;
 } else if(strcmp(stmt,"clearall")==0) {
   ret= Asprog_clear(prog,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(stmt,"clear",5)==0 && stmt[5]!=0) {
   ret= Asprog_keyset_clear(prog,stmt+6,0);
   if(ret!=1)
     goto ex;
 } else if(strncmp(stmt,"helpstart",9)==0 && stmt[9]!=0) {
   /* clear helptext and start new one */
   if(Sregex_string(&(prog->helptext),stmt+10,0)<=0)
     {ret= -1; goto ex;}
 } else if(strncmp(stmt,"helpmore",8)==0 && stmt[8]!=0) {
   /* add new stmt to help text */
   if(prog->helptext!=NULL)
     if(Sregex_string(&(prog->helptext),"\r\n",1)<=0)
       {ret= -1; goto ex;}
   if(Sregex_string(&(prog->helptext),stmt+9,1)<=0)
     {ret= -1; goto ex;}
 } else if(strncmp(stmt,"map",3)==0 && stmt[3]!=0) {
   ret= Asprog_keyset_map(prog,stmt+4,0);
   if(ret!=1)
     goto ex;
 } else if(strncmp(stmt,"maindir",7)==0 && stmt[7]!=0) {
   ret= Asagent_eval(prog->agent,stmt+8,&cpt,0);
   if(ret<=0)
     return(ret);
   ret= Asprog_set_maindir(prog,cpt,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(stmt,"readfrom",8)==0 && stmt[8]!=0) {
   ret= Asagent_eval(prog->agent,stmt+9,&cpt,0);
   if(ret<=0)
     return(ret);
   ret= Asprog_read_kbd(prog,cpt,0);
   if(ret!=1)
      goto ex;
 } else if(strncmp(stmt,"trashdir",8)==0 && stmt[8]!=0) {
   ret= Asagent_eval(prog->agent,stmt+9,&cpt,0);
   if(ret<=0)
     return(ret);
   ret= Asprog_set_trashdir(prog,cpt,0);
   if(ret<=0)
     {ret= -1; goto ex;}
 } else {
   ret= 2;
   goto ex;
 }
ok:;
 ret= 1;
ex:;
 Sregex_string(&cpt,NULL,0);
 return(ret);
}


int Asprog_read_kbd(prog,adr,flag)
struct AsproG *prog;
char *adr;
int flag;
{
 FILE *fp=NULL;
 int ret=1;
 char *fret,line[SfileadrL];

 fp= Sfile_fopen(adr,"r",0);
 if(fp==NULL)
   return(0);
 while(1){
   fret= Sfile_fgets(line,sizeof(line),fp);
   if(fret==NULL)
 break;
   ret= Asprog_define(prog,line,0);
   if(ret!=1)
     goto ex;
 }
 ret= 1;
ex:;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Asprog_read_kbd_v(prog,adr,flag)
struct AsproG *prog;
char *adr;
int flag;
{
 FILE *fp;
 char *fret,line[SfileadrL],main_dir[SfileadrL],target[SfileadrL];
 unsigned char code[2];
 char *cpt,*bpt,*dpt;
 int ret,i,dec,l;

 fp= Sfile_fopen(adr,"r",0);
 if(fp==NULL)
   return(0);

 while(1){
   fret= Sfile_fgets(line,sizeof(line),fp);
   if(fret==NULL)
 break;
   if(line[0]==0||line[0]=='#'||line[0]=='!') 
 continue;
   if(line[0]=='?') { /* helptext line */
     Sregex_trimline(line+1,0);
     strcat(line+1,"\n");
     ret= Asprog_set_helptext(prog,line+1,1);
     if(ret<=0)
       goto ex;
   } else if(line[0]=='>') { /* common main directory */
     strcpy(main_dir,line+1);
     Sregex_trimline(main_dir,0);
     if(main_dir[0]!=0 && prog->trashdir==NULL){
       ret= Asprog_set_trashdir(prog,main_dir,0);
       ret+= Asprog_set_trashdir(prog,"/trashdir",1);
       if(ret<2)
         return(ret);
     }
   } else if(line[0]=='@' && line[1]=='-') { /* trash directory */
     strcpy(target,line+1);
     Sregex_trimline(target,0);
     ret= Asprog_set_trashdir(prog,target,0);
     if(ret<=0)
       return(ret);
   } else {
     Sregex_trimline(line,0);
     bpt= strchr(line,' ');
     if(bpt==NULL)
 continue;
     Sregex_trimline(bpt,0);
     if(*bpt=='^') 
       dpt= strchr(bpt,':');
     if(dpt==NULL) dpt= bpt;
     else          dpt++;
     for(cpt= line; cpt<bpt; cpt++) {
       code[0]= *cpt;
       code[1]= 0;
       if(*cpt=='~' && cpt+1<bpt){
         cpt++;
         if(*cpt=='_') code[0]= ' ';
         else if(*cpt=='~') code[0]= '~';
         else if(cpt+2<bpt) {
           dec= ((*cpt)-48)*100+((*(cpt+1))-48)*10+(*(cpt+2))-48;
           cpt+= 2;
           if(dec<=0||dec>255)
     continue;
           code[0]= dec;
         }
       }
       strcpy(target,"-moveto ");
       l= strlen(target);
       for(i=0;bpt+i!=dpt;i++)
         target[l+i]= bpt[i];
       target[l+i]= 0;
       if(bpt[i]!='/') {
         strcat(target,main_dir);
         strcat(target,"/");
       }
       strcat(target,dpt);
       ret= Asprog_add(prog,code,target,0);
       if(ret<=0)
         goto ex;
     }
   }  
 }
 ret= 1;
ex:;
 if(fp!=NULL) 
   Sfile_fclose(fp);
 return(1);
}


int Asprog_interpreter(prog,command,flag)
struct AsproG *prog;
char *command;
int flag;
{
 int i,ret;

 if(prog->reply!=NULL)
   Imvstatement_destroy(&(prog->reply),0);

 if(command[0]=='-' && strlen(command)>1){
   ret= Imvstatement_new(&(prog->reply),command,command,0);
   return(ret);
 }

 i= Asprog_find_code(prog,command,0);
 if(i<0)
   return(0);
 ret= Imvstatement_new(&(prog->reply),"","",0);
 if(ret<=0)
   return(ret);
 ret= Imvstatement_copy(prog->statements[i],prog->reply,0);
 if(ret<=0)
   return(ret);
 return(1);
}



