
/*
 cc -g -c imgcmp.c
*/

#include <unistd.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "../as/asagent.h"
#include "imgmap.h"
#include "imgsamp.h"

#include "imgcmp.h"


/* ----------------------------- ImgcmP ------------------------------ */


int Imgcmp_new(icmp,agent,mapname,flag)
struct ImgcmP **icmp;
struct AsagenT *agent;
char *mapname;
int flag;
{
 struct ImgcmP *m;

 *icmp= m= TSOB_FELD(struct ImgcmP,1);
 if(m==NULL)
   return(-1);
 Imgcmp_init(m,agent,agent->progname,mapname,"-","-",0);
 return(1);
}


int Imgcmp_destroy(icmp,flag)
struct ImgcmP **icmp;
int flag;
{
 /* remember that storage of imp is included in icmp, not linked outside */
 Imgmap_deactivate(&((*icmp)->imp),0);
 Smem_freE((char *) (*icmp));
 *icmp= NULL;
 return(1);
}


int Imgcmp_init(icmp,agent,progname,mapname,resultziel,meldziel,flag)
struct ImgcmP *icmp;
struct AsagenT *agent;
char *progname,*mapname,*resultziel,*meldziel;
int flag;
{
 int ret;

 icmp->agent= agent;
 Imgmap_init(&(icmp->imp),icmp,mapname,0);
 icmp->find_self= 0;
 icmp->bundle_open= 0;
 strcpy(icmp->progname,progname);
 icmp->client_intern= 1;
 icmp->bundle_count= 0;
 icmp->show_commands= 0;
 icmp->append_nondash= 0;
 return(1);
}


int Imgcmp_copy(source,target,flag)
struct ImgcmP *source,*target;
int flag;
/*
 bit0= for Imgmap_copy : also copy reference to cache 
*/
{
 target->agent= source->agent; 
 Imgmap_copy(&(source->imp),&(target->imp),flag&1);
 target->find_self= source->find_self;
 target->bundle_open= source->bundle_open;
 strcpy(target->progname,source->progname);
 target->client_intern= source->client_intern;
 target->bundle_count= source->bundle_count;
 target->show_commands= source->show_commands;
 return(1);
}


int Imgcmp_set_client_intern(icmp,wert,flag)
struct ImgcmP *icmp;
int wert;
int flag;
{
 int ret,aflag;

 icmp->client_intern= !!wert;
 if(icmp->client_intern==0) 
   aflag= 0;
 else
   aflag= 1;
 ret= Imgcmp_autoclient(icmp,aflag);
 if(ret<0)
   return(ret);
 return(1);
}


int Imgcmp_client_is_intern(icmp,flag)
struct ImgcmP *icmp;
int flag;
{
 return(icmp->client_intern);
}


int Imgcmp_result(icmp,text,flag)
struct ImgcmP *icmp;
char *text;
int flag;
/* 
 bit0= no newline
*/
{
 Asagent_stdout(icmp->agent,text,!(flag&1));
 return(1);
}


int Imgcmp_stderr(icmp,text,flag)
struct ImgcmP *icmp;
char *text;
int flag;
/* 
 bit0= no newline
 bit1= immediate internal display
*/
{
 if(flag&2) {
   fprintf(stderr,"%s",text);
   if(!(flag&1))
     fprintf(stderr,"\n");
 }
 Asagent_stderr(icmp->agent,text,(flag&2)|!(flag&1));
 return(1);
}


int Imgcmp_breakchk(icmp,flag)
/* check wether the program shall end */
struct ImgcmP *icmp;
int flag;
/*
 bit0= exit(2) rather than return(1)
*/
{
 char name[SfileadrL],line[SfileadrL];
 struct stat stbuf;
 char *progname;
 int stat_ret;

 progname= icmp->progname;

 if(Sfile_home_adr(".imgcmp_break",name,0)<=0)
   return(2);
 stat_ret= stat(name,&stbuf);
 if(stat_ret==-1) {
   Imgmap_get_mapname(&(icmp->imp),name,4);
   stat_ret= stat(name,&stbuf);
 }
 if(stat_ret!=-1){
   Imgcmp_stderr(icmp,"",0);
   sprintf(line,"%s : aborted because existence of %s",progname,name);
   Imgcmp_stderr(icmp,line,0);
   if(unlink(name)==0){
     sprintf(line,"%s : removed %s\n",progname,name);
   }else{
     sprintf(line,"%s : couldn't remove %s , errno=%ld",
                    progname,name,errno);
   }
   Imgcmp_stderr(icmp,line,0);
   if(flag&1) exit(2);
   return(1);
 }
 return(0); 
}


int Imgcmp_rc(icmp,flag)
struct ImgcmP *icmp;
int flag;
{
 int ret;
 char name[SfileadrL],*username= NULL;

 if(Sfile_home_adr(".similar_rc",name,0)<=0)
   return(2);

 ret= Asagent_get_current_user(icmp->agent,&username,1);
 if(ret<=0)
   goto ex;
 ret= Asagent_cmd_readfrom(icmp->agent,name,
                            Asagent_get_event_connection(icmp->agent,0),
                            username,0);
ex:
 Sregex_string(&username,NULL,0);
 return(ret);
}


int Imgcmp_append(icmp,in_filename,flag)
struct ImgcmP *icmp;
char *in_filename;
int flag;
/*
 bit0= refresh rather than append
*/
{
 char *progname,msg[2*SfileadrL],filename[SfileadrL];
 int ret;
 struct ImgmaP *imp;
 struct ImgsamplE *samp= NULL;

 progname= icmp->progname;
 imp= &(icmp->imp);

 strcpy(filename,in_filename);
 ret= Sfile_make_adr_global(filename,0);
 if(ret<=0)
   goto ex;
 if(flag&1)
   Imgcmp_stderr(icmp,"refresh in imgmap : ",1);
 else
   Imgcmp_stderr(icmp,"append to imgmap : ",1);

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0)
   goto ex;
 ret= Imgmap_read_rgb(imp,filename,samp,0);
 if(ret<=0) 
   goto ex;
 if(flag&1)
   ret= Imgmap_refresh_filename(imp,filename,samp,0);
 else
   ret= Imgmap_record(imp,filename,samp,0);
 if(ret<=0)
   goto ex;
 Imgcmp_stderr(icmp,filename,0);
 ret= 1;
ex:;
 if(ret<=0) {
   Imgmap_get_errmsg(imp,msg,0);
   if(msg[0]!=0) {
     Imgcmp_stderr(icmp,msg,0);
     Imgmap_reset_errmsg(imp,0);
   }
 }
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgcmp_append_list(icmp,filename,flag)
struct ImgcmP *icmp;
char *filename;
int flag;
/*
 bit2= client mode : send single -append commands to server
return:
 <=0 : error
   1 : all went normal
   3 : end being client
*/
{
 int zaehler= 0,ret;
 char line[SfileadrL],*cmd= NULL;
 FILE *fp= NULL;
 char *progname;

 progname= icmp->progname;

 fp= Sfile_fopen(filename,"rb");
 if(fp==NULL){
   sprintf(line,"-append_list : failed to open %s",filename);
   Imgcmp_stderr(icmp,line,0);
   return(0);
 }
 while(Sfile_fgets(line,SfileadrL-1,fp)!=NULL){
   if(Imgcmp_breakchk(icmp,0)) {
      ret= 0;
      goto abort_message;
   }
   zaehler++;
   if(flag&4) {
     if(Sregex_string(&cmd,"-append:",0)<=0)
       {ret= -1; goto ex;}
     if(Sregex_string(&cmd,line,1)<=0)
       {ret= -1; goto ex;}
     ret= Imgcmp_send_cmd(icmp,cmd,8,0);
   } else {
     ret= Imgcmp_append(icmp,line,0);
   }
   if(ret<0){
abort_message:;
     sprintf(line,"%s : -append_list aborted after %ld input lines",
                  progname,zaehler);
     Imgcmp_stderr(icmp,line,0);
     goto ex;
   }
   ret= Asagent_poll(icmp->agent,1);
   if(ret!=1)
     goto abort_message;
 }
 ret= 1;
ex:;
 if(fp!=NULL && fp!=stdin && fp!=stdout) {
   fclose(fp);
   fp= NULL;
 }
 Sregex_string(&cmd,NULL,0);
 return(ret);
}


int Imgcmp_list_doubles(icmp,flag)
struct ImgcmP *icmp;
int flag;
{
 struct ImgmaP imp_scan;
 char memofilename[SfileadrL];
 char filename[SfileadrL],currname[SfileadrL],start_name[SfileadrL];
 char line[2*SfileadrL];
 struct ImgsamplE *samp= NULL;
 int count=0, found= 0,foundmemo,ret;
 FILE *fp;
 char *progname;
 struct ImgmaP *imp;

 progname= icmp->progname;
 imp= &(icmp->imp);
 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0)
   goto ex;

 /* >>> replace imp_scan by dynamically allocated object */
 Imgmap_copy(imp,&imp_scan,2|4);
 Imgmap_search_reset(&imp_scan,0);
 ret= Imgmap_activate_exempt(imp,0);
 if(ret<0)
   goto ex;

 /* resume search which was possibly interrupted */
 memofilename[0]= 0;
 start_name[0]= 0;
 Imgmap_get_mapname(imp,memofilename,3);
 fp= fopen(memofilename,"rb");
 if(fp!=NULL){
   if(Sfile_fgets(start_name,SfileadrL-1,fp)==NULL)
     strcpy(start_name,"");
   fclose(fp);
   if(start_name[0]){
     sprintf(line,"because %s : searching for %s",memofilename,start_name);
     Imgcmp_stderr(icmp,line,0);
     unlink(memofilename);
   }
 }
 if(start_name[0]){
   ret= Imgmap_search_filename(&imp_scan,start_name,samp,0);
   if(ret<=0) {
     Imgcmp_stderr(icmp,"--- cannot find start file ---",0);
     goto ex;
   }
   count= ret;
   start_name[0]= 0;
   sprintf(line,"skipped %ld lines",count);
   Imgcmp_stderr(icmp,line,0);
   count= 0;
 }

 /* accept any file */
 while(Imgmap_search(&imp_scan,currname,samp,2)>0){
   if(Imgcmp_breakchk(icmp,0)) {
     sprintf(line,"%s : -list_doubles  aborted after %ld input lines\n",
                  progname,count);
     Imgcmp_stderr(icmp,line,0);
     if(memofilename[0]){
       fp= fopen(memofilename,"wb");
       if(fp!=NULL){
         if(start_name[0]==0) {
           fprintf(fp,"%s",currname);
           sprintf(line,"%s written to %s for resuming",currname,memofilename);
         } else {
           fprintf(fp,"%s",start_name);
           sprintf(line,"%s written to %s for resuming",
                        start_name,memofilename);
         }
         Imgcmp_stderr(icmp,line,0);
         fclose(fp);
       }
     }
     ret= found; goto ex;
   }

   if((count%50)==0) {
     Imgcmp_stderr(icmp,"",0);
     sprintf(line,"%8ld : %-8ld  ",count,found);
     Imgcmp_stderr(icmp,line,1|2);
   }
   count++;

   foundmemo= found;
   Imgmap_search_reset(imp,0);
   while(Imgmap_search(imp,filename,samp,4)>0){
     if(strcmp(currname,filename)!=0) {
       Imgcmp_result(icmp,currname,0); 
       found++;
   break;
     }
   }
   if(found>foundmemo)
     Imgcmp_stderr(icmp,"#",1|2);
   else
     Imgcmp_stderr(icmp,"-",1|2);
 }
 if(strlen(Imgmap_get_errmsg(imp,line,0))!=0)
 Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
 sprintf(line,"\n%8ld : %-8ld  ",count,found);
 Imgcmp_stderr(icmp,line,0);
 ret= found;
ex:
 /* >>> Imgmap_deactivate(&imp_scan,1); */
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgcmp_offadr(icmp,filename,foundnames,foundsamples,maxfnames,flag)
struct ImgcmP *icmp;
char *filename;
char **foundnames;
struct ImgsamplE **foundsamples;
int maxfnames;
int flag;
/*
 bit0= report findings
 bit1= report findings as result
 bit2= return only one filename if possible
*/
{
 int ret;

 ret= Imgmap_offadr(&(icmp->imp),filename,foundnames,foundsamples,maxfnames,
                    flag&7);
 return(ret);
}


int Imgcmp_autoclient(icmp,flag)
struct ImgcmP *icmp;
int flag;
/*
 bit0= close connection and do not open a new one
*/
{
 char filename[SfileadrL],servhost[SfileadrL];
 int ret,portnumber,count;
 FILE *fp;

 if(Asagent_get_service_mode(icmp->agent,0)==1) {
   ret= Asagent_tcp_client(icmp->agent,"",0,"","",2|4); /* end client */
   if(ret<=0)
     return(ret);
 }

 Imgmap_set_disabled(&(icmp->imp),0,0);
 Imgmap_get_mapname(&(icmp->imp),filename,5);
 if((fp= fopen(filename,"rb"))==NULL)
   return(0);
 count= fscanf(fp,"%s %ld",servhost,&portnumber);
 fclose(fp);
 if(count<2)
   return(0);
 if((flag&1)||icmp->client_intern) {
   Imgmap_set_disabled(&(icmp->imp),1,0);
 }else{
   ret= Asagent_tcp_client(icmp->agent,servhost,portnumber,
                            "default","encapsulated",1|4);
   if(ret<=0) {
     /* switch off auto_client mode */
     icmp->client_intern= 1;
     ret= -1;
   }
 }
 return(ret);
}


int Imgcmp_send_cmd(icmp,cmd,fnamept,flag)
struct ImgcmP *icmp;
char *cmd;
int fnamept; /* Start of fileaddress in cmd, 0 = no fileaddress. Not used yet */
int flag;
/*
 bit0= for Asagent_send_cmd() : do not prepend -mark
*/
{
 int ret;
 char wdir[SfileadrL];

 if(!icmp->bundle_open) {
   ret= Asagent_send_cmd(icmp->agent,"-bundle:start",1);
   if(ret<=0)
     goto ex;
 }
 strcpy(wdir,"-cd:");
 if(getcwd(wdir+strlen(wdir),sizeof(wdir)-1-strlen(wdir))!=NULL) {
   ret= Asagent_send_cmd(icmp->agent,wdir,0);
   if(ret<=0)
     goto ex;
 }
 if(icmp->find_self) {
   ret= Asagent_send_cmd(icmp->agent,"-search_self_too",1);
   if(ret<=0)
     goto ex;
 } else {
   ret= Asagent_send_cmd(icmp->agent,"-search_others_only",1);
   if(ret<=0)
     goto ex;
 }
 if(cmd[0]!=0) {
   ret= Asagent_send_cmd(icmp->agent,cmd,flag&1);
   if(ret<=0)
     goto ex;
 }
 if(!icmp->bundle_open) {
   ret= Asagent_send_cmd(icmp->agent,"-bundle:end",1);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Imgcmp_search(icmp,filename,foundnames,flag)
struct ImgcmP *icmp;
char *filename;
struct LstrinG **foundnames;
int flag;
/*
 bit0= do not find_self
 bit2= use exempt list despite imp->use_exemptlist==0
*/
{
 int ret;
 char line[SfileadrL],found[SfileadrL];
 dev_t dev_arg,dev_found;
 ino_t inode_arg,inode_found;
 struct stat stbuf;
 struct ImgmaP *imp;
 struct AsagenT *agent;
 struct ImgsamplE *samp= NULL,*foundsample;

 imp= &(icmp->imp);
 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0)
   goto ex;
 agent= icmp->agent;

 Lstring_destroy_all(foundnames,0);
 if(stat(filename,&stbuf)==-1) {
   sprintf(line,"cannot determine properties of file '%s'",filename);
   Imgcmp_stderr(icmp,line,0);
   ret= 0; goto ex;
 }
 inode_arg= stbuf.st_ino;
 dev_arg= stbuf.st_dev;
 ret= Imgmap_read_rgb(imp,filename,samp,0);
 if(ret<=0)
   goto ex;
 if(strlen(Imgmap_get_errmsg(imp,line,0))>0)
   Imgcmp_stderr(icmp,line,0);
 Imgmap_search_reset(imp,0);
 while(Imgmap_search(imp,found,samp,flag&4)>0){
   if((flag&1) || !icmp->find_self){
     inode_found= !inode_arg;
     if(stat(found,&stbuf)!=-1) {
       inode_found= stbuf.st_ino;
       dev_found= stbuf.st_dev;
     }
     if(inode_found==inode_arg && dev_arg==dev_found)
 continue;
   }
   ret= Lstring_new(foundnames,found,*foundnames,1);
 }
 if(strlen(Imgmap_get_errmsg(imp,line,0))>0)
   Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
 ret= 1;
ex:;
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgcmp_helptext(icmp,reply,flag)
struct ImgcmP *icmp;
char **reply;
int flag;
/*
 bit0-3 = mode : 0=intro , 1=command reference , 2=appendice , 3=theory
*/
{
 static Asagent_helplinE intro_text[]= {
"",
"This software maintains a database of image file names and micro thumbnails.",
"Main purpose is to find images which look very similar to a given one.",
"This data base is also called \"map\" here.",
"",
"The program does not alter the image files but possibly creates temporary",
"copies of them in /tmp directory. In most cases the temporary files are",
"unlinked immediately after creation so it is quite unlikely that files",
"like /tmp/imgrgb*.jpg remain.",
"To handle other file formats than JPEG, the program \"convert\" has to",
"be accessible.",
"",
"At start the program tries to open the file",
"  $HOME/.similar_rc",
"and execute the text lines as options. This may be prevented by option",
" -no_rc  which must be the first argument to work properly.",
"",
"The database is depicted by a  dbaddress  which in fact is a",
"commont start text for the names of several files. Usual are :",
" ${dbaddress}_nam    the filenames of the registered images",
" ${dbaddress}_smp16  the fingerprints (samples) of the images",
" ${dbaddress}_exempt list of explicitely distinguished images (if any)",
" ${dbaddress}_break  used to abort long running operations",
" ${dbaddress}_pubno  address of the active guarding server (if any)",
"The default for dbaddress is \"similar_db\".",
"",
"A single instance of this program can act as guarding server of a map.",
"It then serves for the others via a TCP/IP based client-server protocol.",
"If guarded, a map may not be accessed directly by any other instance of",
"this program. Access to the database may be gained by -auto_client:on .",
"This avoids problems of concurrency while altering a database and allows",
"effective use of memory caching.",
"",
"A guarding server publishes its TCP/IP address in file ${dbaddress}_pubno .",
"If this file is found by any other instance of this program which tries to",
"access dbaddress directly then it refuses to open the files. If auto client",
"mode is set \"on\" then the guarding server is contacted automatically",
"whenever a guarded map is selected by command -mapname .",
"",
"Several non-map commands are performed locally by the client program, though.",
"",
"Some of the long lasting operations can be stopped by the existence of",
"the file  ${HOME}/.imgcmp_break  or  ${dbaddress}_break :",
" -append_list , -list_doubles",
"The program tries to remove this file and begins to execute the next option.",
"",
"This software uses a communications module called agent. Its features and",
"commands also are available in a standalone program called sagent which may",
"serve as frontend client or communications node in a net of servers.",
"",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};

 static Asagent_helplinE reference_text[]= {
"",
"                         Command reference",
"",
" \"fileaddress\" is path and name of an accessible file.",
" \"sourceaddress\" is a fileaddress or \"-\" for standard input.",
"",
"-append:fileaddress",
"     Append the image to the database.",
"",
"-append_nondash:on|off",
"     If set to \"on\" the beginning of an -append command may be ommitted",
"     and only the fileaddress needs to be given. This works only with",
"     fileaddresses which do not begin with a dash \"-\". Since agent's",
"     @ command input at the start terminal possibly automatically prepends",
"     dashes, one should only rely on this mode for arguments and commands",
"     sent from frontend clients.",
"",
"-append_list:sourceaddress",
"     Read a list of filenames from source and append them to the database.",
"     Each line of input is used as one filename including all special",
"     characters besides the line's end marks.",
"",
"-auto_client:on|off",
"     Sets the automatic client mode which controls the behavior when",
"     encountering a guarded map (see below, command -guarded_map).",
"     If this mode is \"on\" the program tries to use the server of the",
"     guarded map. If this mode is \"off\" then access to the guarded map",
"     is disabled.",
"     If the program is already connected to a server, this connection is",
"     closed when -auto_client gets executed. A new connection gets is",
"     established if appropriate.",
"     One should consider to set -auto_client:on in file $HOME/.similar_rc",
"     after all local settings (like -security:all:user) are done.",
"",
"-caching:on|off",
"     With  -caching:on  the programm uses a memory copy to speed up search",
"     operations in long program runs like -list_doubles or in guarding mode.",
"     Since reading the cache lasts some time, short runs become slower !",
"     To get an uncompromised cache for a multi-client server, use",
"     -guarded_map: before -caching: .",
"",
"-cd:directoryaddress",
"     Set the working directory for use with fileaddresses.",
"",
"-del_back:fileaddress",
"     Delete fileaddress from the database if it is the newest entry.",
"     The freed space is immediately reusable.",
"",
"-del_dir:fileaddress",
"     Delete all files within the directory given by fileaddress from the",
"     database. Do not delete subdirectories.",
"",
"-del_file:fileaddress",
"     Delete the entry of a file with the same byte content as the one",
"     with the given address from the database. Doesn't act if more than",
"     one matching entry is found.",
"",
"-del_file_adr:fileaddress",
"     Delete all entries of the given fileaddress from the database.",
"",
"-del_tree:fileaddress",
"     Delete all files within the directory given by fileaddress and within",
"     all its subdirectories from the database.",
"",
"-distinguish:fileaddress",
"     Write a record to the list of explicitely distinguished files. The",
"     record will contain the fileaddress and the addresses of all images",
"     that can be found by a -search run for fileaddress. In -list_doubles",
"     these addresses will not show up again as possible duplicates of",
"     fileaddress. Vice versa, fileaddress will not show up as duplicate of",
"     any of these addresses.",
"     Usually this list will only be checked with -list_doubles . If",
"     -use_exemptlist is set to \"on\" it will affect all search operations.",
"     Beware of the possible race condition that a real duplicate might",
"     be appended to the database in the time between checking the old",
"     false duplicates and sending the -distinguish command.",
"     Therefore the permission class of -distinguish is \"dbadmin\" rather",
"     than \"modify\". Better disable it in heavy multiuser situations.",
"",
"-garbage_collection",
"     Copy all valid entries to a new database, rename the old one to a",
"     backup name and rename the new one to the original name.",
"",
"-guarded_map:connection_type",
"     Write own server address into file ${dbaddress}_pubno , provided a",
"     service of the given connection_type is active. This file marks the",
"     database as a guarded map.",
"     Guarded maps may not be accessed directly by instances of this program.",
"     The guarding state ends when another map is selected by -mapname or",
"     when the service ends properly (i.e. no total program crash occurs).",
"     Currently, only connection_type \"tcp\" is recognized. This type writes",
"     hostname and portnumber of the current -tcp_service:2 to the file.",
"",
"-list_all_files",
"     Prints all stored files to standard output.",
"",
"-list_doubles",
"     Prints those stored files to standard output which find other stored",
"     files by -search.",
"",
"-lookup:fileaddress",
"     Show the stored sample values of the file.",
"",
"-imagesize:mode:fileaddress",
"     Print the geometrical size of an image file. Valid modes are:",
"       x   the width of the image measured in pixels",
"       y   the height of the image measured in pixels",
"       z   the depth i.e. number of color components per pixel",
"       xy  width height",
"       xyz width height depth",
"     These informations might be useful to estimate the quality of an image.",
"     The command is performed locally by the client. To avoid an unnecessary",
"     connection to a server, -no_rc might be helpful.",
"",
"-lookup_adr:fileaddress",
"     Print the address of a registered file which has the same byte content",
"     as the file with the given fileaddress. This option retrieves the",
"     \"official\" name of a file.",
"",
"-lookup_adr_all:fileaddress",
"     Like -lookup_adr but printing all known addresses if there is more than",
"     one.",
"",
"-lookup_adr_count:fileaddress",
"     Print the number of addresses which can be found with -lookup_adr_all",
"",
"-mapname:dbaddress",
"     Set name of database. Actual fileaddresses are ${dbaddress}_nam and",
"     ${dbaddress}_smp16 .",
"",
"-match_par:color_tolerance:bw_tolerance:color_ignore:bw_ignore[:method]",
"     Set the parameters for comparation of samples.",
"     color_tolerance sets the maximum color difference for two pixels to",
"     be considered matching. The color values of a pixel are restricted",
"     to a range of 0.0 to 255.0.",
"     color_ignore sets the number of tolerable non-matching pixels",
"     for matching samples. There are 16 pixels in a sample.",
"     bw_tolerance and bw_ignore do the same for images with higher",
"     redundancy such as black-and-white images.",
"     These values should usually be about half of the color values.",
"     The optional parameter method chooses one of the following comparison",
"     algorithms:",
"       color_diff          subtracts the colors of corresponding pixels",
"                           and tests wether the difference is tolerable.",
"       color_double_diff   computes the color average of both images and",
"                           shifts colors to make both averages equal.",
"                           Then it performs color_diff comparison.",
"       color_scaled_diff   computes the color average of both images and",
"                           scales color relative to 64.0 to make both",
"                           averages equal before color_diff comparison.",
"     Both average correcting algorithms are suitable to find heavily colo",
"     corrected images. They are a lot slower though.",
"",
"-no_rc   (this is not really a command)",
"     This special option works only if given as first argument at program",
"     start. In all other cases it will not be recognized.",
"     If given, the program does not read and execute  $HOME/.similar_rc",
"     which usually occurs at the very start.",
"",
"-refresh:fileaddress",
"     Recompute and update the sample values of a fileaddress already",
"     registered in the database.",
"",
"-search:fileaddress",
"     Prints all addresses of stored image files similar to the given one.",
"",
"-search_self_too",
"     Given before option -search : print the address of the given file too",
"     if stored in the database and image still similar.",
"",
"-search_others_only",
"     Given before option -search : do not print the address of the given",
"     file. (default)",
"",
"-statistics",
"     Determines the overall color histograms of red, green and blue.",
"     Prints a shell script to stdout. If executed, this script prints a",
"     text table describing the histograms and the three maximum intensities.",
"     Also it creates and displays 3 files in the working directory :",
"     red.gif , green.gif and blue.gif which show the histograms in a self",
"     adjusting scale. A thin horizontal line marks the average value and a",
"     vertical line marks the color value with the highest density.",
"     Try:   similar -statistics | sh",
"     But be aware of the three image files which are created.",
"",
"-theory",
"     Prints a text which speculates about the reasons why similar possibly",
"     works.",
"",
"-use_exemptlist:mode",
"     If mode is \"on\" then all search operations will respect the list",
"     of explicitely distinguished files (see -distinguish). This may slow",
"     down searching and especially the operation of command -distinguish.",
"     If mode is \"off\" then only -list_doubles will use the list.",
"     A change of database (-mapname) sets -use_exemptlist to \"off\".",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};

 static Asagent_helplinE appendice_text[]= {
"",
"",
"                      Appendix Similar A : Keyset Definitions",
"",
"similar is not much intented for dialog. Usually servers guard databases",
"and clients get their commands from arguments.",
"",
"",
"                      Appendix Similar L : Legal stuff",
"",
"By using libjpeg this software is based in part on the work of",
"the Independent JPEG Group (http://www.ijg.org).",
"",
"This software itself is copyright 2001, Thomas Schmitt stic-source@gmx.net",
"and provided to you without any warranty under an open source BSD license.",
"(see file COPYING)",
"",
"",
"                      Appendix Similar P : Permission Classes",
"",
"This is a list of application permission classes and their commands.",
"see \"Appendix agent P\" for more information about permission classes.",
"",
"  Class       Command (resp. Action)",
"------------------------------------------------------------------------",
"",
"  service     -guarded_map, -mapname , -match_par",
"  dbadmin     -caching , -distinguish , -garbage_collection",
"  modify      -append , -append_list , -del_back , -del_dir , -del_file",
"              -del_file_adr , -del_tree , -refresh , ",
"  env         -append_nondash , -auto_client , -cd , -search_others_only",
"              -search_self_too",
"  info        -imagesize , -list_all_files , -list_doubles , -lookup",
"              -lookup_adr ,  -lookup_adr_all , -lookup_adr_count , -search",
"              -statistics",
"",
"------------------------------------------------------------------------",
"",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};
 static Asagent_helplinE why_text[]= {
"",
"",
"                  Appendix Similar T : Why it probably works",
"",
"similar touches the topic of Content Based Image Retrieval. Most of the",
"CBIR systems use color histograms with sophisticated statistical operations",
"or shape detection or detection of contrast peaks.",
"Why does a simple 4x4 image system find lots of doubles ? Without using any",
"image filtering but the most primitive area-oriented anti-aliasing.",
"It expectetly finds identical images, but why does it find many alterated",
"ones too ? How may this ability be enhanced ?",
"",
"Since the system does not analyze the images at all, it has to depend on",
"a low probability to show one of the very many false replies and on a",
"high probability to detect a duplicate if there is any.",
"",
"If the likelihood of a /random/ alert is very low then an alert has a high",
"likelihood to be caused by some /systematic/ connection between the two",
"pictures. One of those systematic connections could be a common ancestor",
"of both images. This special connection type is what we are looking for.",
"With tolerances not too wide, it really seems to be the dominant one in",
"color_double_diff comparison.",
"",
"How likely is the color_diff comparison algorithm to match two completely",
"random images with N pixels with a tolerance of T ?",
"Neglecting effects at the ends of the scale ",
"  P= 1/((256/T)^(N*3))",
"with N=16 and T=4 this results in 1/(2^288) or about 2e-87 .",
"",
"But this calculation is much too optimistic. Images are not random and that",
"is a very important reason why it works at all. If they were random noise,",
"then out of mere statistical reasons the miniaturization of the images would",
"result in quite uniform samples.",
"Therefore let us now assume quite roughly that with the 16-pixel samples the",
"tolerance effectively distinguishes only about E/T different stages of color.",
"I found good reason in my test collection that E is larger than 32.",
"Maybe i am wrong, statistics was not my favorite topic of math. Actually i",
"found in 125000 samples the most crowded color notch to be only 2 times as",
"dense as the average. That would lead to E > 100 . Aliasing of the histogram",
"cannot be be worse than a factor of three if T>=1.",
"More problematic seems to be a slight bump in the curves of green and blue",
"where red reaches its maximum. Looks like there is some color correlation.",
"Arghh, how to calculate that effect ?",
"( see doc/collection_histogram_*.gif : horizontal line = average density )",
"",
"Recalculating the probability of a false duplicate :  1/(2^144) = 4.5e-44",
"",
"",
"             Finding non-identical but similar images",
"",
"One may wish to extend the search in two main directions :",
"  Images from the same source which show the same world scene and objects",
"  Images from different sources which show altered views of the same picture",
"",
"The object recognition aspect is currently beyond the scope of similar.",
"Not surprisingly, quite a lot of very similar world scenes will show up as",
"altered versions of each other and may be even found during the search for",
"nearly identical images.",
"",
"Another type of human percepted similarity is the recognition of altered",
"2D pictures. This may be different scans of the same picture or postprocessed",
"versions of the same scan.",
"Some types of alteration of the same picture :",
"  Annotations and limited manipulations",
"  Global Color changes",
"  Frames",
"  Different outtakes",
"",
"Annotations are handled by the blind spot feature. A certain number of",
"mismatches is tolerated and regarded as result of limited manipulations.",
"A rough adaption to global color changes is provided by match-method",
"color_double_diff which shifts the colors so that the samples's average",
"colors become the same.",
"The tests with color_scaled_diff did not justify the additional processing",
"time. Maybe one should try to determine possible gamma corrections. (how ?)",
"",
"Frames are naughty because they influence 12 of 16 sample pixels. Maybe one",
"should try to clip them, but how to decide in a fast and suitable way ?",
"Outtakes are even worse because the lack of visible borders.",
"Both may need a kind of holographic (resp. correlative) characterization",
"or partial image comparison which currently go beyond the scope of similar.",
"",
"The influence of B blind spots further reduces selectivity",
"   P= (N over B)/((E/T)^((N-B)*3))",
"where (N over B) is  N!/(B!*(N-B)!) = N*(N-1)*...*(N-B+1) / (B*(B-1)*...*2)",
"With B=4 the calculation yields 1820/(2^108) = 5.5e-30 ",
"Even if we assume E to be as low as 16 it would still be 1820/(2^72) = 4e-19",
"(a lower E means a higher overall similarity of colors within an image",
"collection).",
"",
"So let us assume 1e-20 as upper bound of random fault likelihood.",
"With that builtin selectivity, the likelihood for a completely false",
"duplicate in a collection of 100000 images should be about 1e-15 .",
"",
"This likelihood is valid only for faults that we would consider to be without",
"any reason. For a collection that concentrates on a certain topic, there",
"will be a lot correlation of colors. For black and white images the",
"pessimisticly estimated likelihoot is about 1e-4. A dangerously low value",
"if one has to deal with a large collection.",
"With T=1 and B=2 we get 120/(16^14) = 1.5e-15  for black and white images.",
"",
"Since method color_double_diff allows an arbitrary shift by a color vector",
"the reduction of selectivity is roughly estimated as 1 additional",
"blindspot (an arbitrary color vector too).",
"That would be  4368/(2^66) = 6e-17  for color parameters T=4, B=4",
"an             560/(16^13) = 1.2e-13  for b/w parameters T=1, B=2",
"Remember E=16 is quite a pessimistic value. These results should be upper",
"bounds of the undesired probability of uncorrelated false matches.",
"",
"To estimate the likelihood for the whole cross-checking of a collection,",
"one has to consider the birthday paradox. According to hash lectures, one",
"may expect a first collision in a hash table with S slots after about",
"sqrt(2*S) tries. Our sample space may be seen as a hashtable with at least",
" 1/6e-17 slots (resp. 1/1.2e-13 for b/w). sqrt(2*S) is 1.8e8 (resp. 4e6).",
"",
"So, for the purpose of finding nearly identical images the above parameters",
"T and B should be sufficient at least for a collection of a million images.",
"",
"",
" >>> not finished yet <<<",
"",
" >>> intended color permutation test",
"",
" >>> intended correlative characterization",
"",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};
 Asagent_helplinE *helptext;
 int mode,ret,i;
 char line[SfileadrL];

 if(Sregex_string(reply,"",0)<=0)
   {ret= -1; goto ex;};
 mode= flag&15;
 if(mode==0)
   helptext= intro_text;
 else if(mode==1)
   helptext= reference_text;
 else if(mode==2)
   helptext= appendice_text;
 else if(mode==3)
   helptext= why_text;
 else
   {ret= 0; goto ex;}
 for(i=0;strcmp(helptext[i],
                "@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
                )!=0;i++) {
   sprintf(line,"%s\n",helptext[i]);
   if(Sregex_string(reply,line,1)<=0)
     {ret= -1; goto ex;};
 }
 ret= 1;
ex:;
 return(ret);
}


int Imgcmp_cmd_imagesize(icmp,in_cmd,flag)
struct ImgcmP *icmp;
char *in_cmd;
int flag;
{
 int ret,height,width,depth;
 char *cpt,*mode= NULL,*filename= NULL,line[1024];
 struct ImgrgB *rgb= NULL;

 ret= Asagent_get_permission(icmp->agent,"info","imagesize",0);
 if(ret<=0)
   goto ex;

 cpt= in_cmd;
 ret= Sregex_cut_neck(cpt,':',&mode,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret==2 || strlen(cpt)==0) {
malformed:;
   Asagent_stderr(icmp->agent,"--- malformed options to -imagesize : ",0);
   Asagent_stderr(icmp->agent,in_cmd,1);
   ret= 0; goto ex;
 }
 if(Sregex_string(&filename,cpt,0)<=0)
   {ret= -1; goto ex;}
 ret= Imgmap_open_rgb(&(icmp->imp),filename,&rgb,0);
 if(ret<=0)
   goto ex;
 Imgrgb_get_sizes(rgb,&height,&width,&depth,0);
 line[0]= 0;
 if(strcmp(mode,"x")==0) {
   sprintf(line,"%ld",width);
 } else if(strcmp(mode,"y")==0) {
   sprintf(line,"%ld",height);
 } else if(strcmp(mode,"z")==0) {
   sprintf(line,"%ld",depth);
 } else if(strcmp(mode,"xy")==0) {
   sprintf(line,"%ld %ld",width,height);
 } else if(strcmp(mode,"xyz")==0) {
   sprintf(line,"%ld %ld %ld",width,height,depth);
 } else {
   Imgcmp_stderr(icmp,"unknown option to -imagesize : ",1);
   Imgcmp_stderr(icmp,mode,0);
 }
 if(line[0])
   Imgcmp_result(icmp,line,0);
 ret= 1;
ex:;
 Imgrgb_destroy(&rgb,0);
 Sregex_string(&mode,NULL,0);
 Sregex_string(&filename,NULL,0);
 return(ret);
}

int Imgcmp_cmd_match_par(icmp,in_cmd,flag)
struct ImgcmP *icmp;
char *in_cmd;
int flag;
{
 int ret,color_max_misses,bw_max_misses,search_method= -1;
 char *cpt,*color_eps_text= NULL,*bw_eps_text= NULL,*search_method_text= NULL;
 double color_eps= -1.0,bw_eps= -1.0;

 ret= Asagent_get_permission(icmp->agent,"service","match_par",0);
 if(ret<=0)
   goto ex;

 cpt= in_cmd;
 ret= Sregex_cut_neck(cpt,':',&color_eps_text,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret==2 || strlen(cpt)==0) {
malformed:;
   Asagent_stderr(icmp->agent,"--- malformed options to -match_par : ",0);
   Asagent_stderr(icmp->agent,in_cmd,1);
   ret= 0; goto ex;
 }
 ret= Sregex_cut_neck(cpt,':',&bw_eps_text,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret==2 || strlen(cpt)==0)
   goto malformed;
 ret= Sregex_cut_neck_int(cpt,':',&color_max_misses,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret==2)
   goto malformed;
 ret= Sregex_cut_neck_int(cpt,':',&bw_max_misses,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret==1) {
   ret= Sregex_cut_neck(cpt,':',&search_method_text,&cpt,0);
   if(ret<0)
     goto ex;
 }

 sscanf(color_eps_text,"%lf",&color_eps);
 sscanf(bw_eps_text,"%lf",&bw_eps);
 if(search_method_text!=NULL) {
   if(strcmp(search_method_text,"default")==0
      || strcmp(search_method_text,"color_diff")==0) {
     search_method= 0;
   } else if(strcmp(search_method_text,"color_double_diff")==0) {
     search_method= 1;
   } else if(strcmp(search_method_text,"color_scaled_diff")==0) {
     search_method= 2;
   } else if(strcmp(search_method_text,"-1")!=0) {
     Asagent_stderr(icmp->agent,
                     "--- unknown search-method with -match_par : ",0);
     Asagent_stderr(icmp->agent,search_method_text,1);
     ret= 0; goto ex;
   }
 }

 Imgmap_set_match_par(&(icmp->imp),search_method,color_eps,bw_eps,
                                   color_max_misses,bw_max_misses,0);
 ret= 1;
ex:;
 Sregex_string(&search_method_text,NULL,0);
 Sregex_string(&color_eps_text,NULL,0);
 Sregex_string(&bw_eps_text,NULL,0);
 return(ret);
}


int Imgcmp_cmd_mapname(icmp,mapname,flag)
struct ImgcmP *icmp;
char *mapname;
int flag;
{
 int ret;
 char filename[SfileadrL];
 
 if(mapname[0]==0)
   return(2);

 ret= Asagent_get_permission(icmp->agent,"service","mapname",0);
 if(ret<=0)
   goto ex;
 Imgmap_get_mapname(&(icmp->imp),filename,0);
 if(strcmp(mapname,filename)==0)
   return(2);

 /* eventually release guarding of old map */
 Asagent_publish_tcpadr(icmp->agent,"",2,1);
 Imgmap_set_guarding(&(icmp->imp),0,0);

 ret= Imgmap_set_mapname(&(icmp->imp),mapname,0);
 if(ret<=0)
   return(ret);
 ret= Imgcmp_autoclient(icmp,0);
 if(ret<0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Imgcmp_cmd_distinguish(icmp,filename,flag)
struct ImgcmP *icmp;
char *filename;
int flag;
{
 int ret;
 struct LstrinG *foundnames= NULL,*ls;
 char listname[SfileadrL],line[SfileadrL*2];

 ret= Asagent_get_permission(icmp->agent,"dbadmin","distinguish",0);
 if(ret<=0)
   goto ex;
 ret= Imgcmp_search(icmp,filename,&foundnames,1);
 if(ret<=0)
   goto ex;

 if(foundnames==NULL) {
   sprintf(line,"no matching files found for %s",filename);
   Imgcmp_stderr(icmp,line,0);
 } else {
   /* add to exempt file and eventually to loaded exempt list */;
   Imgcmp_result(icmp,"----------------------------------------",0);
   Imgcmp_result(icmp,filename,0);
   Imgcmp_result(icmp,"          distinguished from",0);
   for(ls= foundnames;ls!=NULL;ls=ls->next)
     Imgcmp_result(icmp,ls->text,0);
   Imgcmp_result(icmp,"----------------------------------------",0);
   /* beware: foundnames may be set to NULL in this call */
   ret= Imgmap_add_exempt(&(icmp->imp),filename,&foundnames,0);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 Lstring_destroy_all(&foundnames,0);
 return(ret);
}


int Imgcmp_cmd_help(icmp,flag)
struct ImgcmP *icmp;
int flag;
{
 int ret,i;
 char line[4096],*helptext= NULL;

 ret= Asagent_get_permission(icmp->agent,"help","help",0);
 if(ret<=0)
   goto ex;


 Asagent_stdout(icmp->agent,"",1);
 Asagent_stdout(icmp->agent,"",1);
 Asagent_stdout(icmp->agent,"                Features and Options of ",0);
 Asagent_stdout(icmp->agent,icmp->agent->progname,1);

 for(i=0;i<3;i++) {
   if(i<2)
     ret= Imgcmp_helptext(icmp,&helptext,i);
   else
     ret= Asagent_helptext(icmp->agent,&helptext,i);
   if(ret<=0)
     goto ex;
   Asagent_stdout(icmp->agent,helptext,4);
   if(i<2)
     ret= Asagent_helptext(icmp->agent,&helptext,i);
   else
     ret= Imgcmp_helptext(icmp,&helptext,i);
   if(ret<=0)
     goto ex;
   Asagent_stdout(icmp->agent,helptext,4);
 }
 Asagent_stdout(icmp->agent,"",1);
 ret= 1;
ex:
 Sregex_string(&helptext,NULL,0);
 return(ret);
}


int Imgcmp_cmd_theory(icmp,flag)
struct ImgcmP *icmp;
int flag;
{
 int ret,i;
 char line[4096],*helptext= NULL;

 ret= Asagent_get_permission(icmp->agent,"help","theory",0);
 if(ret<=0)
   goto ex;

 ret= Imgcmp_helptext(icmp,&helptext,3);
 if(ret<=0)
   goto ex;
 Asagent_stdout(icmp->agent,helptext,4);
 ret= 1;
ex:
 Sregex_string(&helptext,NULL,0);
 return(ret);
}


int Imgcmp_cmd_statistics(icmp,flag)
struct ImgcmP *icmp;
int flag;
{
 char currname[SfileadrL];
 static char color_name[3][8]={"red","green","blue"};
 char line[2*SfileadrL];
 struct ImgsamplE *samp= NULL;
 int ret,i,j,maxrgb_idx[3],idx,sample_count=0;
 double rgb[257][3],maxrgb[3],count= 0.0,scale,avg[3],color;
 struct ImgmaP *imp;
 imp= &(icmp->imp);

 ret= Asagent_get_permission(icmp->agent,"info","statistics",0);
 if(ret<=0)
   goto ex;

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0)
   goto ex;
 Imgmap_search_reset(imp,0);

 for(i=0;i<257;i++) for(j=0;j<3;j++) rgb[i][j]= 0.0;
 avg[0]= avg[1]= avg[2]= 0.0;
 /* accept any file */
 while(Imgmap_search(imp,currname,samp,2)>0){

   /* >>> test wether b/w and count separately */;

   for(i=0;i<samp->spoints;i++) {
     for(j=0;j<3;j++) {
       color= Imgsample_get_coord(samp,j,i,0);
       idx= color;
       rgb[idx][j]+= 1.0;
       avg[j]+= color;
     }
     count+= 1.0;
   }
   sample_count++;
   if((sample_count%1000)==0) {
     sprintf(line,"\r  %ld\r",sample_count);
     Imgcmp_stderr(icmp,line,1|2);
   }
 }
 Imgcmp_stderr(icmp,
               "\r                                                       ",0);
 if(count<=0) {
   Imgcmp_result(icmp,"echo 'no valid samples in database'",0);
   ret= 2; 
   goto ex;
 }

 for(j=0;j<3;j++) 
   avg[j]/= count;

 for(j=0;j<3;j++) {
   maxrgb[j]= 0.0;
   maxrgb_idx[j]= -1;
 }
 Imgcmp_result(icmp,
      "echo '================== Total Histogram of Database ================'",
               0);
 for(i=0;i<256;i++) {
   sprintf(line,
        "echo '%3ld : %10.lf %10.lf %10.lf  :   %2.2lf%%  %2.2lf%%  %2.2lf%%'",
        i,rgb[i][0],rgb[i][1],rgb[i][2],
        rgb[i][0]/count*100.0,rgb[i][1]/count*100.0,rgb[i][2]/count*100.0);
   Imgcmp_result(icmp,line,0); 
   for(j=0;j<3;j++)
     if(maxrgb[j]<rgb[i][j]) {
       maxrgb[j]= rgb[i][j];
       maxrgb_idx[j]= i;
     }
 }
 Imgcmp_result(icmp,
        "echo ====================== Maximum Densities ======================",
               0);
 for(j=0;j<3;j++) {
   sprintf(line,"echo '%7s at intensity %3ld : %10.lf   :   %2.2lf%%'",
                color_name[j],maxrgb_idx[j],maxrgb[j],maxrgb[j]/count*100.0);
   Imgcmp_result(icmp,line,0); 
 }
 Imgcmp_result(icmp,
        "echo ===============================================================",
               0);
 sprintf(line,
 "echo 'total :  samples= %ld   ,   points= %.lf  ,  average= %.lf,%.lf,%.lf'",
              sample_count,count,avg[0],avg[1],avg[2]);
 Imgcmp_result(icmp,line,0); 
 Imgcmp_result(icmp,"echo",0); 

 for(j=0;j<3;j++) {
   scale= maxrgb[j]/100.0;
   sprintf(line,"convert -geometry '256x100!' \\");
   Imgcmp_result(icmp,line,0); 
   sprintf(line,"-pen black -draw 'fillRectangle 0,0 255,99' \\");
   Imgcmp_result(icmp,line,0); 
   sprintf(line,"-pen %s \\",color_name[j]);
   Imgcmp_result(icmp,line,0); 
   for(i=0;i<256;i++) {
     if(rgb[i][j]/scale<1.0) 
   continue;
     sprintf(line,"-draw 'line %ld,99 %ld,%.lf' \\",i,i,100.0-rgb[i][j]/scale);
     Imgcmp_result(icmp,line,0); 
   }
   sprintf(line,"-pen black -draw 'line 0,%.lf 255,%.lf' \\",
                count/256.0/scale,count/256.0/scale);
   Imgcmp_result(icmp,line,0); 
   sprintf(line,"-pen black -draw 'line %ld,%.lf %ld,99' \\",
                maxrgb_idx[j],count/256.0/scale,maxrgb_idx[j]);
   Imgcmp_result(icmp,line,0); 
   sprintf(line,"NULL: %s.gif",color_name[j]); 
   Imgcmp_result(icmp,line,0); 
   sprintf(line,"display %s.gif &",color_name[j]); 
   Imgcmp_result(icmp,line,0); 
 }
ex:
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgcmp_interpreter(icmp,flag)
struct ImgcmP *icmp;
int flag;
/*
 bit2= client mode 
 bit3= forced local mode
*/
/*
 return values:
  <=0 = error
  1 = all went normal
  2 = redo user interaction
  3 = end agent service
  4 = command not recognized
*/
{
 char *cmd= NULL,line[SfileadrL],filename[SfileadrL],*adr,*username= NULL;
 char *filename_pt;
 int ret,idx,no_files= 0,yes= 0,count,i,j,rm_ret,nam_adr,pkt_adr,client_mode= 0;
 int inode_arg,dev_arg,inode_found,dev_found,fnamept= 0,no_prepend= 0;
 struct stat stbuf;
 struct AsagenT *agent;
 struct ImgmaP *imp;
 struct ImgsamplE *samp= NULL,*foundsample;

 imp= &(icmp->imp);
 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0)
   goto ex;
 agent= icmp->agent;
 if(((flag&4) || Asagent_get_service_mode(agent,0)==1) && !(flag&8))
   client_mode= 1;

 agent->input_fd= Asagent_get_event_connection(agent,0);
 ret= Asagent_get_event_cmd(agent,&cmd,0);
 if(ret<=0)
   goto ex;
 if(strncmp(cmd,"--imgcmp",8)==0) {
   /* remove own module prefix if there */
   for(i=8;cmd[i]!=0;i++)
     cmd[i-8]= cmd[i];
   cmd[i]= 0;
 }
 if(icmp->show_commands) {
   sprintf(line,"%ld> ",Asagent_get_event_connection(agent,0));
   Asagent_stdout(agent,line,0);
   Asagent_stdout(agent,cmd,1);
 }

 if(cmd[0]==0||cmd[0]=='#'||cmd[0]=='!') {
   ;
 } else if(strncmp(cmd,"-addfile",8)==0 && cmd[8]!=0) {
   fnamept= 8; 
   goto append_file;
 } else if(strncmp(cmd,"-append",7)==0 && cmd[7]!=0 && cmd[7]!='_') {
   fnamept= 7; 
append_file:;
   ret= Asagent_get_permission(agent,"modify","append",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   ret= Imgcmp_append(icmp,cmd+fnamept+1,0);
   if(ret<0)
     goto ex;
 } else if(strncmp(cmd,"-append_list",12)==0 && cmd[12]!=0) {
   fnamept= 12;
   ret= Asagent_get_permission(agent,"modify","append_list",0);
   if(ret<=0)
     goto ex;
   ret= Imgcmp_append_list(icmp,cmd+fnamept+1,(!!client_mode)*4);
   if(ret<=0)
     goto ex;
   if(ret==3) {
     Imgcmp_autoclient(icmp,1);      /* end client service */
     ret= 1;
   }
 } else if(strncmp(cmd,"-append_nondash",15)==0 && cmd[15]!=0) {
   ret= Asagent_get_permission(agent,"env","append_nondash",0);
   if(ret<=0)
     goto ex;
   if(strcmp(cmd+16,"on")==0) {
     icmp->append_nondash= 1;
   } else if(strcmp(cmd+16,"off")==0) {
     icmp->append_nondash= 0;
   } else {
     Imgcmp_stderr(icmp,"unknown option to -append_nondash : ",1);
     Imgcmp_stderr(icmp,cmd+16,0);
     ret= 0; goto ex;
   }
 } else if(strncmp(cmd,"-auto_client",12)==0 && cmd[12]!=0) {
   ret= Asagent_get_permission(agent,"env","auto_client",0);
   if(ret<=0)
     goto ex;
   if(strcmp(cmd+13,"on")==0)
     yes= 0;
   else
     yes= 1;
   ret= Imgcmp_set_client_intern(icmp,yes,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-bundle",7)==0 && cmd[7]!=0) {
   if(strcmp(cmd+8,"start")==0) {
     icmp->bundle_open= 1;
     icmp->bundle_count++;
     if(Imgmap_get_guarding(imp,0)>0) 
       fprintf(stderr,"\r%79.79s\r  %ld ","",icmp->bundle_count);
   } else {
     icmp->bundle_open= 0;
     if(Imgmap_get_guarding(imp,0)>0) 
       fprintf(stderr,"\r+ %ld\r",icmp->bundle_count);
   }
   no_prepend= 1;
 } else if(strncmp(cmd,"-caching",8)==0 && cmd[8]!=0) {
   ret= Asagent_get_permission(agent,"dbadmin","caching",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   Imgmap_delete_cache(imp,0);
   if(strcmp(cmd+9,"on")==0){
     imp->caching= 1;
     ret= Imgmap_load_cache(imp,0);
     if(ret<=0)
       goto ex;
   } else {
     imp->caching= 0;
   }
 } else if(strncmp(cmd,"-cd",3)==0 && cmd[3]!=0) {
   fnamept= 3;
   ret= Asagent_get_permission(agent,"env","cd",0);
   if(ret<=0)
     goto ex;
   strcpy(filename,cmd+fnamept+1);
   if(chdir(filename)!=0) {
     sprintf(line,"chdir(\"%s\") failed. errno=%ld\n",filename,errno);
     Imgcmp_stderr(icmp,line,0);
     ret= 0; goto ex;
   }
 } else if(strncmp(cmd,"-database",9)==0 && cmd[9]!=0) {
   fnamept= 9;
   ret= Imgcmp_cmd_mapname(icmp,cmd+fnamept+1,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-del_back",9)==0 && cmd[9]!=0) {
   fnamept= 9;
   ret= Asagent_get_permission(agent,"modify","del_back",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   strcpy(filename,cmd+fnamept+1);
   if(Imgmap_del_back(imp,filename,0)>0) {
     sprintf(line,"deleted from imgmap : '%s'\n",filename);
     Imgcmp_stderr(icmp,line,0);
   } else {
     Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
   }
 } else if(strncmp(cmd,"-del_dir",8)==0 && cmd[8]!=0) {
   fnamept= 8;
   ret= Asagent_get_permission(agent,"modify","del_dir",0);
   if(ret<=0)
     goto ex;
   strcpy(filename,cmd+fnamept+1);
   rm_ret= Imgmap_rm_filename(imp,filename,1|4);
   if(rm_ret!=0) {
     sprintf(line,
     "deleted from imgmap : all image files in directory '%s'  (%ld entries)\n",
                    filename,(rm_ret>0?rm_ret:-rm_ret));
     Imgcmp_stderr(icmp,line,0);
   } else {
     Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
   }
 } else if(strncmp(cmd,"-del_file",9)==0 && cmd[9]!=0 && cmd[9]!='_') {
   fnamept= 9;
   ret= Asagent_get_permission(agent,"modify","del_file",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   sprintf(line,"-del_file %s :",cmd+10);
   Imgcmp_stderr(icmp,line,0);
   filename_pt= filename;
   ret= Imgcmp_offadr(icmp,cmd+fnamept+1,&filename_pt,&foundsample,1,1|4);
   if(ret<0) 
     goto ex;
   count= ret;
   if(count==1){
     if(foundsample==NULL)
       rm_ret= Imgmap_rm_filename(imp,filename,0);
     else
       rm_ret= Imgmap_rm_sample(imp,foundsample,0);
     if(rm_ret!=0){
       sprintf(line,"deleted from imgmap : '%s'  (%ld entries)\n",
                     filename,(rm_ret>0?rm_ret:-rm_ret));
       Imgcmp_stderr(icmp,line,0);
     } else {
       Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
     }
   } else if(count==0){
     sprintf(line,"--- nothing deleted. No matching file found.");
     Imgcmp_stderr(icmp,line,0);
   } else {
     sprintf(line,"--- not deleted. More than one matching file address.");
     Imgcmp_stderr(icmp,line,0);
   }
 } else if(strncmp(cmd,"-del_file_adr",13)==0 && cmd[13]!=0) {
   fnamept= 13;
   ret= Asagent_get_permission(agent,"modify","del_file_adr",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   strcpy(filename,cmd+fnamept+1);
   rm_ret= Imgmap_rm_filename(imp,filename,0);
   if(rm_ret!=0){
     sprintf(line,"deleted from imgmap : '%s'  (%ld entries)\n",
                   filename,(rm_ret>0?rm_ret:-rm_ret));
     Imgcmp_stderr(icmp,line,0);
   } else {
     Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
   }
 } else if(strncmp(cmd,"-del_tree",9)==0 && cmd[9]!=0) {
   fnamept= 9;
   ret= Asagent_get_permission(agent,"modify","del_tree",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   strcpy(filename,cmd+fnamept+1);
   rm_ret= Imgmap_rm_filename(imp,filename,8);
   if(rm_ret!=0){
     sprintf(line,"deleted from imgmap : all image files in and below directory '%s'  (%ld entries)",
                  filename,(rm_ret>0?rm_ret:-rm_ret));
     Imgcmp_stderr(icmp,line,0);
   } else {
     Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
   }
 } else if(strncmp(cmd,"-distinguish",12)==0 && cmd[12]!=0) {
   if(client_mode) {
     ret= Asagent_get_permission(agent,"dbadmin","distinguish",0);
     if(ret<=0)
       goto ex;
     goto remote_cmd;
   }
   ret= Imgcmp_cmd_distinguish(icmp,cmd+13,0);
   if(ret<=0)
     goto ex;
 } else if(strcmp(cmd,"-end")==0) {
   ret= Asagent_get_permission(agent,"end","end",2);
   if(ret<=0)
     goto ex;
   if(Asagent_get_service_mode(agent,0)==1) {
     ret= Asagent_set_end_on(agent,"mark:_imgcmp_all_commands_done_",1);
     if(ret<=0)
       goto ex;
     ret= Asagent_send_cmd(icmp->agent,"-mark:_imgcmp_all_commands_done_:",1);
     if(ret<0)
       goto ex;
     ret= 1;
   } else {
     ret= 3;
   }
   goto ex;
 } else if(strcmp(cmd,"-garbage_collection")==0) {
   ret= Asagent_get_permission(agent,"dbadmin","garbage_collection",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   ret= Imgmap_garbage_coll(imp,0);
   if(ret<=0)
     {ret= 0; goto ex;}
   if(Imgmap_get_errmsg(imp,line,0)[0]!=0)
     Imgcmp_stderr(icmp,line,0);
 } else if(strncmp(cmd,"-guarded_map",12)==0 && cmd[12]!=0) {
   ret= Asagent_get_permission(agent,"service","guarded_map",0);
   if(ret<=0)
     goto ex;
   ret= Imgmap_get_mapname(imp,filename,5);
   if(ret<=0)
     goto ex;
   if(Sfile_size(filename,1)!=-1) {
     sprintf(line,"File '%s' already exists.",filename);
     Imgcmp_stderr(icmp,line,0);
     ret= 0; goto ex;
   } else if(Imgmap_get_disabled(imp,0)>0) {
     ret= Imgmap_get_mapname(imp,filename,0);
     if(ret<=0)
       goto ex;
     sprintf(line,"Access to map file '%s' is disabled.",filename);
     Imgcmp_stderr(icmp,line,0);
     ret= 0; goto ex;
   }

   /* >>> Sfile_get_lock(filename,(char *) icmp, microsec ,"guarded map",0); */

   if(strcmp(cmd+13,"tcp")==0) {
     ret= Asagent_publish_tcpadr(agent,filename,2,0);
     if(ret<=0)
       goto ex;
   } else {
     sprintf(line,"Inappropriate connection type '%s' with -guarded_map",
                  cmd+13);
     Imgcmp_stderr(icmp,line,0);

     /* >>> remove locked file */

     ret= 0; goto ex;
   }
   Imgmap_set_guarding(imp,1,0);
 } else if(strncmp(cmd,"-help",5)==0){
   ret= Imgcmp_cmd_help(icmp,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-imagesize",10)==0 && cmd[10]!=0) {
   ret= Imgcmp_cmd_imagesize(icmp,cmd+11,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-karte",6)==0 && cmd[6]!=0) {
   fnamept= 6;
   ret= Imgcmp_cmd_mapname(icmp,cmd+fnamept+1,0);
   if(ret<=0)
     goto ex;
 } else if(strcmp(cmd,"-list_all_files")==0) {
   ret= Asagent_get_permission(agent,"info","list_all_files",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   Imgmap_search_reset(imp,0);
   filename[0]= 0;
   while(Imgmap_search_filename(imp,filename,samp,8|32)>0){
     Imgcmp_result(icmp,filename,0);
     filename[0]= 0;
   }
   if(Imgmap_get_errmsg(imp,line,0)[0])
     Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
 } else if(strcmp(cmd,"-list_doubles")==0) {
   ret= Asagent_get_permission(agent,"info","list_doubles",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   ret= Imgcmp_list_doubles(icmp,0);
   if(ret<0)
     goto ex;
   sprintf(line,"number of suspects : %ld",ret);
   Imgcmp_stderr(icmp,line,0);
 } else if(strncmp(cmd,"-lookup",7)==0 && cmd[7]!=0 && cmd[7]!='_') {
   fnamept= 7;
   ret= Asagent_get_permission(agent,"info","lookup",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   strcpy(filename,cmd+fnamept+1);
   if(Imgmap_lookup(imp,filename,samp,0)>0){
     sprintf(line,"lookup '%s' : ", filename);
     Imgcmp_result(icmp,line,1);
     for(j=0;j<Imgmap_get_spoints(imp,0);j++) {
       sprintf(line,"r=%3.3lf g=%3.3lf b=%3.3lf   ",
                    Imgsample_get_coord(samp,0,j,0),
                    Imgsample_get_coord(samp,1,j,0),
                    Imgsample_get_coord(samp,2,j,0));
       Imgcmp_result(icmp,line,1);
     }
     Imgcmp_result(icmp,"",0);
   } else {
     Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
   }
 } else if(strncmp(cmd,"-lookup_adr",11)==0 && cmd[11]!=0 && cmd[11]!='_') {
   fnamept= 11;
   ret= Asagent_get_permission(agent,"info","lookup_adr",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   filename_pt= filename;
   count= Imgcmp_offadr(icmp,cmd+fnamept+1,&filename_pt,&foundsample,1,0);
   if(count>0) 
     Imgcmp_result(icmp,filename,0);
   else {
     sprintf(line,"No matching file found for : %s",cmd+12);
     Imgcmp_stderr(icmp,line,0);
   }
 } else if(strncmp(cmd,"-lookup_adr_all",15)==0 && cmd[15]!=0) {
   fnamept= 15;
   ret= Asagent_get_permission(agent,"info","lookup_adr_all",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   filename_pt= filename;
   count= Imgcmp_offadr(icmp,cmd+fnamept+15,&filename_pt,&foundsample,1,2);
   if(count==0) {
     sprintf(line,"No matching file found for : %s",cmd+16);
     Imgcmp_stderr(icmp,line,0);
   }
 } else if(strncmp(cmd,"-lookup_adr_count",17)==0 && cmd[17]!=0) {
   fnamept= 17;
   ret= Asagent_get_permission(agent,"info","lookup_adr_count",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   filename_pt= filename;
   count= Imgcmp_offadr(icmp,cmd+fnamept+1,&filename_pt,&foundsample,1,0);
   sprintf(line,"%ld",count);
   Imgcmp_result(icmp,line,0);
 } else if(strncmp(cmd,"-mapname",8)==0 && cmd[8]!=0) {
   fnamept= 8;
   ret= Imgcmp_cmd_mapname(icmp,cmd+fnamept+1,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-match_par",10)==0 && cmd[10]!=0) {
   if(client_mode) {
     ret= Asagent_get_permission(agent,"service","match_par",0);
     if(ret<=0)
       goto ex;
     goto remote_cmd;
   }
   ret= Imgcmp_cmd_match_par(icmp,cmd+11,0);
   if(ret<=0)
     goto ex;
 } else if(strcmp(cmd,"-noop")==0) { 
   ret= Asagent_get_permission(agent,"client","noop",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   if(!icmp->bundle_open) {
     icmp->bundle_count++;
     if(Imgmap_get_guarding(imp,0)>0) 
       fprintf(stderr,"\r%79.79s\r+ %ld ","",icmp->bundle_count);
   }
 } else if(strncmp(cmd,"-refresh",8)==0 && cmd[8]!=0) { 
   fnamept= 8;
   ret= Asagent_get_permission(agent,"modify","refresh",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   strcpy(filename,cmd+fnamept+1);
   ret= Imgcmp_append(icmp,filename,1);
   if(ret<=0){
     Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
     goto ex;
   }
 } else if(strncmp(cmd,"-search",7)==0 && cmd[7]!=0 && cmd[7]!='_') {
   fnamept= 7;
   ret= Asagent_get_permission(agent,"info","search",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   strcpy(filename,cmd+fnamept+1);
   if(stat(filename,&stbuf)==-1) {
     sprintf(line,"cannot determine properties of file '%s'",filename);
     Imgcmp_stderr(icmp,line,0);
     ret= 0; goto ex;
   }
   inode_arg= stbuf.st_ino;
   dev_arg= stbuf.st_dev;
   ret= Imgmap_read_rgb(imp,filename,samp,0);
   if(ret<=0) 
     goto ex;
   if(strlen(Imgmap_get_errmsg(imp,line,0))>0)
     Imgcmp_stderr(icmp,line,0);
   if(ret<=0)
     goto ex;
   Imgmap_search_reset(imp,0);
   while(Imgmap_search(imp,filename,samp,0)>0){
     if(icmp->find_self){
       Imgcmp_result(icmp,filename,0);
     }else{
       inode_found= !inode_arg;
       if(stat(filename,&stbuf)!=-1) {
         inode_found= stbuf.st_ino;
         dev_found= stbuf.st_dev;
       }
       if(inode_found!=inode_arg || dev_arg!=dev_found)
         Imgcmp_result(icmp,filename,0);
     }
   }
   if(strlen(Imgmap_get_errmsg(imp,line,0))>0)
     Imgcmp_stderr(icmp,Imgmap_get_errmsg(imp,line,0),0);
 } else if(strcmp(cmd,"-search_others_only")==0) {
   ret= Asagent_get_permission(agent,"env","search_others_only",0);
   if(ret<=0)
     goto ex;
   icmp->find_self= 0;
 } else if(strcmp(cmd,"-search_self_too")==0) {
   ret= Asagent_get_permission(agent,"env","search_self_too",0);
   if(ret<=0)
     goto ex;
   icmp->find_self= 1;
 } else if(strcmp(cmd,"-statistics")==0){
   if(client_mode) {
     ret= Asagent_get_permission(agent,"info","statistics",0);
     if(ret<=0)
       goto ex;
     goto remote_cmd;
   }
   ret= Imgcmp_cmd_statistics(icmp,0);
   if(ret<=0)
     goto ex;
 } else if(strcmp(cmd,"-theory")==0){
   ret= Imgcmp_cmd_theory(icmp,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-use_exemptlist",15)==0 && cmd[15]!=0) {
   ret= Asagent_get_permission(agent,"dbadmin","use_exemptlist",0);
   if(ret<=0)
     goto ex;
   if(client_mode)
     goto remote_cmd;
   if(strcmp(cmd+16,"on")==0){
     ret= Imgmap_activate_exempt(imp,0);
     if(ret<0)
       goto ex;
     Imgmap_set_use_exemptlist(imp,1,0);
   } else if(strcmp(cmd+16,"off")==0) {
     Imgmap_set_use_exemptlist(imp,0,0);
   } else {
     Imgcmp_stderr(icmp,"unknown option to -use_exemptlist : ",1);
     Imgcmp_stderr(icmp,cmd+16,0);
     ret= 0; goto ex;
   }
 } else {
   ret= Asagent_interpreter(icmp->agent,1|2|(flag&(4|8)));
   if(ret<=0 || ret==3)
     goto ex;
   if(ret==4) {
     if(icmp->append_nondash) {
       ret= Asagent_get_permission(agent,"modify","append",0);
       if(ret<=0)
         goto ex;
       if(client_mode)
         goto remote_cmd;
       ret= Imgcmp_append(icmp,cmd,0);
       if(ret<0)
         goto ex;   
     } else {
       sprintf(line,"--- command not recognized : '%s'",cmd);
       Imgcmp_stderr(icmp,line,0);
       goto ex;
     }
   }
 }
 if(ret==0)
   ret= 1;
ex:;
 Sregex_string(&cmd,NULL,0);
 Sregex_string(&username,NULL,0);
 Imgsample_destroy(&samp,0);
 return(ret);

remote_cmd:;
 /* get original command again */
 ret= Asevent_get_cmd(agent->event,&cmd,0);
 if(ret<=0)
   goto ex;

 ret= Imgcmp_send_cmd(icmp,cmd,fnamept,no_prepend);
 if(ret<=0)
   goto ex;

 ret= 1; goto ex;
}


