
/*
 cc -g -c imgsamp.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

#include "imgrgb.h"

#include "imgsamp.h"



/* ------------------------------- ImgsamplE ------------------------------ */


int Imgsample_new(sample,flag)
struct ImgsamplE **sample;
int flag;
{
 int i;
 struct ImgsamplE *s;

 *sample= s= TSOB_FELD(struct ImgsamplE,1);
 if(s==NULL)
   return(-1);
 s->invalid= 15;
 s->id= 0;
 s->samp_adr= 0;
 s->nam_adr= 0;
 s->filename= NULL;
 s->filename_is_reference= 0;
 s->crc= 0;
 s->spoints= 0;
 for(i=0;i<Imgsamp_max_sizE;i++) 
   s->red[i]= s->green[i]= s->blue[i]= 0;
 s->prev= s->next= NULL;
 return(1);
}


int Imgsample_destroy(sample,flag)
struct ImgsamplE **sample;
int flag;
{
 if(*sample==NULL)
   return(0);
 if((*sample)->filename!=NULL && !(*sample)->filename_is_reference)
   Sregex_string(&((*sample)->filename),NULL,0);
 Smem_freE((char *) *sample);
 *sample= NULL;
 return(1);
}


int Imgsample_copy(source,target,flag)
struct ImgsamplE *source;
struct ImgsamplE *target;
int flag;
/*
 bit0= establish reference link to filename rather than copying it
*/
{
 int i,ret;

 target->invalid= source->invalid;
 target->id= source->id;
 target->samp_adr= source->samp_adr;
 target->nam_adr= source->nam_adr;
 ret= Imgsample_set_filename(target,source->filename,flag&1);
 if(ret<=0)
   return(ret);
 target->crc= source->crc;
 target->spoints= source->spoints;
 for(i=0;i<Imgsamp_max_sizE;i++) {
   target->red[i]= source->red[i];
   target->green[i]= source->green[i];
   target->blue[i]= source->blue[i];
 }
 return(ret);
}


int Imgsample__set_coord(coord,value,flag)
Imgsamp_coorD *coord;
double value;
int flag;
{
 if(value>=Imgsamp_scalE)
   *coord= Imgsample_max_coorD;
 else if(value<0.0)
   *coord= 0;
 else
  *coord= value*Imgsamp_scalE;
 return(1);
}


double Imgsample_get_coord(sample,color,idx,flag)
struct ImgsamplE *sample;
int color; /* 0=red , 1=green , 2=blue */
int idx;
int flag;
{
 if(idx>=sample->spoints || idx<0)
   return(0.0);
 switch (color) {
 case 0: return(((double) sample->red[idx])/((double) Imgsamp_scalE));
 case 1: return(((double) sample->green[idx])/((double) Imgsamp_scalE));
 case 2: return(((double) sample->blue[idx])/((double) Imgsamp_scalE));
 }
 return(0.0);
}


int Imgsample_set_rgb(sample,spoints,red,green,blue,flag)
struct ImgsamplE *sample;
int spoints;
double *red,*green,*blue;
int flag;
{
 int i;

 if(spoints>Imgsamp_max_sizE)
   return(0);
 sample->spoints= spoints;
 for(i=0;i<spoints;i++){
   Imgsample__set_coord(&(sample->red[i]),red[i],0);
   Imgsample__set_coord(&(sample->green[i]),green[i],0);
   Imgsample__set_coord(&(sample->blue[i]),blue[i],0);
 }
 sample->invalid&= ~8;
 return(1);
}


int Imgsample_set_nam_adr(sample,nam_adr,flag)
struct ImgsamplE *sample;
unsigned long nam_adr;
int flag;
{
 sample->nam_adr= nam_adr;
 sample->invalid&= ~2;
 return(1);
}


int Imgsample_set_samp_adr(sample,samp_adr,flag)
struct ImgsamplE *sample;
unsigned long samp_adr;
int flag;
{
 sample->samp_adr= samp_adr;
 sample->invalid&= ~1;
 return(1);
}


int Imgsample_set_filename(sample,filename,flag)
struct ImgsamplE *sample;
char *filename;
int flag;
/*
 bit0= establish reference link to filename rather than copying it
*/
{
 if(flag&1) {
   if(!sample->filename_is_reference)
     Sregex_string(&(sample->filename),NULL,0);
   sample->filename= filename;
   sample->filename_is_reference= 1;
 } else {
   sample->filename_is_reference= 0;
   if(Sregex_string(&(sample->filename),filename,0)<=0)
     return(-1);
 }
 return(1);
}


