
/*
  cc -g -c nntpclient.c
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include "../s_tools/smem.h"

#include "../s_tools/sregex.h"
#include "../s_tools/sfile.h"

#include "nntpconn.h"
#include "nntpfiler.h"
#include "nntphead.h"
#include "nntpnov.h"
#include "nntpfilter.h"
#include "nntpload.h"
#include "nntpgroup.h"
#include "nntppost.h"
#include "nntpstats.h"

#include "nntpclient.h"


/* --------------------- Nntpclient -------------------- */


int Nntpclient_new(client,servername,flag)
struct NntpclienT **client;
char *servername;
int flag;
/*
 bit0= do not establish connection to server
*/
{
 int ret;
 struct NntpclienT *c;

 *client= c= TSOB_FELD(struct NntpclienT,1);
 if(c==NULL) {
   fprintf(stderr,"Cannot create client object : %s\n",strerror(errno));
   return(-1);
 }
 strcpy(c->progname,"nntpclient");
 getcwd(c->start_dir,sizeof(c->start_dir));
 strcpy(c->servername,servername);
 c->port= 119;
 c->conn= NULL;
 c->timeout= 86400;
 c->timeout_bell= 1;
 c->r_baudlimit= 0.0;
 c->use_nov= 1;
 c->nov= NULL;
 c->login[0]= 0;
 c->password[0]= 0;
 c->group= NULL;
 c->stats= NULL;
 c->head= NULL;
 c->loadjob= NULL;
 c->download_dir[0]= 0;
 c->overwrite= 0;
 c->html= 1;
 c->short_form= 1;
 c->filer= NULL;
 Nntpfilter_new(&(c->filter),0,0x7FFFFFFF,2);
 c->use_own_msgid= 0;
 strcpy(c->hostid,"somewhere");
 strcpy(c->username,"anonymous");
 strcpy(c->organization,"none");
 strcpy(c->reference,"");
 c->posting_style= 0;
 c->user_messages[0]= 0;
 c->argc= 0;
 c->argv= NULL;
 c->argidx= -1;
 c->check_for_short_article= 1;
 c->check_for_slow_server= 1;
 c->short_article_counter= 0;
 c->short_article_maxtry= 5;
 c->slow_server_minsize= 50000.0;
 c->slow_server_limit= 120.0;

 ret= Sfile_home_adr(Nntpclient_basediR,c->basedir,0);
 if(ret<=0) {
   Nntpclient_destroy(client,0);
   return(ret);
 } 
 if(!(flag&1)){
   ret= Nntpclient_open(c,servername,0);
   if(ret<=0){
     Nntpclient_destroy(client,0);
     return(ret);
   }
 } else {
   ret= Nntpconn_new(&(c->conn),c->servername,c->port,2);
   if(ret<=0){
     Nntpclient_destroy(client,0);
     return(ret);
   }
    Nntpconn_set_auth(c->conn,c->login,c->password,0);
 }
 Nntpconn_set_timeout(c->conn,c->timeout,0);
 ret= Nntpfiler_new(&(c->filer),c->conn,c->basedir,"",!!c->html);
 if(ret<=0){
   Nntpclient_destroy(client,0);
   return(ret);
 }
 ret= Nntpstats_new(&(c->stats),NULL,NULL,NULL,0);
 if(ret<=0){
   Nntpclient_destroy(client,0);
   return(ret);
 }
 Nntpconn_set_filer(c->conn,c->filer,0);
 return(1);
}


int Nntpclient_destroy(client,flag)
struct NntpclienT **client;
int flag;
{ 
 struct NntpclienT *c;
 
 c= *client;
 if(c==NULL)
   return(0);
 Nntpclient_close(c,0);
 if(c->group!=NULL)
   Nntpgroup_destroy(&(c->group),0);
 if(c->head!=NULL)
   Nntphead_destroy(&(c->head),0);
 if(c->loadjob!=NULL)
   Nntpload_destroy(&(c->loadjob),0);
 if(c->filter!=NULL)
   Nntpfilter_destroy(&(c->filter),0);
 if(c->filer!=NULL)
   Nntpfiler_destroy(&(c->filer),0);
 if(c->stats!=NULL)
   Nntpstats_destroy(&(c->stats),0);
 free((char *) c);
 *client= NULL;
 return(1);
}


int Nntpclient_open(client,servername,flag)
struct NntpclienT *client;
char *servername;
int flag;
{
 int ret;

 Nntpclient_close(client,0);
 client->servername[0]= 0;
 ret= Nntpconn_new(&(client->conn),servername,client->port,2);
 if(ret<=0)
   return(ret);
 Nntpconn_set_auth(client->conn,client->login,client->password,0);
 Nntpconn_set_timeout(client->conn,client->timeout,0);
 Nntpconn_set_slow_server(client->conn,client->slow_server_limit,0);
 Nntpconn_set_r_baudlimit(client->conn,client->r_baudlimit,0);
 Nntpconn_set_filer(client->conn,client->filer,0);
 Nntpfiler_set_conn(client->filer,client->conn,0);
 ret= Nntpconn_open(client->conn,0);
 if(ret<=0)
   return(ret);
 if(client->servername!=servername)
   strcpy(client->servername,servername);
 return(1);
}


int Nntpclient_close(client,flag)
struct NntpclienT *client;
int flag;
{
 int ret;

 if(client->conn==NULL)
   return(0);

 ret= Nntpconn_destroy(&(client->conn),0);
 return(ret);
}


int Nntpclient_set_progname(client,progname,flag)
struct NntpclienT *client;
char *progname;
int flag;
{
 if(strlen(progname)>=sizeof(client->progname))
   return(0);
 strcpy(client->progname,progname);
 return(1);
}


int Nntpclient_set_timeout(client,timeout,flag)
struct NntpclienT *client;
int timeout;
int flag;
{
 client->timeout= timeout;
 Nntpconn_set_timeout(client->conn,timeout,0);
 return(1);
}


int Nntpclient_set_r_baudlimit(client,limit,flag)
struct NntpclienT *client;
double limit;
int flag;
{
 client->r_baudlimit= limit;
 Nntpconn_set_r_baudlimit(client->conn,limit,0);
 return(1);
}


int Nntpclient_set_auth(client,login,password,flag)
struct NntpclienT *client;
char *login,*password;
int flag;
{
 strcpy(client->login,login);
 strcpy(client->password,password);
 Nntpconn_set_auth(client->conn,login,password,0);
 return(1);
}
  

int Nntpclient_open_group(client,groupname,flag)
struct NntpclienT *client;
char *groupname;
int flag;
/*
 bit0= error if not NEXT message readable
 bit1= do not read header from server
*/
{
 int ret,gflag= 0;
 char cmd[Nntpclient_strmaX],name[Nntpgroup_strmaX];
 struct NntpconN *conn;

 conn= client->conn;

 strcpy(name,groupname);        /* possible:  groupname==client->group->name */
 if(client->group!=NULL)
   Nntpgroup_destroy(&(client->group),0);       /* kills client->group->name */
 ret= Nntpgroup_new(&(client->group),conn,name,0);      /* but name survived */
 if(ret<=0)
   return(ret);
 Nntpfiler_set_groupname(client->filer,name,0);
 if(client->head!=NULL)
   Nntphead_destroy(&(client->head),0);
 if(flag&2) {
   ret= Nntphead_new(&(client->head),conn,"",1);
   if(ret<=0)
     return(ret);
   return(1);
 }
 ret= Nntpconn_transaction(conn,"NEXT","",0,"",0);
 if(ret<=0)
   return(ret);
 if(conn->status_code!=223){
   if(flag&1){
     conn->errn= -8;
     sprintf(conn->errmsg,
             "Cannot access first message in group %s :\n%s",
             groupname,conn->status_line);
     return(0);
   }
   gflag= 1;
 }
 ret= Nntphead_new(&(client->head),conn,"",2|gflag);
 if(ret<=0)
   return(ret);
 return(1);
}


int Nntpclient_keep_alive(client,flag)
struct NntpclienT *client;
int flag;
{
 int ret;

 fprintf(stderr,"keeping connection alive with ");
 if(client->group!=NULL)
   if(client->group->read) {
     fprintf(stderr,"NEXT ...\n");
     ret= Nntpconn_transaction(client->conn,"NEXT","",0,"",0);
     fprintf(stderr,"done\n");
     if(ret<=0)
       goto ex;
     return(1);
   }
 if(client->conn->active){
     fprintf(stderr,"HELP ...\n");
   ret= Nntpconn_transaction(client->conn,"HELP","",0,"",0);
   fprintf(stderr,"done\n");
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Nntpclient_open_message(client,msgno,msgname,flag)
struct NntpclienT *client;
int msgno;
char *msgname;
int flag;
/*
 bit0= permission to use  client->nov  if valid
 bit1= do not see invalid article as error
 bit2= reopening of the same article. preserve byte_count
*/
{
 int ret,byte_count_merk;
 char count_text[81];

 if(client->group==NULL){
   client->conn->errn= -9;
   sprintf(client->conn->errmsg,
           "Message required but no group opened : messagename= %s",msgname);
   return(0);
 }
 byte_count_merk= client->head->byte_count;

 Nntphead_set_msgadr(client->head,msgno,msgname,0);
 if((flag&1) && client->nov!=NULL) {
   ret= Nntpnov_fill_head(client->nov,client->head,msgno,0);
   if(ret>0)
     client->head->status= 2;
   if(ret==0) ret= 1;
 } else {
   ret= Nntphead_read(client->head,flag&2); 
   if((flag&4) && client->head->byte_count<0)
     client->head->byte_count= byte_count_merk;
 }
 return(ret);
}


int Nntpclient_result(client,text,flag)
struct NntpclienT *client;
char *text;
int flag;
/*
 bit0= to stdout (rather than stderr)
*/
{
 if(flag&1)
   fprintf(stdout,"%s",text);
 else
   fprintf(stderr,"%s",text);
 return(1);
}


int Nntpclient_pause(client,filename,time_to_pause,keep_alive_after,flag)
struct NntpclienT *client;
char filename[4096];
double time_to_pause,keep_alive_after;
int flag;
/*
 bit0= pause  time_to_pause  seconds without regarding control files
*/
{
 struct stat stbuf;
 int i,ret= 1,to_pause;
 FILE *fp;

 if(!(flag&1)){
   ret= Sfile_file_to_double(filename,&time_to_pause,0);
   if(ret<=0)
 return(0);
 }
 to_pause= time_to_pause;
 for(i=1;i<=to_pause;i++) {
   fprintf(stderr,"\r                         \rpause (%ld of %ld)",
                  i,to_pause);
   if(flag&1){
     ;
   }else{
     ret= Sfile_file_to_double(filename,&time_to_pause,0);
     if(ret<=0)
 break;
     to_pause= time_to_pause;
   }
   sleep(1);
   if((i%((int)keep_alive_after))==0) {
     fprintf(stderr,"\r                         \r");
     Nntpclient_keep_alive(client,0);
   }
 }
 ret= 1;
ex:
 fprintf(stderr,"\r                         \rpaused %ld seconds",i-1);
 return(ret);
}


int Nntpclient_read_index(client,filename,flag)
struct NntpclienT *client;
char *filename;
int flag;
{
 int ret,num;

 ret= Sfile_file_to_int(filename,&num,0);
 if(ret<=0 || num<0)
   return(0);
 client->filter->msgno_idx= num;
 return(1);
}


int Nntpclient_read_baudlimit(client,filename,flag)
struct NntpclienT *client;
char *filename;
int flag;
{
 Nntpconn_enforce_limits(client->conn,filename,1|2);
 Nntpclient_set_r_baudlimit(client,client->conn->r_baudlimit,0);
 return(1);
}


int Nntpclient_intrptchk(client,flag)
struct NntpclienT *client;
int flag;
{
 struct stat stbuf;
 char filename[4096];
 double num= -1,time_to_pause= 60.0;
 int ret;

 sprintf(filename,"%s/%s",client->download_dir,"nntpclient_pause");
 if(lstat(filename,&stbuf)!=-1){
   ret= Nntpclient_pause(client,filename,time_to_pause,120.0,0);
   unlink(filename);
   if(ret<=0)
     return(-1);
 }
 sprintf(filename,"%s/%s",client->basedir,"nntpclient_baudlimit");
 if(lstat(filename,&stbuf)!=-1){
   ret= Nntpclient_read_baudlimit(client,filename,0);
   /* unlink(filename); */
   if(ret<=0)
     return(-1);
 }
 sprintf(filename,"%s/%s",client->download_dir,"nntpclient_baudlimit");
 if(lstat(filename,&stbuf)!=-1){
   ret= Nntpclient_read_baudlimit(client,filename,0);
   /* unlink(filename); */
   if(ret<=0)
     return(-1);
 }
 sprintf(filename,"%s/%s",client->download_dir,"nntpclient_index");
 if(lstat(filename,&stbuf)!=-1){
   ret= Nntpclient_read_index(client,filename,0);
   unlink(filename);
   if(ret<=0)
     return(-1);
 }
 sprintf(filename,"%s/%s",client->download_dir,"nntpclient_stop");
 if(lstat(filename,&stbuf)!=-1){
   sprintf(client->user_messages,"interrupt file found : %s",filename);
   unlink(filename);
   return(1);
 }
 sprintf(filename,"%s/%s",client->basedir,"nntpclient_stop");
 if(lstat(filename,&stbuf)!=-1){
   sprintf(client->user_messages,"interrupt file found : %s",filename);
   unlink(filename);
   return(1);
 }
 sprintf(filename,"%s/%s",client->basedir,"nntpclient_stop_all");
 if(lstat(filename,&stbuf)!=-1){
   sprintf(client->user_messages,"interrupt file found : %s",filename);
   return(1);
 }
 return(0);
}


int Nntpclient_std_download_dir(client,flag)
struct NntpclienT *client;
int flag;
/*
 bit0= set working directory to download_dir
*/
{
 int ret;

 if(client->group==NULL)
   return(0);

 Nntpfiler_std_dir(client->filer,client->download_dir,0);
 Nntpgroup_set_dir(client->group,client->download_dir,0);
 if(flag&1){
   ret= Nntpgroup_chdir(client->group,0);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Nntpclient_message_id_key(client,adr,adrsiz,flag)
struct NntpclienT *client;
char *adr;
int adrsiz;
int flag;
/*
 bit0= get reference rather than msgid
*/
{
 int l,hc,ret;
 char line[Nntpclient_strmaX];

 strcpy(adr,"");
 if(client->head==NULL)
   return(2);
 if(client->head->status!=2)
   return(2);
 if(flag&1)
   strcpy(line,client->head->reference);
 else
   strcpy(line,client->head->msgid);

 ret= Nntpfiler_message_id_key(client->filer,line,adr,adrsiz,0);
 return(ret);
}


int Nntpclient_set_known_message(client,flag)
struct NntpclienT *client;
int flag;
/*
 bit0= remove message id mark
 bit1= message has been loaded : 
         write message file address relative to basedir and some header items
*/
{
 int ret,l;
 char adr[SfileadrL],line[SfileadrL],msgadr[SfileadrL];

 ret= Nntpclient_message_id_key(client,adr,sizeof(adr),0);
 if(ret<=0)
   return(ret);
 if(ret==2) /* no usable message id */
   return(0); 
 if(flag&1){
   unlink(adr);
 }else{
   if(flag&2){
     ret= Nntpload_msg_target(client->loadjob,line,msgadr,
                              client->loadjob->head->msgno,1);
     if(ret<=0)
       return(ret);
     strcpy(line,msgadr+strlen(client->basedir)+1);
     strcat(line,"\n");
     ret= Sfile_string_to_file(adr,line,1);
     sprintf(line,"References: %s\n",client->loadjob->head->reference);
     ret= Sfile_string_to_file(adr,line,1|2);
     sprintf(line,"Subject: %s\n",client->loadjob->head->subject);
     ret= Sfile_string_to_file(adr,line,1|2);
     sprintf(line,"From: %s\n",client->loadjob->head->from);
     ret= Sfile_string_to_file(adr,line,1|2);
     sprintf(line,"Date: %s\n",client->loadjob->head->date);
     ret= Sfile_string_to_file(adr,line,1|2);
   } else {
     line[0]= 0;
     ret= Sfile_string_to_file(adr,line,1);
   }
 }
 return(ret);
}


int Nntpclient_is_known_message(client,flag)
struct NntpclienT *client;
int flag;
/*
 bit0= mark as known if not already known
*/
{
 int ret,l;
 char adr[SfileadrL],ret_text[SfileadrL],line[SfileadrL];
 struct stat stbuf;

 ret= Nntpclient_message_id_key(client,adr,sizeof(adr),0);
 if(ret<=0)
   return(ret);
 if(ret==2) /* no usable message id */
   return(0); 
 if(lstat(adr,&stbuf)!=-1) {
   return(1);
 }else if(flag&1) {
   /* perform locking */
   ret= Sfile_get_lock(adr,(char *) client,Nntpclient_lock_waiT," loading",0);
   if(ret<=0)
     return(1);
 } 
 return(0);
}


int Nntpclient_msg_target(client,targetname,targetadr,msgno,flag)
struct NntpclienT *client;
char *targetname,*targetadr;
int msgno;
int flag;
/*
 bit0= do not check for existing message
 bit1= do not copy result to targetname
 bit2= do not copy result to targetadr
 bit3= do not check header id 
 bit4= Nntpclient_is_known_message:bit0= claim message id if not already known
*/
{
 int ret,kret;
 char filename[4096],fileadr[4096];
 struct stat stbuf;

 Nntpclient_std_download_dir(client,0);
 ret= Nntpfiler_msg_target(targetname,targetadr,
                          client->download_dir,msgno,flag&7);
 if(ret==1 && !(flag&1)){
   kret= Nntpclient_is_known_message(client,(flag&16)>>4);
   if(kret<0)
     return(kret);
   if(kret>0) ret= 2;
 }
 if(ret==2 && (!client->overwrite) && !(flag&3)){
   client->conn->errn= -11;
   sprintf(client->conn->errmsg,
           "Message already downloaded (no overwrite) :\n %s",fileadr);
   return(0);
 }
 return(ret);
}


int Nntpclient_fetch_message(client,flag)
struct NntpclienT *client;
int flag;
/*
 bit0= overwrite existing message
 bit1= do not see invalid article as error
 bit3= conn: do not reset transaction count
*/
{
 int ret,message_valid= 0,multi_bin=0;
 char filename[SfileadrL],fileadr[SfileadrL];
 char text[Nntpclient_strmaX];
 struct stat stbuf;
 double t_time= 0.0, t_bytes= 0.0;

 if(client->head->status!=2){
   client->conn->errn= -10;
   sprintf(client->conn->errmsg,
           "Message download required but no message opened");
   return(0);
 }
 sprintf(filename,"%s/%s",client->download_dir,"nntpclient_multi_bin_inline");
 if(stat(filename,&stbuf)!=-1)
   multi_bin= 1;
 sprintf(filename,"%s/%s",client->download_dir,"nntpclient_multi_bin_tee");
 if(stat(filename,&stbuf)!=-1)
   multi_bin= 1;
 sprintf(filename,"%s/%s",client->download_dir,"nntpclient_multi_bin_tee_2dg");
 if(stat(filename,&stbuf)!=-1)
   multi_bin= 2;
 ret= Nntpclient_msg_target(client,filename,fileadr,
                            client->head->msgno,(flag&1)|16);
 if(ret<=0)
   return(ret);

 if(client->loadjob!=NULL)
   Nntpload_destroy(&(client->loadjob),0);
 ret= Nntpload_new(&(client->loadjob),client->conn,client->head,
                   client->download_dir,filename,
                   (client->short_form<<1)|((multi_bin&3)<<2));
 if(ret<=0)
   goto ex;                  
 Nntpload_set_filter(client->loadjob,client->filter,0);
 Nntpload_set_html(client->loadjob,client->html,0);
 Nntpload_set_boss(client->loadjob,client,0);

 ret= Nntpload_article((client->loadjob),(flag&(8|2)));
 if(ret<=0){
   if(client->conn->errn==-19) {
     ret= 2;
     goto ex;
   } if(client->conn->errn==-27) { /* negative reply on ARTICLE HEAD or BODY */
     if(flag&2)
       ret= 3;
     goto ex;
   } else {
     /* Failure : remove possibly downloaded partial message */
     unlink(fileadr);
     /* unlink last downloaded binary file */
     if(client->loadjob!=NULL)
       if(client->loadjob->bin_name!=NULL)
         if(client->loadjob->bin_name[0]!=0)
           unlink(client->loadjob->bin_name);
     goto ex;
   }
 }
 /* if the subjectfile was actually downloaded : record it now */
 if(client->filter->sfrec_dirlist!=NULL
    && client->filter->sfrec_name!=NULL
    && client->filter->sfrec_name_loaded)
   if(client->filter->sfrec_name[0]!=0)
     Sfile_string_to_hashdir(client->filter->sfrec_dirlist,
                             client->filter->sfrec_name,"",1);
 message_valid= 1;
 client->short_article_counter= 0;
 if(client->head->reference[0]!=0) {
   Nntpfiler_add_reference(client->filer,client->head->reference,
                           client->head->msgno,client->head->subject,
                           client->head->from,0);
 }
 ret= 1;
ex:;
 if(client->conn->show_io_stat)
   Nntpconn_show_stat(client->conn,8|16);
 /* close id mark : release locking or remove message id from list */;
 Nntpclient_set_known_message(client,2|!message_valid);
 Nntpconn_get_stat(client->conn,&t_time,&t_bytes,0);
 if(client->check_for_slow_server &&
    t_bytes>client->slow_server_minsize && t_time>0)
   if(t_bytes/t_time<client->slow_server_limit) { 
     sprintf(text,"\n********** SLOW SERVICE **********\n");
     Nntpclient_result(client,text,0);
     if(client->timeout_bell){
       Nntpclient_result(client,"\007",0);
       sleep(1);
       Nntpclient_result(client,"\007",0);
     }
     Nntpclient_result(client,"will try to reconnect in 5 seconds\n",0);
     sleep(5);
     ret= Nntpclient_try_reconnect(client,0);
   }
 return(ret);
}


int Nntpclient_msgmark_adr(client,adr,flag)
struct NntpclienT *client;
char *adr;
int flag;
/*
 bit0-7= mode: 0=list_newest , 1=list_start , 2=remove_old_files_allowed
*/
{
 int mode;
 
 Nntpclient_std_download_dir(client,0);
 mode= flag&255;
 if(mode==0) {
   sprintf(adr,"%s/list_newest",client->download_dir);
 } else if(mode==1) {
   sprintf(adr,"%s/list_start",client->download_dir);
 } else {
   sprintf(adr,"/=/=/Nntpclient_msgmark_adr(): unknown mode %ld/=/=\n",mode);
 }
 return(1);
}


int Nntpclient_write_msgmark(client,msgno,flag)
struct NntpclienT *client;
int msgno;
int flag;
/*
 bit0-7= mode for Nntpclient_msgmark_adr
*/
{
 int ret;
 FILE *fp= NULL;
 char text[4096];
 
 Nntpclient_msgmark_adr(client,text,flag&255);
 fp= fopen(text,"w");
 if(fp==NULL){
   client->conn->errn= errno;
   sprintf(client->conn->errmsg,"Error %ld on opening output file : %s\n%s",
                                errno,text,strerror(errno));
   ret= 0;
   goto ex;
 }
 fprintf(fp,"%ld\n",msgno);
 ret= 1;
ex:;
 if(fp!=NULL) fclose(fp);
 return(ret);
}


int Nntpclient_read_msgmark(client,flag)
struct NntpclienT *client;
int flag;
/*
 bit0-7= mode for Nntpclient_msgmark_adr
 bit8= return(0) if file is missing (else return(-1) and errmsg)
*/
{
 int ret,mode;
 FILE *fp= NULL;
 char text[4096];
 
 Nntpclient_msgmark_adr(client,text,flag&255);
 fp= fopen(text,"r");
 if(fp==NULL){
   if(flag&256) {
     ret= 0;
     goto ex;
   }
   client->conn->errn= errno;
   sprintf(client->conn->errmsg,
           "Error %ld on opening file for reading : %s\n%s",
           errno,text,strerror(errno));
   ret= -1;
   goto ex;
 }
 ret= -1;
 fscanf(fp,"%ld",&ret);
ex:;
 if(fp!=NULL) fclose(fp);
 return(ret);
}


int Nntpclient_read_msgnodef(client,text,flag)
struct NntpclienT *client;
char *text;
int flag;
/*
 bit0= default 0x7FFFFFFF rather than -1
*/
{
 int ret,no_boundary_test= 0;

 if(strcmp(text,"first")==0 || strcmp(text,"start")==0) {
   ret= client->group->first_msgno;
 } else if(strcmp(text,"last")==0 || strcmp(text,"end")==0) {
   ret= client->group->last_msgno;
 } else if(strcmp(text,"list_start")==0) {
   ret= Nntpclient_read_msgmark(client,1);
   if(ret<=0)
     return(-1);
 } else if(strcmp(text,"list_end")==0) {
   ret= Nntpclient_read_msgmark(client,0);
   if(ret<=0)
     return(-1);
 } else if(strcmp(text,"new")==0) {
   ret= Nntpclient_read_msgmark(client,0);
   if(ret<0)
     return(-1);
   ret++;
 } else if(sscanf(text,"%ld",&ret)<=0) {
   if(flag&1) ret= 0x7FFFFFFF;
   else       ret= -1;
   no_boundary_test= 1;
 }
 if(!no_boundary_test && ret>0) {
   if(ret<client->group->first_msgno)
     ret= client->group->first_msgno;
   if(ret>client->group->last_msgno)
     ret= client->group->last_msgno+(flag&1);
 }
 return(ret);
}


Nntpclient_read_intervaldef(client,argv,ard,flag)
struct NntpclienT *client;
char **argv;
int *ard;
int flag;
/*
 bit0= allow empty intervals
*/
{
 int ret;
 (*ard)++;
 ret= Nntpclient_read_msgnodef(client,argv[(*ard)],1);
 if(ret<0)
   return(ret);
 client->filter->msgno_start= ret;
 (*ard)++;
 ret= Nntpclient_read_msgnodef(client,argv[(*ard)],0);
 if(ret<0)
   return(ret);
 client->filter->msgno_end= ret;
 if(client->filter->msgno_start<=0)
   client->filter->msgno_start= 1;
 if(!(flag&1))
   if(client->filter->msgno_start>client->filter->msgno_end) {
     client->conn->errn= -12;
     sprintf(client->conn->errmsg,"Bad message range : %ld to %ld",
                                  client->filter->msgno_start,
                                  client->filter->msgno_end);
     return(0);
   }
 return(1);
}


int Nntpclient_list_result(client,fp,text,flag)
struct NntpclienT *client;
FILE *fp;
char *text;
int flag;
/*
 bit0= do not escape HTML-brackets , omit if not client->html
 bit1= do not print to result channel
 bit2= remove HTML tags from result channel output
*/
{
 int ret;
 char *buffertext,buffertext_data[4096],*de_html_text= NULL;

 if(client->html && !(flag&1)) {
   ret= Nntpfiler_escape_html(buffertext_data,sizeof(buffertext_data),
                              text,0);
   if(ret<=0)
     goto ex;
   buffertext= buffertext_data;
 } else {
   if((flag&1)&&!client->html)
     {ret= 2; goto ex;}
   buffertext= text;
 }

 if(!(flag&2)) {
   if(flag&4) {
     de_html_text= Smem_malloC(strlen(buffertext)+1);
     if(de_html_text==NULL)
       {ret= -1; goto ex;}
     Sregex_de_html(buffertext,de_html_text,0);
   } else {
     if(Sregex_string(&de_html_text,buffertext,0)<=0)
       {ret= -1; goto ex;}
   }
   Nntpclient_result(client,de_html_text,1);
 }
 fprintf(fp,"%s",buffertext);
 ret= 1;
ex:;
 Sregex_string(&de_html_text,NULL,0);
 return(ret);
}


int Nntpclient_fetch(client,msgid,flag)
struct NntpclienT *client;
char *msgid;
int flag;
{
 char text[4096];
 int ret,msgno;

 sprintf(text,"fetching message %s ...\n",msgid);
 Nntpclient_result(client,text,0);
 msgno= -1;
 sscanf(msgid,"%ld",&msgno);
 ret= Nntpclient_open_message(client,msgno,msgid,2);
 if(ret<=0)
   return(ret);
 if(Nntphead_is_validmsg(client->head,0)) {
   ret= Nntpclient_fetch_message(client,8);
   if(ret<=0) {
     if(client->conn->errn==-11) {
       sprintf(text,"... omitted because already downloaded.\n");
       Nntpclient_result(client,text,0);
     } else
       return(ret);
   } else if(ret!=3) {
     sprintf(text,"... message %s downloaded.\n",msgid);
     Nntpclient_result(client,text,0);
   }
 } else {
   sprintf(text,"... message %s : not a valid message\n",msgid);
   Nntpclient_result(client,text,0);
 }
 return(1);
}


int Nntpclient_load_nov(client,flag)
struct NntpclienT *client;
int flag;
{
 char fileadr[SfileadrL],dir[SfileadrL];
 int ret;
 FILE *fp= NULL;

 Nntpfiler_std_dir(client->filer,dir,0);
 sprintf(fileadr,"%s/xover_%-8.8ld_%-8.8ld",
                 dir,client->nov->range_start,client->nov->range_end);
 fp= fopen(fileadr,"r");
 if(fp==NULL)
   return(0);
 ret= Nntpnov_read_file(client->nov,fp,0);
 fclose(fp);
 return(ret);
}
 

int Nntpclient_store_nov(client,flag)
struct NntpclienT *client;
int flag;
{
 char fileadr[SfileadrL],dir[SfileadrL];
 int ret;
 FILE *fp= NULL;

 Nntpfiler_std_dir(client->filer,dir,0);
 sprintf(fileadr,"%s/xover_%-8.8ld_%-8.8ld_open",
                 dir,client->nov->range_start,client->nov->range_end);
 fp= fopen(fileadr,"w");
 if(fp==NULL) {
   client->conn->errn= errno;
   sprintf(client->conn->errmsg,"Error %ld on opening output file : %s\n%s",
                                errno,fileadr,strerror(errno));
   return(0);
 }
 ret= Nntpnov_write_file(client->nov,fp,0);
 fclose(fp);
 if(ret>0){
   /* rename the new file to the official name wthout suffix '_open' */
   strcpy(dir,fileadr);
   fileadr[strlen(fileadr)-5]= 0;
   rename(dir,fileadr);
 }
 return(ret);
}
 

int Nntpclient_fetch_nov(client,msgno_start,msgno_end,flag)
struct NntpclienT *client;
int msgno_start,msgno_end;
int flag;
/*
 bit0= do not save the xover interval to group directory
*/
{
 int ret;

 if(client->nov!=NULL)
   Nntpnov_destroy(&(client->nov),0);
 ret= Nntpnov_new(&(client->nov),client->conn,msgno_start,msgno_end,0);
 if(ret<=0)
   return(ret);

 /* look wether a cached xover interval exists and load */
 ret= Nntpclient_load_nov(client,0);
 if(ret<0)
   goto failed;
 if(ret>0)
   goto success;

 /* read xover from server */
 ret= Nntpnov_read(client->nov,0);
 if(ret<=0)
   goto failed;
 Nntpconn_show_stat(client->conn,2|16);

 /* try to store xover interval to group directory */
 if(!(flag&1))
   Nntpclient_store_nov(client,0);
 
success:;
 ret= 1;
 goto ex;

failed:;
 client->use_nov= 0;
 Nntpnov_destroy(&(client->nov),0);
 Nntpconn_show_stat(client->conn,1);
 goto ex;

ex:;
 return(ret);
}


int Nntpclient_fetch_interval(client,argv,ard,flag)
struct NntpclienT *client;
char **argv;
int *ard;
int flag;
{
 int ret,reject_count,use_nov= 0,idx,try_nov= 0,nov_start=0,nov_end= -1;
 int fetch_flag,do_not_store_nov=0;
 char msgid[Nntpclient_strmaX];
 char line[4096],text[4096];

 if(client->group==NULL){
   client->conn->errn= -9;
   sprintf(client->conn->errmsg,
           "No group opened with -fetch_interval");
   return(0);
 }
 ret= Nntpclient_read_intervaldef(client,argv,ard,1);
 if(ret<=0)
   return(ret);
 if(client->filter->msgno_start>client->filter->msgno_end)
   return(2);

 Nntpclient_std_download_dir(client,0);
 reject_count= 0;

 if(Nntpfilter_is_restrictive(client->filter,0) &&
    Nntpclient_use_nov(client,0)>0){
   try_nov= 1;
 }
 client->filter->msgno_prev[1]= client->filter->msgno_prev[0]= 0;
 for(client->filter->msgno_idx= client->filter->msgno_start; 
     client->filter->msgno_idx<=client->filter->msgno_end; 
     client->filter->msgno_idx++) {
   if(Nntpclient_intrptchk(client,0)){
     client->conn->errn= -14;
     sprintf(client->conn->errmsg,"-fetch_interval stopped because : %s",
                                  client->user_messages);
     return(0);
   }
   /* index can be set externally */
   if(client->filter->msgno_idx>client->filter->msgno_end) 
 break;
   idx= client->filter->msgno_idx;

   /* check wether message file aready exists */
   if(Nntpclient_msg_target(client,"","",idx,2|4|8)<=0) { 
     if(reject_count==0)
       Nntpclient_result(client,"\n",0);
     sprintf(text,"--- omitted because already downloaded: msg%8.8ld\n",idx);
     Nntpclient_result(client,text,0);
     reject_count++;
 continue;
   }

   if(try_nov && (idx<nov_start || idx>nov_end) ){
     do_not_store_nov= 0;
     nov_start= 1+(idx-1)-((idx-1)%Nntpclient_xover_chunksizE);
     if(nov_start<client->group->first_msgno) {
       nov_start= client->group->first_msgno;
       do_not_store_nov= 1;
     }
     nov_end= nov_start+Nntpclient_xover_chunksizE-1;
     if(nov_end>client->group->last_msgno) {
       nov_end= client->group->last_msgno;
       do_not_store_nov= 1;
     }
     ret= Nntpclient_fetch_nov(client,nov_start,nov_end,do_not_store_nov);
     use_nov= (ret>0);
   }

   sprintf(msgid,"%ld",idx);
   ret= Nntpclient_open_message(client,idx,msgid,2|use_nov);
   if(ret<=0)
     return(ret);
   if(!Nntphead_is_validmsg(client->head,0)) {
     if(reject_count==0)
       Nntpclient_result(client,"\n",0);
     sprintf(text,"--- message %s : not a valid message\n",msgid);
     Nntpclient_result(client,text,0);
     reject_count++;
 continue;
   }

   ret= Nntpfilter_head_match(client->filter,client->head,line,0);
   if(ret<0)
     return(ret);
   if(ret==0){
     sprintf(text,"--- message %ld rejected by -filter\n",
                    client->head->msgno);
     Nntpclient_result(client,text,0);
     reject_count++;
 continue;     
   }

   reject_count= 0;
   sprintf(text,"\nfetching message %s from %s\n",
                  msgid,client->group->name);
   Nntpclient_result(client,text,0);
   if(Nntphead_get_property(client->head,"subject:",line,sizeof(line),0)>0){
     sprintf(text,"Subject: %s\n",line);
     Nntpclient_result(client,text,0);
   }

   fetch_flag= 2;   
   ret= Nntpclient_fetch_message(client,fetch_flag);
   if(ret<=0) {
     if(client->conn->errn==-11) {
       if(reject_count==0)
         Nntpclient_result(client,"\n",0);
       sprintf(text,
            "--- omitted because already downloaded: msg%8.8ld.html\n    %s\n",
               idx,client->head->msgid);
       Nntpclient_result(client,text,0);
       reject_count++;
 continue;
     } else if(client->conn->errn==-27){
       if(reject_count==0)
         Nntpclient_result(client,"\n",0);
       sprintf(text,"--- message %s : not a valid message\n",msgid);
       Nntpclient_result(client,text,0);
       reject_count++;
 continue;
     } else
       return(ret);
   } else {
     sprintf(text,"+++ message %s downloaded from %s\n\n",
                    msgid,client->group->name);
     Nntpclient_result(client,text,0);
     client->short_article_counter= 0;
   }
   client->filter->msgno_prev[1]= client->filter->msgno_prev[0];
   client->filter->msgno_prev[0]= idx;
 }

 return(1);
}


int Nntpclient_list_groups(client,flag)
struct NntpclienT *client;
int flag;
{
 int ret;
 char line[4096],dir[4096],fileadr[4096];
 FILE *fp= NULL;
 struct NntpgrouP *group= NULL;

 Nntpfiler_std_dir(client->filer,dir,1);

 sprintf(fileadr,"%s/grouplist",dir);
 fp= fopen(fileadr,"w");
 if(fp==NULL){
   client->conn->errn= errno;
   sprintf(client->conn->errmsg,"Error %ld on opening output file : %s\n%s",
                                errno,fileadr,strerror(errno));
   ret= 0;
   goto ex;
 }

 ret= Nntpconn_transaction(client->conn,"LIST","",0,"",1024);
 if(ret<=0)
   return(ret);
 if(client->conn->status_code!=215){
   client->conn->errn= -8;
   sprintf(client->conn->errmsg,"Cannot get list of groups :\n%s",
           client->conn->status_line);
   ret= 0;
   goto ex;
 }
 while(Nntpconn_getline(client->conn,line,sizeof(line),0)>0){
   Nntpclient_list_result(client,fp,line,0);
   if(Sregex_complines(".",line,0))
 break;
 }

ex:;
 if(fp!=NULL) fclose(fp);
 return(ret);
}


int Nntpclient_use_nov(client,flag)
struct NntpclienT *client;
int flag;
{
 char adr[SfileadrL];
 struct stat stbuf;

 if(client->use_nov<=0)
   return(0);
 if(client->use_nov==1)
   return(1);
 Nntpfiler_std_dir(client->conn->filer,adr,1);
 strcat(adr,"/nntpclient_no_xover");
 if(stat(adr,&stbuf)!=-1)
   return(0);
 return(1);
}


int Nntpclient_list_file(client,flag)
struct NntpclienT *client;
int flag;
/*
 bit0= copy result to list_00000000_000000000.html, remove msg_new.html
*/
{
 int ret,locked= 0,idx;
 int try_nov= 0,do_not_store_nov= 0,use_nov=1,nov_start= 0,nov_end= -1;
 char msgid[Nntpclient_strmaX];
 char line[4096],text[4096],fileadr[4096],msgname[4096],msgadr[4096];
 char linkadr[4096],filename[321],*cpt,link[321];
 FILE *fp= NULL;

 sprintf(filename,"list_%-8.8ld_%-8.8ld.html",
                 client->filter->msgno_start,client->filter->msgno_end);
 sprintf(fileadr,"%s/%s",client->download_dir,filename);
 sprintf(linkadr,"%s/list_00000000_00000000.html",client->download_dir);
 fp= fopen(fileadr,"w");
 if(fp==NULL){
   client->conn->errn= errno;
   sprintf(client->conn->errmsg,"Error %ld on opening output file : %s\n%s",
                                errno,fileadr,strerror(errno));
   ret= 0;
   goto ex;
 }

 sprintf(link,"</PRE><CENTER><KBD><A HREF=\"list_index.html#%s\">",filename);
 strcat(link,"group index</A> | ");
 strcpy(line,client->group->name);
 Sfile_esc_filename(line,sizeof(line),2);
 sprintf(text,"<A HREF=\"../../main_index.html#%s\">",line);
 strcat(text,"main index</A></KBD></CENTER><PRE>\n");
 strcat(link,text);

 sprintf(text,"<HTML><TITLE>%s  %s  %ld %ld</TITLE><BODY><PRE>\n",
              client->servername,client->group->name,
              client->filter->msgno_start,client->filter->msgno_end);
 Nntpclient_list_result(client,fp,text,1|4);
 Nntpclient_list_result(client,fp,link,1|2|4);
 if(client->html){
   sprintf(text,"</PRE><CENTER><KBD>\nServer <B>%s</B>",client->servername);
   Nntpclient_list_result(client,fp,text,1|2|4);
   sprintf(text," , Port %ld<BR>\n",client->port);
   Nntpclient_list_result(client,fp,text,1|2|4);
   sprintf(text,"Group <B>%s</B><BR>\n",client->group->name);
   Nntpclient_list_result(client,fp,text,1|2|4);
   sprintf(text,"Messages %ld to %ld<BR>\n",
                 client->filter->msgno_start,client->filter->msgno_end);
   Nntpclient_list_result(client,fp,text,1|2|4);
   sprintf(text,"<HR><BR><A HREF=\"msg_new.html\">");
   Nntpclient_list_result(client,fp,text,1|2|4);
   sprintf(text,"first downloaded message</A>\n<HR><BR>");
   Nntpclient_list_result(client,fp,text,1|2|4);
   sprintf(text,"</KBD></CENTER><PRE>\n");
   Nntpclient_list_result(client,fp,text,1|2|4);
 }else{
   sprintf(text,"\n\n -server %s  -group %s  -list_interval %ld %ld\n\n\n",
                client->servername,client->group->name,
                client->filter->msgno_start,client->filter->msgno_end);
   Nntpclient_list_result(client,fp,text,0|4);
 }

 try_nov= 0;
 if(Nntpclient_use_nov(client,0)>0)
   try_nov= 1;

 Nntpclient_std_download_dir(client,0);
 client->filter->msgno_prev[1]= client->filter->msgno_prev[0]= 0;
 for(client->filter->msgno_idx= client->filter->msgno_start; 
     client->filter->msgno_idx<=client->filter->msgno_end; 
     client->filter->msgno_idx++) {
   if(Nntpclient_intrptchk(client,0)){
     client->conn->errn= -14;
     sprintf(client->conn->errmsg,"-list_interval stopped because : %s",
                                  client->user_messages);
     ret= 0;
     goto ex;
   }
   /* index can be set externally */
   if(client->filter->msgno_idx>client->filter->msgno_end)
 break;
   idx= client->filter->msgno_idx;
   sprintf(msgid,"%ld",client->filter->msgno_idx);

   if(try_nov && (idx<nov_start || idx>nov_end) ){
     do_not_store_nov= 0;
     nov_start= 1+(idx-1)-((idx-1)%Nntpclient_xover_chunksizE);
     if(nov_start<client->group->first_msgno) {
       nov_start= client->group->first_msgno;
       do_not_store_nov= 1;
     }
     nov_end= nov_start+Nntpclient_xover_chunksizE-1;
     if(nov_end>client->group->last_msgno) {
       nov_end= client->group->last_msgno;
       do_not_store_nov= 1;
     }
     ret= Nntpclient_fetch_nov(client,nov_start,nov_end,do_not_store_nov);
     use_nov= (ret>0);
   }

   ret= Nntpclient_open_message(client,client->filter->msgno_idx,msgid,
                                use_nov|2);
   if(ret<=0)
     goto ex;
   /* publish current end of list */
   ret= Nntpclient_write_msgmark(client,client->filter->msgno_idx,0);
   if(ret<=0)
     goto ex;
   if(!Nntphead_is_validmsg(client->head,0)) {
     sprintf(text,"--- message %s : not a valid message\n\n",msgid);
     Nntpclient_list_result(client,fp,text,0|4);
 continue;
   }
   ret= Nntpfilter_head_match(client->filter,client->head,line,0);
   if(ret<0)
     goto ex;
   if(ret==0){
     sprintf(text,"--- message %ld rejected by -filter\n\n",
                    client->head->msgno);
     Nntpclient_list_result(client,fp,text,0|4);
 continue;
   }
   ret= Nntpclient_msg_target(client,msgname,msgadr,
                              client->head->msgno,1);
   if(ret<=0)
     return(ret);
   sprintf(text,"<A HREF=\"%s\" NAME=\"%s\"> %ld   ",msgname,msgname,
                                                     client->head->msgno);
   Nntpclient_list_result(client,fp,text,1|4);
   Nntpclient_list_result(client,fp,client->head->msgid,0|4);
   Nntpclient_list_result(client,fp,"</A>\r\n",1|4);
   Nntpclient_list_result(client,fp,"<BR>",1|2);
   sprintf(text,"from   : %s\n",client->head->from);
   Nntpclient_list_result(client,fp,text,0|4);
   Nntpclient_list_result(client,fp,"<BR>",1|2);
   sprintf(text,"date   : %s\n",client->head->date);
   Nntpclient_list_result(client,fp,text,0|4);
   Nntpclient_list_result(client,fp,"<BR>",1|2);
   sprintf(text,"size   : ");
   Nntpclient_list_result(client,fp,text,0|4);
   if(Nntphead_get_property(client->head,"bytes:",text,sizeof(text),0)>0)
     if(text[0]!=0 && text[0]!='?') {
       Nntpclient_list_result(client,fp,text,0|4);
       strcpy(text," in ");
       Nntpclient_list_result(client,fp,text,0|4);
     }
   if(Nntphead_get_property(client->head,"lines:",text,sizeof(text),0)>0)
   Nntpclient_list_result(client,fp,text,0|4);
   Nntpclient_list_result(client,fp,"\n",0|4);
   Nntpclient_list_result(client,fp,"<BR>",1|2);
   if(client->html){
     sprintf(text,"subject: <B>");
     Nntpclient_list_result(client,fp,text,1|4);
     sprintf(text,"%s",client->head->subject);
     Nntpclient_list_result(client,fp,text,0|4);
     sprintf(text,"</B>\n");
     Nntpclient_list_result(client,fp,text,1|4);
   }else{
     sprintf(text,"subject: %s\n",client->head->subject);
     Nntpclient_list_result(client,fp,text,0|4);
   }
   Nntpclient_list_result(client,fp,"<BR>",1|2);
   Nntpclient_list_result(client,fp,"\n",0|4);
   Nntpclient_list_result(client,fp,"<BR>",1|2);

   client->filter->msgno_prev[1]= client->filter->msgno_prev[0];
   client->filter->msgno_prev[0]= client->filter->msgno_idx;
 }
 ret= 1;

 sprintf(text,"\nlist also stored in file :\n");
 Nntpclient_result(client,text,0);
 Nntpclient_list_result(client,fp,"<BR>",1|2);
 sprintf(text,"%s\n",fileadr);
 Nntpclient_result(client,text,0);
 Nntpclient_list_result(client,fp,"<BR>",1|2);
 Nntpclient_list_result(client,fp,"<BR><HR><BR>\n",1|2|4);
 Nntpclient_list_result(client,fp,link,1|2|4);
 Nntpclient_list_result(client,fp,"</PRE></HTML></BODY>\r\n",1|2|4);
ex:;
 if(fp!=NULL) fclose(fp);
 if((flag&1) && ret>0){
   /* copy new list to  list_00000000_00000000.html  */
   unlink(linkadr);
   Sfile_copy_file(fileadr,linkadr,
                   &(client->conn->errn),client->conn->errmsg,0);
   sprintf(line,"%s/msg_new.html",client->download_dir);
   unlink(line);
 }
 if(client->nov!=NULL)
   Nntpnov_destroy(&(client->nov),0);
 return(ret);
}


int Nntpclient_list_interval(client,argv,ard,flag)
struct NntpclienT *client;
char **argv;
int *ard;
int flag;
{
 int ret,locked= 0,msgno_start= 0,msgno_end= -1,i;
 char text[4096],fileadr[4096],filename[321],*cpt;

 Nntpclient_std_download_dir(client,0);
 ret= Nntpclient_read_intervaldef(client,argv,ard,1);
 if(ret<=0)
   return(ret);
 if(client->filter->msgno_start>client->filter->msgno_end) {
   ret= Nntpclient_write_msgmark(client,client->filter->msgno_end+1,1);
   if(ret<=0)
     return(ret);
   return(2);
 }

 ret= Nntpfiler_mntc_lock(client->filer,"listing messages",1);
 if(ret<=0)
   goto ex;
 locked= 1;

 /* publish start of list interval */
 ret= Nntpclient_write_msgmark(client,client->filter->msgno_start,1);
 if(ret<=0)
   goto ex;

 /* multiple list files are created if the list is very long */
 msgno_start= client->filter->msgno_start;
 msgno_end= client->filter->msgno_end;
 for(i= msgno_start; i<=msgno_end; i+= Nntpclient_max_list_itemS) {
   client->filter->msgno_start= i;
   client->filter->msgno_end= i+Nntpclient_max_list_itemS-1;
   if(client->filter->msgno_end>msgno_end)
     client->filter->msgno_end= msgno_end;
   ret= Nntpclient_list_file(client,(i==msgno_start));
   if(ret<=0)
     goto ex;
 }

 /* make an index of listfiles */
 ret= Nntpfiler_list_index(client->conn->filer,0);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 client->filter->msgno_start= msgno_start;
 client->filter->msgno_end= msgno_end;
 if(locked)
   Nntpfiler_mntc_lock(client->filer,"listing messages",2);
 if(client->filter->msgno_idx!=client->filter->msgno_end+1) {
   /* correct list name if early abort */
   cpt= strrchr(filename,'_');
   sprintf(cpt+1,"%-8.8ld",client->filter->msgno_idx-1);
   strcpy(text,fileadr);
   cpt= strrchr(fileadr,'_');
   sprintf(cpt+1,"%-8.8ld.html",client->filter->msgno_idx-1);
   rename(text,fileadr);
 }

 return(ret);
}


int Nntpclient_group_stats(client,reportfile,flag)
struct NntpclienT *client;
char *reportfile;
int flag;
{
 struct NntpstatS *stats= NULL;
 int ret;

 if(client->group==NULL){
   client->conn->errn= -9;
   sprintf(client->conn->errmsg,
           "group statistics required but no group opened ");
   return(0);
 }
 if(!Nntpclient_use_nov(client,0)>0){
   client->conn->errn= -31;
   sprintf(client->conn->errmsg,
           "XOVER not enabled for %s.port%ld",
           client->conn->servername,client->conn->portnumber);
   return(0);
 }

 stats= client->stats;
 if(client->stats==NULL){
   ret= Nntpstats_new(&(client->stats),client->conn,client->group,
                      client->head,0);
   if(ret<=0) {
     client->conn->errn= -21;
     sprintf(client->conn->errmsg,"creation of group statistics object failed");
     return(ret);
   }
   stats= client->stats;
 } else {
   Nntpstats_accu_reset(stats,0);
   Nntpstats_set_conn(stats,client->conn,0);
   Nntpstats_set_group(stats,client->group,0);
   Nntpstats_set_head(stats,client->head,0);
 }
 Nntpstats_set_filter(stats,client->filter,0);
 strcpy(stats->largest_article_group,stats->group->name);
 Nntpstats_set_reportfile(stats,reportfile,0);
 if(client->group->first_msgno<=client->group->last_msgno &&
    client->group->first_msgno>0 ) {
   ret= Nntpstats_scan(stats,0);
   if(ret<=0)
     goto ex;
 }
 ret= Nntpstats_report(stats,0);
 if(ret<=0)
   goto ex;

ex:;
 Nntpstats_set_conn(stats,NULL,0);
 Nntpstats_set_group(stats,NULL,0);
 Nntpstats_set_head(stats,NULL,0);
 return(ret);
}


int Nntpclient_grouplist_stats(client,start_expr,expr,reportfile,flag)
struct NntpclienT *client;
char *start_expr,*expr;
char *reportfile;
int flag;
{
 int ret,offst=0,negate= 0, start_negate= 0, started= 0;
 char groupname[Nntpgroup_strmaX],fileadr[SfileadrL],line[SfileadrL];
 struct SregeX *rexpr= NULL, *start_rexpr= NULL;
 struct NntpstatS *sum_stats= NULL, *largest_stats= NULL, *lart_stats= NULL;
 FILE *fp= NULL;

 offst= 0;
 if(strncmp(start_expr,"-not:",5)==0){
   offst= 5;
   start_negate= 1;
 }
 if(Sregex_new(&start_rexpr,start_expr+offst,0)<=0){
err_no_regex:;
   client->conn->errn= -3;
   sprintf(client->conn->errmsg,
           "cannot initialize regular expressions for group statistics");
   ret= -1;
   goto ex;
 }
 offst= 0;
 if(start_expr[0]==0)
   started= 1;
 if(strncmp(expr,"-not:",5)==0){
   offst= 5;
   negate= 1;
 }
 if(Sregex_new(&rexpr,expr+offst,0)<=0)
   goto err_no_regex;
 ret= Nntpstats_new(&sum_stats,client->conn,NULL,NULL,0);
 if(ret<=0) {
err_no_stats:;
   client->conn->errn= -21;
   sprintf(client->conn->errmsg,"creation of group statistics object failed");
   return(ret);
 }
 Nntpstats_set_reportfile(sum_stats,reportfile,0);
 Nntpstats_copy_limits(sum_stats,client->stats);
 ret= Nntpstats_new(&largest_stats,client->conn,NULL,NULL,1);
 if(ret<=0)
   goto err_no_stats;
 Nntpstats_set_reportfile(largest_stats,reportfile,0);
 Nntpstats_copy_limits(largest_stats,client->stats);
 ret= Nntpstats_new(&lart_stats,client->conn,NULL,NULL,1);
 if(ret<=0)
   goto err_no_stats;
 Nntpstats_set_reportfile(lart_stats,reportfile,0);
 Nntpstats_copy_limits(lart_stats,client->stats);

 /* open grouplist in server directory */
 Nntpfiler_std_dir(client->filer,line,1);
 sprintf(fileadr,"%s/grouplist",line);
 fp= Sfile_fopen(fileadr,"rb",1);
 if(fp==NULL){
   client->conn->errn= errno;
   sprintf(client->conn->errmsg,"cannot open '%s' \n  %s",
                                fileadr,strerror(errno));
   ret= 0;
   goto ex;
 }
 while(1){
   /* read line and get group name */
   if(Sfile_fgets(line,sizeof(line),fp)==NULL)
 break;
   Sregex_trimline(line,0);
   if(strcmp(line,".")==0)
 break;
   if(strlen(line)>sizeof(groupname))
 continue;
   sscanf(line,"%s",groupname);
   if(!started){
     if(start_negate^(!Sregex_is_match(start_rexpr,groupname,0)))
 continue;
     started= 1;
   }
   if(negate^(!Sregex_is_match(rexpr,groupname,0)))
 continue;
   ret= Nntpclient_open_group(client,groupname,2);
   if(ret<=0) {
      Nntpstats_print(client->stats,
                 "\n++++++++++++++++++++++++++++++++++++++++++++++++++++\n",0);
      sprintf(line,"\nError %ld with group : %s\n",
                   client->conn->errn,groupname);
      Nntpstats_print(client->stats,line,0);
      Nntpstats_print(client->stats,client->conn->errmsg,0);
      Nntpstats_print(client->stats,
                 "\n++++++++++++++++++++++++++++++++++++++++++++++++++++\n",0);
     if(client->conn->errn==-7) {  /* no such group */
       Nntpconn_errmsg_reset(client->conn,0);
 continue;
     }
     if(client->conn->errn==-6 ) { /* timeout */
       Nntpconn_close(client->conn,1<<15);
       ret= Nntpconn_open(client->conn,0);
       if(ret<=0) {
         client->conn->errn= -36;
         goto ex; 
       }
       Nntpconn_errmsg_reset(client->conn,0);
 continue;
     }
     goto ex;
   }
   ret= Nntpclient_group_stats(client,reportfile,0);
   if(ret<=0)
     goto ex;
   if(Nntpstats_report(client->stats,2)) /* over limits */
     Nntpstats_sum(sum_stats,client->stats,0);
   Nntpstats_largest_group(largest_stats,client->stats,0);
   Nntpstats_largest_article(lart_stats,client->stats,0);
 }
 Nntpstats_print(client->stats,
                 "\n===========================================================================\n\n",0);
#ifdef NIX
 Nntpstats_print(lart_stats,"\n===== Group with largest article =====\n");
 Nntpstats_report(lart_stats,1);
 if(ret<=0)
   goto ex;
#endif
 Nntpstats_print(lart_stats,"\n===== Largest group =====\n",0);
 Nntpstats_report(largest_stats,1);
 if(ret<=0)
   goto ex;
 Nntpstats_print(lart_stats,
                 "\n===== Total sizes and largest article =====\n",0);
 Nntpstats_report(sum_stats,1);
 if(ret<=0)
   goto ex;
 Nntpstats_print(client->stats,
                 "\n===========================================================================\n",0);
 ret= 1;
ex:;
 Nntpstats_destroy(&sum_stats,0);
 Nntpstats_destroy(&largest_stats,0);
 Nntpstats_destroy(&lart_stats,0);
 Sregex_destroy(&rexpr,0);
 if(fp!=NULL) fclose(fp);
 return(ret);
}


int Nntpclient_create_rc(client,flag)
struct NntpclienT *client;
int flag;
/*
 bit0= permission to overwrite existing file
*/
{
 char filename[SfileadrL],text[Nntpclient_strmaX];
 char dir[SfileadrL];
 int ret;

 Nntpfiler_std_dir(client->filer,dir,0);

 sprintf(filename,"%s/.%src",dir,client->progname);
 text[0]= 0;
 sprintf(text+strlen(text),"\n");
 sprintf(text+strlen(text),"# Startupfile for %s\n",client->progname);
 sprintf(text+strlen(text),"# sets base directory, server and group\n");
 sprintf(text+strlen(text),
    "# although automatically created, this file may be altered manually\n");
 sprintf(text+strlen(text),"\n");
 sprintf(text+strlen(text),"-basedir  %s\n",client->basedir);
 sprintf(text+strlen(text),"-server   %s\n",client->servername);
 sprintf(text+strlen(text),"-group    %s\n",client->group->name);
 ret= Sfile_string_to_file(filename,text,flag&1);
 return(ret>=0);
}


int Nntpclient_hull(client,in_client,servername,portname,groupname,flag)
/* A virgin client which inherits a few properties from the input client */ 
struct NntpclienT **client;
struct NntpclienT *in_client;
char *servername,*portname,*groupname;
int flag;
{
 struct NntpclienT *c;
 int ret,port= -1;
 char *sname,*pname,*gname,in_portname[16];

 if(servername[0]!=0 && !(servername[0]=='-' && servername[1]==0))
   sname= servername;
 else
   sname= in_client->servername;
 sprintf(in_portname,"%ld",in_client->port);
 if(portname[0]!=0 && !(portname[0]=='-' && portname[1]==0))
   pname= portname;
 else
   pname= in_portname;
 if(groupname[0]!=0 && !(groupname[0]=='-' && groupname[1]==0)) {
   gname= groupname;
 } else {
   if(in_client->group!=NULL)
     gname= in_client->group->name;
   else
     gname= groupname;
 }
 ret= Nntpclient_new(client,sname,1);
 if(ret<=0)
   goto ex;
 c= *client;
 sscanf(pname,"%ld",&port);
 if(port<=0)
   port= 119;
 c->port= port;
 Nntpfiler_cpy(c->filer,in_client->filer,1); /* do not copy conn */
 strcpy(c->basedir,in_client->basedir);
 strcpy(c->progname,in_client->progname);
 strcpy(c->start_dir,in_client->start_dir);
 /* A virgin group */
 ret= Nntpgroup_new(&(c->group),c->conn,gname,1);
 if(ret<=0)
   goto ex;
 Nntpconn_set_groupname(c->conn,c->group->name,0);
 Nntpclient_std_download_dir(c,0);
 Nntpgroup_set_dir(c->group,c->download_dir,0);
 ret= Nntpload_new(&(c->loadjob),c->conn,c->head,c->download_dir,"msg",1);
 if(ret<=0)
   goto ex;
 return(1);
ex:;
 Nntpclient_destroy(client,0);
 return(0);
}


int Nntpclient_adjust_links(in_client,servername,portname,groupname,flag)
struct NntpclienT *in_client;
char *servername,*portname,*groupname;
int flag;
/*
 bit0= do not aquire maintenance lock
*/
{
 struct NntpclienT *client;
 int ret;
 
 ret= Nntpclient_hull(&client,in_client,servername,portname,groupname,0);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_adjust_links(client->filer,0);
 if(ret<=0)
   goto ex;
ex:;
 if(client!=NULL) {
   in_client->conn->errn= client->conn->errn;
   strcpy(in_client->conn->errmsg,client->conn->errmsg);
   Nntpclient_destroy(&client,0);
 }
 return(ret);
}


int Nntpclient_remove_old_files(in_client,servername,portname,groupname,
                                timecode,flag)
struct NntpclienT *in_client;
char *servername,*portname,*groupname;
char *timecode;
int flag;
/*
 bit0-7= mode : 0=group files , 1=message id list
*/
{
 struct NntpclienT *client;

 int ret,valid;
 time_t agelimit;

 ret= Nntpclient_hull(&client,in_client,servername,portname,groupname,0);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_remove_old_files(client->filer,timecode,flag);
ex:;
 if(client!=NULL) {
   in_client->conn->errn= client->conn->errn;
   strcpy(in_client->conn->errmsg,client->conn->errmsg);
   Nntpclient_destroy(&client,0);
 }
 return(ret);

}


int Nntpclient_main_index(in_client,flag)
struct NntpclienT *in_client;
int flag;
{
 struct NntpclienT *client;
 int ret;

 ret= Nntpclient_hull(&client,in_client,"no_server","119","no_group",0);
 if(ret<=0)
   goto ex;
 ret= Nntpfiler_main_index(client->filer,client->progname,0);
ex:;
 if(client!=NULL) {
   in_client->conn->errn= client->conn->errn;
   strcpy(in_client->conn->errmsg,client->conn->errmsg);
   Nntpclient_destroy(&client,0);
 }
 return(ret);
}


int Nntpclient_subscribe(in_client,servername,portname,groupname,flag)
struct NntpclienT *in_client;
char *servername,*portname,*groupname;
int flag;
{
 struct NntpclienT *client;
 int ret,k;

 ret= Nntpclient_hull(&client,in_client,servername,portname,groupname,0);
 if(ret<=0)
   goto ex;

 ret= Nntpfiler_create_std_dir(client->filer,client->progname,0);
 if(ret<=0)
   goto ex;
 if(ret==3 || ret==4) /* no server or no group given */
   {ret= 1; goto main_index;}
 if(ret!=2) { /* directory was not already there */
   for(k=0;k<2;k++){
     /* init list_newest */ /* init list_start */ 
     ret= Nntpclient_write_msgmark(client,0,k);
     if(ret<=0)
       goto ex;
   }
 }
 ret= Nntpclient_create_rc(client,0);
 if(ret<0)
   return(0);
main_index:;
 ret= Nntpclient_main_index(client,0);
 if(ret<0)
   return(0);
 ret= 1;
ex:;
 if(client!=NULL) {
   in_client->conn->errn= client->conn->errn;
   strcpy(in_client->conn->errmsg,client->conn->errmsg);
   Nntpclient_destroy(&client,0);
 }
 return(ret);
}


int Nntpclient_post_interpreter(client,argc,argv,ard,flag)
struct NntpclienT *client;
int argc;
char **argv;
int *ard;
int flag;
{
 int ret,i,post_end,style,first_msg,msgid_flag;
 struct NntpposT *postjob;
 char *subject,*body,line[Nntppost_stringleN],*cpt,msgid[Nntppost_stringleN];
 char binfile[SfileadrL];

 if(client->group==NULL){
   client->conn->errn= -26;
   sprintf(client->conn->errmsg,
           "Posting required but no group opened");
   return(0);
 }
 if((*ard)+2>=argc)
   return(0);
 for(i=(*ard);i<argc;i++)
   if(strcmp(argv[i],"-post_end")==0)
 break;
 post_end= i;
 
 subject= argv[(*ard)+1];
 body= argv[(*ard)+2];
 (*ard)+= 2;

 msgid_flag= (!!client->use_own_msgid)<<1;
 if(Sregex_complines("file:",body,1)) {
   ret= Nntppost_new(&postjob,client->group,client->hostid,client->username,
                     client->organization,subject,body+5,"",1|msgid_flag);
 }else{
   ret= Nntppost_new(&postjob,client->group,client->hostid,client->username,
                     client->organization,subject,body,"",0|msgid_flag);
 }
 if(ret<=0)
   return(ret);
 Nntppost_set_reference(postjob,client->reference,0);
 if((*ard)+1>=post_end){ /* without binary file */
   sprintf(line,"\nposting message to %s\n",postjob->group->name);
   Nntpclient_result(client,line,0);
   ret= Nntppost_article(postjob,0);
   if(client->conn->show_io_stat)
     Nntpconn_show_stat(client->conn,8|16);
   if(ret<=0)
     goto ex;
   sprintf(line,"+++ message uploaded to %s\n",postjob->group->name);
   Nntpclient_result(client,line,0);
   ret= 1;
   goto ex;
 }

 /* with binary files */
 style= client->posting_style&15;
 first_msg= 2;
 for((*ard)++;(*ard)<post_end;(*ard)++){
   if(first_msg>0) first_msg--;
   if(Nntpclient_intrptchk(client,0)){
     client->conn->errn= -14;
     sprintf(client->conn->errmsg,"-post stopped because : %s",
                                  client->user_messages);
     ret= 0;
     goto ex;
   }
   cpt= strrchr(argv[(*ard)],'/');
   if(cpt==NULL)
     cpt= argv[(*ard)];
   else
     cpt++;
   if(argv[(*ard)][0]=='/')
     strcpy(binfile,argv[(*ard)]);
   else
     sprintf(binfile,"%s/%s",client->start_dir,argv[(*ard)]);
   Nntppost_set_bin_file(postjob,binfile,0);
   sprintf(line,"%s - %s (%ldK)",subject,cpt,Sfile_size(binfile,0)/1024);
   Nntppost_set_subject(postjob,line,0);
   sprintf(line,"\nposting message to %s\n",postjob->group->name);
   Nntpclient_result(client,line,0);
   ret= Nntppost_article(postjob,!first_msg);
   if(client->conn->show_io_stat)
     Nntpconn_show_stat(client->conn,8|16);
   if(ret<=0)
     goto ex;
   sprintf(line,"+++ message uploaded to %s\n",postjob->group->name);
   Nntpclient_result(client,line,0);
   if((*ard)<post_end-1 && postjob->msgid[0]!=0)
    if((first_msg && style==1) || style==2) {
     Nntppost_set_reference(postjob,postjob->msgid,0);
     ret= Nntppost_set_all_references(postjob,postjob->msgid,1);
     if(ret<=0)
       goto ex; 
   }
 } 
 if((*ard)>=argc) (*ard)= argc-1;
 else if((*ard)<argc-1) (*ard)++;         /* consume -post_end */

 ret= 1;
ex:;
 Nntppost_destroy(&postjob,0);
 return(ret);
}


int Nntpclient_control(client,control_msg,flag)
struct NntpclienT *client;
char *control_msg;
int flag;
{
 int ret;
 struct NntpposT *postjob;

 if(client->group==NULL){
   client->conn->errn= -26;
   sprintf(client->conn->errmsg,"Posting required but no group opened");
   return(0);
 }
 if(control_msg[0]==0) {
   client->conn->errn= -30;
   sprintf(client->conn->errmsg,"Attempt to post an empty control message");
   return(0);
 }
 
 ret= Nntppost_new(&postjob,client->group,client->hostid,client->username,
                   client->organization,control_msg,control_msg,"",0);
 if(ret<=0)
   return(ret);
 Nntppost_set_control_msg(postjob,control_msg,0);
 ret= Nntppost_article(postjob,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Nntppost_destroy(&postjob,0);
 return(ret);
}


int Nntpclient_cancel(client,msgid,flag)
struct NntpclienT *client;
char *msgid;
int flag;
{
 int ret;
 char subject[Nntppost_stringleN];
#ifdef NIX
 struct NntpposT *postjob;

 if(client->group==NULL){
   client->conn->errn= -26;
   sprintf(client->conn->errmsg,
           "Posting required but no group opened");
   return(0);
 }
 if(msgid[0]==0)
   return(0);
 
 sprintf(subject,"cancel %s",msgid);
 ret= Nntppost_new(&postjob,client->group,client->hostid,client->username,
                   client->organization,subject,"cancel","",0);
 if(ret<=0)
   return(ret);
 Nntppost_set_control_msg(postjob,subject,0);
 ret= Nntppost_article(postjob,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Nntppost_destroy(&postjob,0);
 return(ret);

#else /* NIX */

 if(msgid[0]==0) {
   client->conn->errn= -30;
   sprintf(client->conn->errmsg,"Attempt to cancel a message with empty ID");
   return(0);
 }
 sprintf(subject,"cancel %s",msgid);
 ret= Nntpclient_control(client,subject,0);
 return(ret);

#endif /* ! NIX */

}


int Nntpclient_filter_parser(client,argc,argv,ard,flag)
struct NntpclienT *client;
int argc;
char **argv;
int *ard;
int flag;
{
 struct NntpfilteR *filter;
 int ret;

 filter= client->filter;
 
 for(;(*ard)<argc;(*ard)++) {

   if(strcmp(argv[(*ard)],"-filter")==0){

     if(filter!=NULL)
       Nntpfilter_destroy(&(filter),0);
     ret= Nntpfilter_new(&(filter),0,0x7FFFFFFF,2);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-not")==0) {

     ret= Nntpfilter_not(filter,0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-and")==0){

     ret= Nntpfilter_and(filter,0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-or")==0){

     ret= Nntpfilter_or(filter,0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],")")==0 || strcmp(argv[(*ard)],"-subend")==0){

     ret= Nntpfilter_close_bracket(filter,0);
     if(ret<=0) {
       return(ret);
     }

   } else if(strcmp(argv[(*ard)],"(")==0 || strcmp(argv[(*ard)],"-sub")==0){

     ret= Nntpfilter_open_bracket(filter,0);
     if(ret<=0)
       return(ret);

   } else if((*ard)+1<argc){ /* a test : pair of property and expression */

     if(argv[(*ard)][0]==0)
       goto no_filter_option;
     if(argv[(*ard)][strlen(argv[(*ard)])-1]!=':')
       goto no_filter_option;

     if(Nntpfilter_cursor_complete(filter,0)) {
        filter->errn= -2;
        sprintf(filter->errmsg,"Binary operator expected, expression found");
        return(0);
     }
     (*ard)+= 1;
     ret= Nntpfilter_set_regex(filter,argv[*ard-1],argv[*ard],0);
     if(ret<=0){
       client->conn->errn= -15;
       sprintf(client->conn->errmsg,"Cannot set filter expression : %s %s",
                                    argv[(*ard)-1],argv[(*ard)]);
       return(ret);
     }

   } else { 

no_filter_option:;
     (*ard)--;
 break; 

   } 

 }
 return(1);
}


int Nntpclient_read_from(client,filename,flag)
struct NntpclienT *client;
char *filename;
int flag;
/*
 bit0= do not complain verbously about failure to open the file
*/
{
 char **argv= NULL,*buf= NULL;
 int argc,i,ret,l,l_lim,pass;
 char line[4096];
 FILE *fp;

 fp= Sfile_fopen(filename,"r",flag&1);
 if(fp==NULL){
   client->conn->errn= errno;
   sprintf(client->conn->errmsg,"Error %ld on opening argument file: %s\n%s",
                                errno,filename,strerror(errno));
   return(0);
 }
 for(pass=0;pass<2;pass++){
   l= 0;
   fseek(fp,0,0);
   while(Sfile_fgets(line,sizeof(line),fp)!=NULL){
     Sregex_trimline(line,0);
     if(line[0]==0 || line[0]=='!' || line[0]=='#')
   continue;
     if(pass>0) {
       if(l+strlen(line)+1>l_lim) {
         client->conn->errn= -33;
         sprintf(client->conn->errmsg,"Filesize changed during -read_from %s",
                                      filename);
         ret= 0;
         goto ex;
       }
       strcpy(buf+l," ");
       strcpy(buf+l+1,line);
     }
     l+= strlen(line)+1;
   }
   if(pass==0){
     buf= TSOB_FELD(char,l+1);
     if(buf==NULL) {
       client->conn->errn= -17;
       sprintf(client->conn->errmsg,"Not enough memory for -read_from %s",
                                    filename);
       ret= -1;
       goto ex;
     }
     buf[0]= 0;
     l_lim= l;
   }
 }

 ret= Sfile_make_argv(client->progname,buf,&argc,&argv,0);
 if(ret<=0)
    goto ex;
 i= 1;
 while(i<argc){
   ret= Nntpclient_outer_fence(client,argc,argv,&i,0);
   if(ret<=0)
     {ret= -1; goto ex;}
   while(++i<argc)
     if(argv[i][0]=='-')
       break;
 }
 ret= 1;
ex:;
 Sfile_fclose(fp);
 Sfile_make_argv("","",&argc,&argv,2); /* release argv */
 if(buf!=NULL)
   free((char *) buf);
 return(ret);
}


int Nntpclient_help(client,flag)
struct NntpclienT *client;
int flag;
{

#define MELDEN(x) Nntpclient_result(client,x,1)

static char options_text[][84]={
"",
"                       Features and Options of nntpclient",
"",
"This software connects to news servers via NNTP and fetches messages which",
"it stores to the filesystem. Image files in formats uuencode and",
"MIME-multipart are extracted, decoded, and stored separately.",
"",
"The program can also upload simple messages with at most one binary file",
"attached to each of them. Nevertheless more than one file can be given",
"to produce an according number of messages. The name of each binary file",
"is attached to the subject line of the message.",
"",
"The directory structure for storing downloaded data is given by a base",
"directory, the servername, the portnumber, and the group name:",
" $HOME/snntpbatch_download/${servername}.port${portnumber}/${groupname}",
"The base directory (default: $HOME/snntpbatch_download) can be set with",
"the option -basedir .",
"This directory structure can be created by option -subscribe if the",
"base directory already exists.",
"",
"Message texts are stored as \"msg${messagenumber}.html\" in their group's",
"directory. ${messagenumber} is padded with zeros to 8 digits.",
"Binary files (images) are stored in a subdirectory called \"bin\".",
"",
"The image filenames are read from the message content. If they collide",
"with a filename which is already in the target directory, then they are",
"extended by \"_\" and a 7 digit number. A suffix after the last dot in",
"the filename is preserved.",
"",
"Messages can be selected directly via their message id or by filtering",
"with regular expressions (see man regex, man grep, man ex ...)",
"If a message filename is already in the target directory, then the message",
"including all images gets omitted if not \"-overwrite on\" is set.",
"",
"Content summaries (generated by -list_interval) are stored in the group's",
"directory as files with name  list_${first_number}_${last_number}.html.",
"The file \"list_start\" contains the first number. The file \"list_newest\" ",
"contains the number of the last message header actually read. In case of",
"connection problems this number can be smaller than original ${last_number}.",
"The list file is renamed accordingly then. A copy of the most recently",
"successfully created list file is available as  list_00000000_00000000.html .",
"",
"Some simple HTML features are inserted into lists and messages.",
"With your browser bookmark the file  main_index.html  in the basedirectory.",
"This index of servers and groups is automatically maintained by -subscribe.",
"",
"Each link of the main index leads to a group index  list_index.html  which",
"lists links to content summaries. Since the links lead to single message",
"files, such a summary is called message index.",
"",
"A message index gets a TITLE which tells about server, group and desired",
"message range. A header part provides links upward to the main index and",
"the group index. Another header link points to the first message of the",
"list which has actually been downloaded.",
"Also there is a hyperlink (HREF) and a jumplabel (NAME) to the filename",
"of each message listed. Usually the message files get downloaded later",
"so these links do not work immediately after creating the message index.",
"",
"Messages get a TITLE which tells their filename. Three hyperlinks at start",
"and end of the message text serve for navigation.",
"\"up\" points to list_00000000_00000000.html#{messagefilename}.",
"The messages which are fetched by a single -fetch_interval form a chain:",
"\"previous\" points to the  message fetched before.",
"\"next\" points to the message which was fetched afterwards. This link gets",
"its final content only after downloading of this following message started.",
"Messages downloaded by -fetch simply point to the neighbouring message",
"numbers wether they exist or not."
"Naturally, images are embedded by IMG tags.",
"",
"If a message contains references to other messages, the last one is checked",
"wether it points to a downloaded message in the same group. If this is true,",
"then a link to that message is inserted in the header of the one which",
"is freshly downloaded. Also a link back is inserted into the referenced",
"message.",
"",
"The final link structure is generated by option -adjust_links after the",
"desired messages have been loaded. This operation also generates the group",
"index (all available message indice in the group). It should be performed as",
"often as bunch of new messages has been downloaded from a group.",
"",
"This software is copyright 2001, Thomas Schmitt stic-source@gmx.net and",
"provided to you without any warranty under an open source BSD license.",
"(see file COPYING)",
"",
"",
" ---------------------------------------------------------------------------",
"                              Getting Started",
" ---------------------------------------------------------------------------",
"",
"                      --- Setup of Servers and Groups ---",
"",
"Create the base directory, the directory for your server and the",
"directories for holding the message id list :",
"",
"    $ basedir=$HOME/snntpbatch_download",
"    $ newsserver=INSERT_NEWSERVER_ADDRESS_HERE",
""
"    $ mkdir $basedir",
"    $ cd $basedir",
"    $ snntpbatch -subscribe $newsserver 119 -",
"",
"You may want to start your browser and bookmark the main index page. It is",
"still quite empty yet :",
"",
"    $ netscape $HOME/snntpbatch_download/main_index.html &",
"---------------------------------------------------------------------------",
"no groups subscribed",
"---------------------------------------------------------------------------",
"",
"If your news service requires a user-id and a password, you should edit the",
"main startup-file and insert a line  -login <username> <password>",
"",
"    $ touch $HOME/.snntpbatchrc",
"    $ chmod og-rw $HOME/.snntpbatchrc",
"    $ vi $HOME/.snntpbatchrc",
"",
"a typical login line would look like",
"",
"-login JOHNDOE 8HM1RX",
"",
"If you use more than one news server then you may put this line in a local",
"startup file (see below). But for a first try, $HOME/.snntpbatchrc is ok.",
"",
"The following action may last some time and you may skip that until you",
"really need it.",
"A list of available newsgroups can be obtained by this command :",
"    $ snntpbatch -server $newsserver -list_groups",
"Read the result file in the server's directory",
"    $ view ${newsserver}.port119/grouplist",
"presented in raw format as sent from the server. Four fields per line :",
"        groupname   last_msg_number   first_msg_number   posting_yes_or_no",
"",
"Create the directories and files needed for the first news group :",
"",
"    $ newsgroup=alt.binaries.pictures.fantasy-sci-fi",
"    $ snntpbatch -subscribe $newsserver 119 $newsgroup",
"",
"If you reload the bookmarked main index, you will see that group listed now.",
"---------------------------------------------------------------------------",
"                Serverdirectory YOUR_NEWSSERVER.port119",
"---------------------------------------------------------------------------",
"Group alt.binaries.pictures.fantasy-sci-fi",
"first message |group index |most recent index (start message) |last message",
"---------------------------------------------------------------------------",
"The links are still dead until a first -list_interval and -fetch_interval",
"has been performed.",
"",
"                            --- Daily usage ---",
"",
"Create a message index. It will contain some header information about those",
"messages which have been added since the last -list_interval in $newsgroup",
"on server $newsserver :",
"",
"    $ snntpbatch -server $newsserver \\",
"                 -group  $newsgroup \\",
"                 -list_interval new last",
"",
"This creates a HTML file list_????????_????????.html in the group directory",
"which can be accessed also as copy  list_00000000_00000000.html .",
"On the main index page, the link \"most recent index\" of the group should",
"be alive now. Also \"group index\".",
"",
"Fetch all recently listed articles with subject about \"Royo\"",
"but not about \"Repost\" . After downloading, the HTML hyperlinks in",
"the messages and index files are adjusted :",
"",
"    $ snntpbatch -server $newsserver \\",
"                 -group  $newsgroup \\",
"                 -filter subject: '[Rr][Oo][Yy][Oo]' \\",
"                         -and -not subject: '[Rr][Ee][Pp][Oo][Ss][Tt]' \\",
"                 -fetch_interval list_start list_end \\",
"                 -adjust_links - - -",
"",
"If there is no '-filter' then all messages in the given range get fetched.",
"As soon as the first message has been downloaded, it is accessible via",
"the link \"(start message)\" on the main index page. The \"next\" link",
"within a message page gets adjusted when the download of the next message",
"begins. So wait until that happens before you enter a message page. Or",
"reload the page if the \"next\" link is still dead.",
"The links \"first message\" and \"last message\" will get alive only after",
"the -adjust_links command.",
"",
"If there are many Royo images or if you do not like them, you may cancel",
"the run by creating an empty file  nntpclient_stop  in the group directory",
"    $ touch $basedir/${newsserver}.port119/$newsgroup/nntpclient_stop",
"this stops the whole program run after the current article is downloaded.",
"A less rigid way is to set the control index to a ridiculous high value.",
"    $ echo 2000000000 \\",
"           > $basedir/${newsserver}.port119/$newsgroup/nntpclient_index",
"this will only end the -fetch_interval command and not the following ones.",
"You may resume the download with just the same snntpbatch command again.",
"",
"Usually the indexing, fetching and relinking is done within shell scripts.",
"So updating all groups can be a matter of single command without any",
"arguments.",
"",
"If you see interesting messages in the list which are not fetched by",
"your automats, then you can get them by simple calls like these two.",
"",
"Fetch the single message 110233 :",
"",
"    $ snntpbatch -server $newsserver \\",
"                 -group  $newsgroup \\",
"                 -fetch  110233",
"",
"Fetch messages 110200 to 110399 . Get only those with subject about",
"\"Space\" or \"space\" :",
"",
"    $ snntpbatch -server $newsserver \\",
"                 -group  $newsgroup \\",
"                 -filter subject: '[Ss]pace' \\",
"                 -fetch_interval 110200 110399",
"",
"A filter to prevent double downloading of files, if the names can be guessed",
"from the subject line. It accepts subjects with '[Ss]paceship' or",
"'[Qq][Mm][Aa][Nn]'.",
"If the subject texts of the 'Qman's contain a filename with more than 4",
"characters before the first dot, this name is checked. The message is only",
"accepted if the filename (case independend) doesn't already exists in the",
"hashed directory $filerecord . After the file is accepted, an empty",
"file with this name is created within the directory. The hashing structure",
"is created automatically, but at least the parent directory of $filerecord",
"has already to exist.",
"For details, see the the explanation of option -filter below and keep in",
"mind, that a test after -and is only performed if the test before this",
"-and was true. So only 'Qman's which are not already known get recorded",
"by \"subjectfilerec:\".",
"",
"                 -filter \\",
"                 subject: '[Ss]paceship' \\",
"                 -or -sub \\",
"                    subject: '[Qq][Mm][Aa][Nn]' \\",
"                    -and -not -sub \\",
"                       subjectfile: '......*\\..*' \\",
"                       -and subjectfilefound: \"$filerecord\" \\",
"                    -subend \\",
"                    -and subjectfilerec: \"$filerecord -return_true\" \\",
"                 -subend \\",
"",
"Complicated filters should be written into a separate file and employed",
"by -read_from . In that case one does not need \\ at line ends and has not to",
"be aware of most characters which are special to the shell. Quotation marks",
"'' and \"\" are interpreted like with the shell. No variable substition is",
"performed, so you have to insert $filerecord as a constant text :",
"",
"                 -filter",
"                 subject: '[Ss]paceship'",
"                 -or (",
"                    subject: '[Qq][Mm][Aa][Nn]'",
"                    -and -not (",
"                       subjectfile: '......*\\..*'",
"                       -and subjectfilefound: \"INSERT_$FILERECORD\"",
"                    )",
"                    -and subjectfilerec: \"INSERT_$FILERECORD -return_true\"",
"                 )",
"",
"If you expect a lot of downloaded data and your $HOME filesystem is small",
"you may consider to copy $HOME/snntpbatch_download to a larger filesystem.",
"Either you may set the new basedirectory by",
"  -basedir {new_name_of_snntpbatch_download}",
"at the very start of $HOME/.snntpbatchrc or you may establish a link from",
"$HOME/snntpbatch_download to the new location. One may also move some",
"subtrees of the base directory to other filesystems if needed. Here, links",
"are the only means to keep it working.",
"A happy snntbatch installion can easily download hundreds of megabytes per",
"day. So keep an eye on your disk usage.",
"",
"",
"                            --- Disk Cleaning ---",
"",
"Remove from the newsgroup directory all messages and binary files which are",
"older than 14 days. Remove message index files which only contain expired",
"messages. Also this option adjusts the links of messages and index lists.",
"",
"   $ snntpbatch -remove_old_files $newsserver 119 $newsgroup -14d",
"",
"Remove message ids from the directories below  msg_idlist  which are older",
"than 30 days. These files form a database which prevents multiple",
"downloading of identical messages. Also they contain some header information",
"",
"   $ snntpbatch -remove_old_ids  -30d",
"",
"When creating new directories, -subscribe imposes a limit of at least 3 days",
"for groups and message ids (via file  sfile_remove_old_files_allowed ).",
"",
"",
"                              --- Posting ---",
"",
"For posting messages, a server and a group have to be set. Usually some",
"id informations are set before posting is performed.",
"",
"Post three images from stic's sample collection to a test newsgroup :",
"",
"    $ cd $(cat $HOME/.stic_main_dir)",
"    $ newsgroup=alt.test.binary.posting",
"    $ snntpbatch -server  $newsserver \\",
"                 -group   $newsgroup \\",
"                 -post_id my_fantasy_host my_fantasy_name my_organization \\",
"                 -post    \"please ignore - just a test\" \"some text\" \\",
"                          images/birds/*",
"",
"This will result in three messages with subject lines :",
"    please ignore - just a test - 00000010.jpg (10K)",
"    please ignore - just a test - 00000055.jpg (15K)",
"    please ignore - just a test - 00000058.jpg (17K)",
"and the body texts will just consist of the words \"some text\".",
"",
"After each posted message the program waits for confirmation by the server.",
"    message sent. waiting for confirmation ...",
"    ... accepted.",
"In case of failure, you will get a reaction like this",
"    message sent. waiting for confirmation ...",
"    ... rejected.",
"    status = 441 Invalid domain in from-header", 
"this server does not like my_fantasy_host, give it a more realistic one.",
"",
"Since some servers do send error message without CR LF at the end, the",
"program might get stuck in this phase. The received part of the error",
"message should be visible at the terminal's bottom line then. Sorry to say,",
"there is no remedy against this effect yet.",
"",
"You will have to -subscribe $newsserver 119 $newsgroup to see the result.",
"It may last some time until the messages appear on your news server.",
"",
"Post a reply to the message with id <38B1C855.3D17F85A@home> :",
"",
"    $ snntpbatch -server   $newsserver \\",
"                 -group    $newsgroup \\",
"                 -post_id  my_fantasy_host my_fantasy_name my_organization \\",
"                 -post_ref '<38B1C855.3D17F85A@home>' \\",
"                 -post     \"please ignore - just a test\" \\",
"                           \"testing follow-ups\" \\",
"                           images/misc/00000059.jpg",
"",
"It is also ok to give the message id without brackets 38B1C855.3D17F85A@home",
"",
"Cancel the message with id <09400808BE5838D17FCF8088@my_fantasy_host> :",
"",
"    $ snntpbatch -server   $newsserver \\",
"                 -group    $newsgroup \\",
"                 -post_id  my_fantasy_host my_fantasy_name my_organization \\",
"                 -cancel   '<09400808BE5838D17FCF8088@my_fantasy_host>'",
"",
"Cancelling does not work with all news servers.",
"Especially if -post_own_msgid is set to \"on\".",
"",
"",
" ---------------------------------------------------------------------------",
"                                 Options",
" ---------------------------------------------------------------------------",
"",
"If the first option is # or ! then the whole list of options is ignored.",
"(Useful for remarks in jobfiles)",
"Options are processed strictly sequentially. Presets like -server or",
"-filter only apply to the options that follow them. When in doubt, use the",
"sequence of options presented here. (i.e -timeout before -server)",
"",
"Intervals of messages are given by the lowest and highest message number.",
"The special codes \"first\" (or \"start\") and \"last\" (or \"end\") are",
"replaced by the first and last message number as obtained when opening the",
"group.",
"The code \"list_start\" and \"list_end\" give the first and last number of",
"the most recent -list_interval run in the group. The code \"new\" gives",
"the first unlisted message number (i.e. list_end+1).",
"",
"Some options require timedefinitions.",
"Absolute timepoints are defined by the input format MMDDhhmm[[CC]YY][.ss]]",
"of the UNIX command 'date'. For example:",
"   021619302000  =  16 Feb 2000 19:30:00 local time",
"Relative times are preceded by '-' or '+'. They are added to the current",
"time. '-' therefore points to the past, '+' points to the future. The basic",
"unit is 1 second. It can be modified by a letter at the end of the number:",
"   seconds 1s=1 , hours 1h=3600s , days 1d=24h ,",
"   weeks 1w=7d , months 1m=31d , years y : Xy=X*365.25d+1d",
"Example:",
"  -3w  =  3 weeks in the past  =  -1814400s ",
"Please note that 'm' and 'y' are possibly slightly larger than the actual",
"calendar time range.",
"",
" -basedir directoryname   set the basedir for downloading messages",
"                      (default: $HOME/snntpbatch_download)",
"",
" -login username password  set the login parameters for news servers which",
"                      require authentication. This option should be stored",
"                      in the file .snntpbatchrc in the server directory",
"                      (e.g. $basedir/news.isp.port119/.snntpbatchrc ).",
"                      This file should have read permission only for its",
"                      owner. If you do not trust your superuser, store the",
"                      file on a floppy and make .snntpbatchrc a symbolic link",
"                      to the floppyfile's name. A missing link target has",
"                      the same effect as a missing .snntpbatchrc , i.e. none.",
"",
" -subscribe servername portnumber groupname",
"                      create the directories necessary as targets for down-",
"                      loading. If portnumber is 0 or negative, 119 is used.",
"                      If the group directory did not already exist, then",
"                      also create some controlfiles with initial values.",
"                      If the group name is '-' , then no group directories",
"                      are created. If the servername is '-' , then no server",
"                      directory is created (just the basedir and the",
"                      msg_idlist directories). Directories are created only",
"                      if they did not already exist. No files are removed.", 
"",
" -timeout seconds     set the number of seconds to wait for reply",
"                      After this limit is exceeded, the program closes the",
"                      connection to the server and tries to re-establish it.",
"                      Timeout numbers should be larger than the time needed",
"                      to fetch a large article since an interrupted article",
"                      has to be re-read from start. Default is 120",
" -timeout_bell on|off",
"                      enable|disable the acustic signal on timout.",
"                      Default is \"on\"",
" -baudlimit baudrate  restricts the average read bandwith to the given rate.",
"                      This is done by waiting after too many bytes have been",
"                      read. Since a few kB are buffered, the bandwidth at",
"                      the communications port may vary.",
"",
" -port portnumber     use the given port at the server. If a server aready",
"                      has been specified, then the connection is closed and",
"                      re-established.",
"                      Usually this option is not necessary since port 119",
"                      is the default and should work fine.",
" -server servername   connect to the given server at the preset port.",
"                      This should be a very early option since nothing can",
"                      be fetched and no group can be selected without -server",
"",
" -list_groups         write a list of groups into the file grouplist in the",
"                      server+port's directory",
" -group groupname     choose the group on the connected server. This should",
"                      be an early option since nothing can be fetched ",
"                      without -group",
" -filter   -not | -or | -and | '(' | -sub | ')' | -subend |",
"           from: | date: | subject: | subjectfile: regular_expression", 
"           subjectfilefound: | subjectfilerec:  option_list", 
"                      set a filter for -fetch_interval  or  -list_interval .",
"                      The default filter does not reject anything.",
"                      Basically a filter consists of tests which use regular",
"                      expressions (like grep and sed). An expression can",
"                      test one of the message header properties",
"                         \"from:\" \"date:\" \"subject:\"",
"                      so a test actually consists of two arguments.",
"                      Special tests with pseudo properties are:",
"                         \"subjectfile:\"",
"                               try to read a filename from \"subject:\"",
"                               and test wether it matches the expression",
"                         \"subjectfilefound:\"",
"                               try to read a filename from \"subject:\"",
"                               option_list is a list of directories",
"                               separated by blanks. Nevertheless it has",
"                               to be a single shell argument. Use quotation",
"                               marks to protect the blanks. The filename is",
"                               searched in these directories. The test is",
"                               not case sensitive unless \"-case\" is",
"                               given at start of the directory list.",
"                               This test can handle hashed directories.",
"                         \"subjectfilerec:\"",
"                               guess a filename from the subject line and",
"                               create an empty file in each of the hashed",
"                               directories listed in option_list. This is",
"                               done at the end of downloading if a file",
"                               with the guessed name was actually extracted.",
"                               Case sensitivity is tiggered with -case",
"                               and -nocase. The result of the test is",
"                               'true' if a subject filename could be guessed",
"                               else it is 'false'. It also can be forced by",
"                               -return_true resp. -return_false within the",
"                               list of directories.",
"                               If a directory does not exist or is not",
"                               hashed, then it is made a hashed directory.",
"                               Usually the hashed directories are tested",
"                               by \"subjectfilefound:\" to avoid duplicates.",
"                      Tests can be combined by the logical operations -or,",
"                      -and, -not and can be grouped by brackets or pairs of",
"                      -sub and -subend. Precedence: -not, -and, -or .", 
"                      Regular expressions, option lists and brackets should",
"                      always be surrounded by quotes to prevent the shell",
"                      parser from interpreting them. With -read_from quotes",
"                      are only needed to include blank characters.",
"                      The tests are performed in strict left-to-right order.",
"                      An important fact concerning speed as well as the",
"                      effects of \"subjectfilerec:\" :",
"                        The right side of -and is only performed if the left",
"                        side is true. The right side of -or is only",
"                        performed if the left side is false.",
" -overwrite on|off    enable|disable overwriting of already downloaded",
"                      messages. Image files don't get overwritten but the",
"                      newly loaded images get new names. Default is \"off\"",
" -html  on|off        enable|disable insertion of HTML tags in lists and",
"                      messages. Default is \"on\"",
" -group_stat_limits min_articles min_total_size min_max_article_size",
"                      sets limits for the options -group_statistics and",
"                      -grouplist_statistics . A group's overview result gets",
"                      only reported if the group contains at least as many",
"                      articles as given by parameter min_articles.",
"                      The total sum of article sizes has to be at least",
"                      min_total_size. The size of the largest article has",
"                      to be at least min_max_article_size.",  
" -group_statistics target_file",
"                      print a summarized overview of the current group to",
"                      standard output. If target_file is not '-' or '.' then",
"                      the output is also appended to this file.",
"                      The result has the following form",
"                        Group : ${groupname}",
"                        Age   : ${oldest}  to  ${youngest}  days",
"                        Size  : ${x}  bytes in ${y} articles with ${z} lines",
"                        Avg.  : ${average}  bytes per article",
"                        Max.  : ${max_size}  bytes in article #{msg_no}",
"                        Server: ${server}  Port: ${portnumber}  GMT: ${time}",
"                      If a group contains no articles, then the lines 'Age'",
"                      'Avg.' and 'Max.' are omitted.",
"                      If a filter is set, then messages get counted only",
"                      if they pass that filter.",
" -grouplist_statistics start_expression filter_expression target_file",
"                      print a -group_statistics report for each group in the",
"                      grouplist (to be generated by -list_groups).",
"                      Reading and reporting does start with the first group",
"                      which matches the regular expression start_expression.",
"                      To be read, a group has to match the regular",
"                      expression filter_expression and (if set) the limits",
"                      given by option -group_stat_limits.",
"                      Both expressions may be inverted by the prefix '-not:'",
"",
" -list_interval start_number end_number",
"                      create a new message index. This file will contain",
"                      all headers of messages with numbers between",
"                      start_number and end_number (both included). Test the",
"                      headers with the current filter (which matches",
"                      everything if not restricted by option -filter ).",
" -fetch message_id    download the message with the given id from the preset",
"                      server and group. The id may be the message number or",
"                      the unique message id (looks like a mail address).",
" -fetch_interval start_number end_number",
"                      fetch all messages with numbers between start_number",
"                      and end_number (both included). Test the messages with",
"                      the current filter (which matches everything if not",
"                      restricted by option -filter ). ",
"",
" -adjust_links servername portnumber groupname",
"                      adjust the links in the group's message files to avoid",
"                      all skipped messages and to to connect messages from",
"                      different download runs. Also generate the group index",
"                      list_index.html in the group's directory.",
"                      If servername, portnumber or groupname are '-' then",
"                      the settings of the most recent -server, -port and",
"                      -group options are used.",
" -remove_old_files  servername portnumber groupname timedefinition",
"                      check the files related to the group wether they are",
"                      older than the defined time and eventually remove",
"                      them. This option finally runs  -adjust_links .",
"                      If servername, portnumber or groupname are '-' then",
"                      the settings of the most recent -server, -port and",
"                      -group options are used.",
"                      The group's directory has to contain a file with name",
"                       sfile_remove_old_files_allowed . This file may contain",
"                      a timedefinition which prevents remove runs with more",
"                      recent timepoints. If empty, no limit is imposed.",
" -remove_old_ids  timedefinition",
"                      check the files in the idlist directories wether they",
"                      are older than the given time.",
"                      The directory  msg_idlist  has to contain a file with",
"                      name  sfile_remove_old_files_allowed . This file may",
"                      contain a timedefinition which prevents remove runs",
"                      with more recent timepoints. No limit imposed if empty.",
" -release_all_locks   above operations are permitted only once at a time.",
"                      If the program aborts while such an operation, a lock",
"                      file 'maintenance_lock' may remain and prevent all",
"                      further actions. This option removes all lock files.",
"",
" -post_id  hostname  username  organization",
"                      set the components of the sender information in posted",
"                      messages. These will be used with the header lines",
"                      From , Organization and eventually Message-ID .",
"                      The default is \"somewhere\", \"anonymous\", \"none\"",
"                      Personal data from system files or mail settings will",
"                      not be published by this program.",
" -post_own_msgid  on|off",
"                      if -post_own_msgid is set to \"on\" then posted",
"                      messages will get a message id generated from local",
"                      properties and using the hostname given with -post_id",
"                      If set to \"off\" posted messages will rely on the",
"                      id generator of your news server. Default is \"off\".",
" -post_ref  message_id",
"                      set the message id for the \"Reference:\" header line.",
"                      The following messages will be posted as follow-ups to",
"                      the message with this message id.",
" -post_style  flat|indent|stair",
"                      set the posting style for multiple binary files.",
"                      In style \"flat\" all messages are refering to the one",
"                      defined by -post_ref (or to none without -post_ref).",
"                      Style \"indent\" lets only the first message refer",
"                      like in flat style. All other messages refer to this",
"                      first one.",
"                      Style \"stair\" lets each message refer to the previous",
"                      message. The first one refers like in flat style.", 
" -post  subject message_text [ binary_file [ ... binary_file ]]",
" -post_end",
"                      post messages to the preset server and group. At least",
"                      the two arguments subject and message_text have to",
"                      be given. All further arguments up to \"-post_end\"",
"                      are considered to be names of binary files.",
"                      -post_end is not necessary if no other options follow",
"                      the binary list.",
"                      If there is more than one binary file given, copies of",
"                      the message subject and message_text are posted",
"                      together with these additional files.",
"                      The format complies to RFC850 (pre-MIME). Binary files",
"                      are attached as uuencoded lines after message_text.",
" -cancel message_id   remove the depicted message from the news servers.",
"                      -post_id has to be the same as used when posting the",
"                      article.",
" -post_control control_text",
"                      posts a message with  control_text  as content of the",
"                      header lines Subject: and Control: and also as message",
"                      body. Use this option only after reading RFC850 and",
"                      its successors to learn about the meanings of various",
"                      control texts.",
"",
" -hashfilerec directory file [... file] [-hashfilerec_end]",
"                      records the given files in the hashdirectories below",
"                      the given directory. This can be used for -filter test",
"                      \"subjectfilefound:\"",
"",
" -read_from filename  read options from file. Empty lines and lines which",
"                      start with '#' or '!' are ignored. All others have to",
"                      contain at least one option and all its additional",
"                      arguments.",
"",
" -help               Print this text.",
"",
"",
" ---------------------------------------------------------------------------",
"                                 Control Files",
" ---------------------------------------------------------------------------",
"",
"The run can be controled by files in the base directory or in the",
"group's directory :",
"",
"  nntpclient_stop ..... if this file exists, the run ends before the next",
"                        message is fetched or the next option is executed.",
"                        The file gets removed by nntpclient.",
"  nntpclient_stop_all . similar to nntpclient_stop but this file does not",
"                        get removed by nntpclient. Do not forget to remove",
"                        this file when not needed any more.",
"  nntpclient_baudlimit  contains a number which limits the bandwidth used",
"                        by the client. See option -baudlimit above.",
"                        If a baudlimit is set, this file is checked after",
"                        each transmission at the communication port.",
"                        If no limit is set then this file is checked only",
"                        before fetching messges or executing options. Just",
"                        like the other files.",
"",
"The following files are only working in the group's directory :",
"",
"  nntpclient_pause .... a decimal number in the file tells how many seconds",
"                        to pause. The number is re-read once in a second.",
"                        If the file is removed externally then the",
"                        run is resumed. If the time elapsed is larger than",
"                        the current number in the file, then the file",
"                        gets removed and the run is resumed.",
"  nntpclient_index .... a decimal number in the file sets the current",
"                        message number in the current -fetch_interval or",
"                        -list_interval. Useful to skip nasty messages. But",
"                        use this option with great care.",
"  nntpclient_multi_bin_tee",
"                        if this file exists and the subject line matches the",
"                        regular expression .*[[(]1/[2-9][])].* then the",
"                        binary data are also stored undecoded in a file in",
"                        the group's bin directory. The file's name is",
"                        the subject line with some characters replaced by a",
"                        '+' and their ASCII-Code in hex. This is done for",
"                        those characters which are illegal or inconvenient",
"                        for file names. This filename gets the additional",
"                        ending .tee .",
"                        If the subject matches .*[[(][2-9]/[2-9][])].* then",
"                        the whole article body gets stored in such a .tee",
"                        file. The intention is to gather all parts of a",
"                        divided article's binary and put them together with",
"                        a pipe of cat and uudecode.",
"",
"This file works in the server+port directory :",
"",
"  nntpclient_no_xover   disables the use of NNTP extension XOVER which saves",
"                        a lot of time with -list_interval and filtered",
"                        -fetch_interval . XOVER might not be available on",
"                        all servers and the xover_*_* cache files occupy",
"                        some disk space.",
"",
" ---------------------------------------------------------------------------",
"                               Directory Structure",
" ---------------------------------------------------------------------------",
"",
"Root of the tree is the base directory. Usually $HOME/snntpbatch_download .",
"",
"   main_index.html",
"                HTML index of all subscribed newsgroups",
"   msg_idlist   Directory which memorizes unique message ids to prevent",
"                duplicate downloading.",
"      00 .. FA  subdirectories which are used as fields of a hashtable",
"      sfile_remove_old_files_allowed",
"                Gives permission to run -remove_old_ids and may contain a",
"                time limit which protects young files.",
"",
"   ${servername}.port${portnumber}",
"                Directory with all server specific data",
"      grouplist",
"                Text file with list of groups (result of NNTP command LIST)",
"      ${groupname}",
"                Directory with all data specific to a single group",
"         bin    Directory with extracted binary files",
"",
"         list_index.html",
"                Group index. Links to message index files.",
"         list_00000000_00000000.html",
"                Copy of newest message index.",
"         list_${startnumber}_${endnumber}.html",
"                Message index to messages within number range shown by name",
"         list_newest",
"                File containing the number of the last indexed message",
"         list_start",
"                File containing the number of the first message in the most",
"                recent message index",
"         maintenance_lock",
"                indicates that a maintenance process (-list_interval",
"                -adjust_links, -remove_old_files) is in progress and",
"                no other may be started. The file's content tells about",
"                hostname,process-id,client-id and the operation in progress.",
"         msg_${messagenumber}.html",
"                File containing the message text and HTML links to binary",
"                files which have been extracted.",
"         msg_first.html",
"                Copy of the first message in the directory",
"         msg_new.html",
"                Copy of the first message of newest message index",
"         msg_last.html",
"                Copy of the last message in the directory",
"         sfile_remove_old_files_allowed",
"                Gives permission to run -remove_old_files and may contain a",
"                time limit which protects young files.",
"",
" ---------------------------------------------------------------------------",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
 };
static char rc_text[][84]={
"On start the program looks in the users $HOME directory for the above",
"file. If it exists then it is executed like with option -read_from .",
"After this the same is done within the current working directory's parent",
"and afterwards with the current working directory itself.",
"The option -subscribe creates such a file in the group's directory if it",
"does not already exist. So if the program starts in a group's directory",
"it automatically connects to the appropriate server and group and has",
"its base directory set in a suitable way.",
"Login name and password for a news server may be defined in the server's",
"directory. Deny read permission to anyone else than yourself.",
"(see above: option -login).",
"If a global option is needed (like a different -timeout) then $HOME",
"is a good place to define it.",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};
 char line[4096];
 int i;
 
 MELDEN("\n");
 sprintf(line,"embedded in application %s :",client->progname);
 MELDEN(line);
 MELDEN("\n");
 
 for(i=0;strcmp(options_text[i],
                "@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
                )!=0;i++){
   sprintf(line,"%s\n",options_text[i]);
   MELDEN(line);
 }
 MELDEN("\n");

 MELDEN("----------------------------------------------------------------------------\n");
 sprintf(line,"Startupfiles with name '.%src' :\n",client->progname);
 MELDEN(line);
 MELDEN("----------------------------------------------------------------------------\n");
 MELDEN("\n");
 for(i=0;strcmp(rc_text[i],
                "@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
                )!=0;i++){
   sprintf(line,"%s\n",rc_text[i]);
   MELDEN(line);
 }
 MELDEN("\n");

 return(1);
}


int Nntpclient_complete_msgid(client,text,msgid,flag)
struct NntpclienT *client;
char *text;
char *msgid;
int flag;
{
 msgid[0]= 0;
 if(text[0]!=0 && text[0]!='<')
   strcat(msgid,"<");
 strcat(msgid,text);
 if(text[0]!=0)
   if(text[strlen(text)-1]!='>')
     strcat(msgid,">");
 return(1);
}


int Nntpclient_interpreter(client,argc,argv,ard,flag)
struct NntpclienT *client;
int argc;
char **argv;
int *ard;
int flag;
/*
 bit0= start at *ard rather than 1
*/
{
 char msgid[Nntpclient_strmaX];
 char line[4096],text[4096],*dir;
 int i,ret,reject_count=0,msgno;
 double limit;

 static verbous=0;

 client->argc= argc;
 client->argv= argv;
 client->argidx= -1;

 if(argc<=0) return(0);

 /* Remarks start with a lonely ! or # */
 if(argc>1)
   if(argv[1][0]=='#' || argv[1][0]=='!') {
     *ard= argc;
     return(2);
   }

 if(!(flag&1))
   *ard= 1;
 for(;(*ard)<argc;(*ard)++){
    if(Nntpclient_intrptchk(client,0)){
       client->conn->errn= -14;
       sprintf(client->conn->errmsg,
               "Nntpclient_interpreter stopped because : %s",
               client->user_messages);
       return(0);
   }
   client->argidx= *ard;
   client->filter->msgno_idx= -1;
   if(client->conn->errn)
     Nntpconn_errmsg_reset(client->conn,0);

   if(client->group!=NULL)
     if(client->group->name[0]!=0)
       Nntpclient_std_download_dir(client,1); /* set working directory */

   if(strcmp(argv[(*ard)],"-server")==0 && (*ard)+1<argc){

     (*ard)++;
     ret= Nntpclient_open(client,argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-port")==0 && (*ard)+1<argc){

     (*ard)++;
     if(sscanf(argv[(*ard)],"%ld",&(client->port))==1){
       if(client->servername[0]!=0) {
         ret= Nntpclient_open(client,client->servername,0);
         if(ret<=0)
           return(ret);
       }
     }

   } else if(strcmp(argv[(*ard)],"-group")==0 && (*ard)+1<argc){

     (*ard)++;
     ret= Nntpclient_open_group(client,argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-group_stat_limits")==0 && (*ard)+3<argc){

     double total_articles_limit;
     double total_bytes_limit,largest_bytes_limit;

     (*ard)++;
     sscanf(argv[(*ard)],"%lf",&total_articles_limit);
     (*ard)++;
     sscanf(argv[(*ard)],"%lf",&total_bytes_limit);
     (*ard)++;
     sscanf(argv[(*ard)],"%lf",&largest_bytes_limit);
     ret= Nntpstats_set_limits(client->stats,total_articles_limit,
                               total_bytes_limit,largest_bytes_limit,0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-group_statistics")==0 && (*ard)+1<argc){

     (*ard)++;
     ret= Nntpclient_group_stats(client,argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-grouplist_statistics")==0 && (*ard)+3<argc){

     (*ard)+= 3;
     ret= Nntpclient_grouplist_stats(client,argv[(*ard)-2],argv[(*ard)-1],
                                     argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-read_from")==0 && (*ard)+1<argc) {

     (*ard)++;
     ret= Nntpclient_read_from(client,argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-baudlimit")==0 && (*ard)+1<argc) {

     (*ard)++;
     sscanf(argv[(*ard)],"%lf",limit);
     Nntpclient_set_r_baudlimit(client,limit,0);

   } else if(strcmp(argv[(*ard)],"-subscribe")==0 && (*ard)+3<argc) {

     (*ard)+= 3;
     ret= Nntpclient_subscribe(client,argv[(*ard)-2],argv[(*ard)-1],
                                      argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-adjust_links")==0 && (*ard)+3<argc) {

     (*ard)+= 3;
     ret= Nntpclient_adjust_links(client,argv[(*ard)-2],argv[(*ard)-1],
                                         argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-remove_old_files")==0 && (*ard)+4<argc) {

     (*ard)+= 4;
     ret= Nntpclient_remove_old_files(client,argv[(*ard)-3],argv[(*ard)-2],
                                             argv[(*ard)-1],argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-remove_old_ids")==0 && (*ard)+1<argc) {

     (*ard)+= 1;
     ret= Nntpclient_remove_old_files(client,"no_server","0","no_group",
                                      argv[(*ard)],1);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-release_all_locks")==0) {

     ret= Nntpfiler_del_mntc_locks(client->filer,0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-timeout")==0 && (*ard)+1<argc){

     (*ard)++;
     ret= -1;
     sscanf(argv[(*ard)],"%ld",&ret);
     if(ret>=1)
       Nntpclient_set_timeout(client,ret,0);
     else {
       client->conn->errn= -16;
       sprintf(client->conn->errmsg,"Bad number with -timeout : %s",
                                    argv[(*ard)]);
       return(ret);
     }

   } else if(strcmp(argv[(*ard)],"-basedir")==0 && (*ard)+1<argc){

     (*ard)++;
     strcpy(client->basedir,argv[(*ard)]);
     Nntpfiler_set_basedir(client->filer,client->basedir,0);

   } else if( ( ( strcmp(argv[(*ard)],"-filter")==0 || 
                  strcmp(argv[(*ard)],"-not")==0 ||
                  strcmp(argv[(*ard)],"-or")==0 ||
                  strcmp(argv[(*ard)],"-and")==0 ||
                  strcmp(argv[(*ard)],"(")==0 ||
                  strcmp(argv[(*ard)],"-sub")==0
                ) && 
                 (*ard)+2<argc 
              ) || (
                strcmp(argv[(*ard)],")")==0 ||
                strcmp(argv[(*ard)],"-subend")==0
              )
            ) {

     ret= Nntpclient_filter_parser(client,argc,argv,ard,0);
     if(ret<=0){
       client->conn->errn= -18;
       if(client->filter->errmsg[0]!=0){
         strcpy(client->conn->errmsg,client->filter->errmsg);
       } else {
         sprintf(client->conn->errmsg,"Cannot create filter structure");
       } 
       return(ret);
     }

   } else if(strcmp(argv[(*ard)],"-set_test")==0 && (*ard)+2<argc){

     (*ard)+= 2;
     ret= Nntpfilter_set_regex(client->filter,argv[*ard-1],argv[*ard],0);
     if(ret<=0){
       client->conn->errn= -15;
       sprintf(client->conn->errmsg,"Cannot set filter expression : %s %s %s",
                                    argv[(*ard)-2],argv[(*ard)-1],argv[(*ard)]);
       return(ret);
     }

   } else if(strcmp(argv[(*ard)],"-timeout_bell")==0 && (*ard)+1<argc){

     (*ard)++;
     if(strcmp(argv[(*ard)],"on")==0)
       client->timeout_bell= 1;
     else if(strcmp(argv[(*ard)],"off")==0)
       client->timeout_bell= 0;
     else {
       client->conn->errn= -13;
       sprintf(client->conn->errmsg,"Unknown option with -timeout_bell : %s",
                                    argv[(*ard)]);
       return(0);
     }
   } else if(strcmp(argv[(*ard)],"-overwrite")==0 && (*ard)+1<argc){

     (*ard)++;
     if(strcmp(argv[(*ard)],"on")==0)
       client->overwrite= 1;
     else if(strcmp(argv[(*ard)],"off")==0)
       client->overwrite= 0;
     else {
       client->conn->errn= -13;
       sprintf(client->conn->errmsg,"Unknown option with -overwrite : %s",
                                    argv[(*ard)]);
       return(0);
     }

   } else if(strcmp(argv[(*ard)],"-html")==0 && (*ard)+1<argc){

     (*ard)++;
     if(strcmp(argv[(*ard)],"on")==0)
       client->html= 1;
     else if(strcmp(argv[(*ard)],"off")==0)
       client->html= 0;
     else {
       client->conn->errn= -13;
       sprintf(client->conn->errmsg,"Unknown option with -html : %s",
                                    argv[(*ard)]);
       return(0);
     }
     Nntpfiler_set_html(client->filer,client->html,0);

   } else if(strcmp(argv[(*ard)],"-list_groups")==0 ){

     ret= Nntpclient_list_groups(client,0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-list_interval")==0 && (*ard)+2<argc){

     ret= Nntpclient_list_interval(client,argv,ard,0);
     if(ret<=0)
       return(ret); 
     
   } else if(strcmp(argv[(*ard)],"-fetch")==0 && (*ard)+1<argc){

     (*ard)++;
     ret= Nntpclient_fetch(client,argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-fetch_interval")==0 && (*ard)+2<argc){

     ret= Nntpclient_fetch_interval(client,argv,ard,0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-post_id")==0 && (*ard)+3<argc){

     (*ard)+= 3;
     strcpy(client->hostid,argv[(*ard)-2]);
     strcpy(client->username,argv[(*ard)-1]);
     strcpy(client->organization,argv[(*ard)]);

   } else if(strcmp(argv[(*ard)],"-post_ref")==0 && (*ard)+1<argc){

     (*ard)+= 1;
     ret= Nntpclient_complete_msgid(client,argv[(*ard)],client->reference,0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-post_style")==0 && (*ard)+1<argc){

     (*ard)+= 1;
     if(strcmp(argv[(*ard)],"flat")==0)
       client->posting_style= (client->posting_style&~15) | 0;
     else if(strcmp(argv[(*ard)],"indent")==0)
       client->posting_style= (client->posting_style&~15) | 1;
     else if(strcmp(argv[(*ard)],"stair")==0)
       client->posting_style= (client->posting_style&~15) | 2;
     else {
       client->conn->errn= -13;
       sprintf(client->conn->errmsg,"Unknown option with -post_style : %s",
                                    argv[(*ard)]);
       return(0);
     }

   } else if(strcmp(argv[(*ard)],"-post_own_msgid")==0 && (*ard)+1<argc){

     (*ard)+= 1;
     if(strcmp(argv[(*ard)],"on")==0)
       client->use_own_msgid= 1;
     else
       client->use_own_msgid= 0;

   } else if(strcmp(argv[(*ard)],"-post")==0 && (*ard)+2<argc){

     ret= Nntpclient_post_interpreter(client,argc,argv,ard,0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-post_control")==0 && (*ard)+1<argc){

     (*ard)+= 1;
     ret= Nntpclient_control(client,argv[(*ard)],0);
     if(ret<=0)
       return(ret);

   } else if(strcmp(argv[(*ard)],"-cancel")==0 && (*ard)+1<argc){

     (*ard)+= 1;
     ret= Nntpclient_complete_msgid(client,argv[(*ard)],line,0);
     if(ret<=0)
       return(ret);
     ret= Nntpclient_cancel(client,line,0);
     if(ret<=0)
       return(ret);
 
   } else if(strcmp(argv[(*ard)],"-hashfilerec")==0 && (*ard)+1<argc){
      
     (*ard)+= 1;
      dir= argv[*ard];
      for((*ard)++;(*ard)<argc;(*ard)++){
        if(strcmp(argv[*ard],"-hashfilerec_end")==0)
      break;
        ret= Sfile_string_to_hashdir(dir,argv[*ard],"",1|2);
        if(ret<=0) {
          client->conn->errn= errno;
          sprintf(client->conn->errmsg,
                  "Error %ld while -hashfilerec %s %s\n%s",
                  errno,dir,argv[*ard],strerror(errno));
          return(ret);
        }
      }

   } else if(strcmp(argv[(*ard)],"-login")==0 && (*ard)+2<argc){

     (*ard)+= 2;
     Nntpclient_set_auth(client,argv[(*ard)-1],argv[(*ard)],0);

   } else if(strcmp(argv[(*ard)],"-help")==0){

     Nntpclient_help(client,0);

   } else {
     
     client->conn->errn= -13;
     sprintf(client->conn->errmsg,"Unknown or incomplete option : %s",
                                  argv[(*ard)]);
     return(-1);

   } 
 }

 return(1);
}


int Nntpclient_try_reconnect(client,flag)
struct NntpclienT *client;
int flag;
{
 int tio_count= 0,ret= 0;
 char text[Nntpclient_strmaX];

 while(1) {
   Nntpconn_errmsg_reset(client->conn,0);
   sprintf(text,"\ntrying to reconnect ...\n");
   Nntpclient_result(client,text,0);
   if(client->conn==NULL){
     ret= 0;
     goto ex;
   }
   Nntpconn_close(client->conn,1<<15);
   ret= Nntpconn_open(client->conn,0);
   if(ret>0) 
 break;
   if(client->conn->errn!=-6)
     goto ex;
   Nntpconn_close(client->conn,1<<15);
   tio_count++;
   sprintf(text,"... failed. Timeout again. Sleeping for %ld seconds.\n",
                 tio_count*5);
   Nntpclient_result(client,text,0);
   sleep(tio_count*5);
 }
 /* reopen group */
 if(client->group!=NULL){
   ret= Nntpclient_open_group(client,client->group->name,0); 
   if(ret<=0)
     goto ex;
 }

 ret= 1;
ex:;
 return(ret);
}


int Nntpclient_outer_fence(client,argc,in_argv,ard,flag)
/* a wrapper around Nntpclient_interpreter() to handle connection problems */
struct NntpclienT *client;
int argc;
char **in_argv;
int *ard;
int flag;
/*
 bit0= start at *ard rather than 1
*/
{
 char text[Nntpclient_strmaX],targetname[4096],targetadr[4096];
 char **argv,*argdat;
 int ret,i,l,tio_count= 0;
 int t_int;
 double t_time= 0.0, t_bytes= 0.0, t_wait= 15.0,sleep_scale= 1.0;

 /* create a copy of arguments which can be altered */
 l= 16; /* so any number will fit */
 for(i=0;i<argc;i++)
   if(strlen(in_argv[i])>l) l= strlen(in_argv[i]);
 argv= TSOB_FELD(char *,argc);
 argdat= TSOB_FELD(char,(l+1)*argc);
 if(argv==NULL || argdat==NULL) {
   client->conn->errn= -17;
   sprintf(client->conn->errmsg,"Cannot create copy of options.");
   return(-1);
 }
 for(i=0;i<argc;i++) {
   argv[i]= &(argdat[i*(l+1)]);
   strcpy(argv[i],in_argv[i]);
 }

 while(1) {
   ret= Nntpclient_interpreter(client,argc,argv,ard,flag);
   if(ret>0) 
     goto ex;
   if(client->conn==NULL)
     goto ex;
   flag|= 1; /* next time start interpreter at current *ard */

   if(client->conn->errn==-6) { /* timeout : try to resume the job */

     /* emergency close and re-establish connection */
     while(1){
       Nntpconn_close(client->conn,1<<15);
       tio_count++;
       sprintf(text,"\n********** TIMEOUT **********\n");
       Nntpclient_result(client,text,0);
       sprintf(text,"  %s\n",client->conn->errmsg);
       Nntpclient_result(client,text,0);
       sprintf(text,"while working on message #%ld\n\n",
                    client->filter->msgno_idx);
       Nntpclient_result(client,text,0);
bell_and_reconnect:;
       if(client->timeout_bell && tio_count<=3){
         Nntpclient_result(client,"\007",0);
         sleep(1);
         Nntpclient_result(client,"\007",0);
         sleep(1);
         Nntpclient_result(client,"\007",0);
       }
try_reconnect:;
       ret= Nntpclient_try_reconnect(client,0);
       if(ret<=0)
         goto ex;
     break;
     }
     *ard= client->argidx;

     /* if -*_interval : fake new start number */;
     ret= Sregex_is_match_t(".*_interval",argv[*ard],0);
     if(ret<0){
       client->conn->errn= -17;
       sprintf(client->conn->errmsg,
               "Cannot set regular expression for restart");
       return(-1);
     } else if (ret>0) {
       sprintf(argv[(*ard)+1],"%ld",client->filter->msgno_idx);
     }

     Nntpconn_errmsg_reset(client->conn,0);
     tio_count= 0;

   } else if(client->conn->errn==-11){ /* message was already downloaded */

     fprintf(stderr,"--- %s\n",client->conn->errmsg);
     Nntpconn_errmsg_reset(client->conn,0);

   } else if(client->conn->errn==-29){

     Nntpconn_close(client->conn,1<<15);
     sprintf(text,"\n********** SOCKET RETURNS EMPTY DATA **********\n");
     Nntpclient_result(client,text,0);
     goto bell_and_reconnect;

   } else if(client->conn->errn==-37){ /* server returns much less
                                          article bytes than announced */
     if(!client->check_for_short_article) {
       Nntpconn_errmsg_reset(client->conn,0);
       ret= 1; 
       goto ex;
     }
     Nntpconn_close(client->conn,1<<15);
     sprintf(text,"\n********** SERVER FAILURE **********\n");
     Nntpclient_result(client,text,0);
     sprintf(text,"while working on message #%ld (attempt %ld of %ld)\n\n",
                  client->filter->msgno_idx,client->short_article_counter+1,
                  client->short_article_maxtry);
     Nntpclient_result(client,text,0);
     Nntpclient_result(client,client->conn->errmsg,0);
     if(client->timeout_bell){
       Nntpclient_result(client,"\007",0);
       sleep(1);
       Nntpclient_result(client,"\007",0);
       sleep(1);
       Nntpclient_result(client,"\007",0);
       sleep(1);
       Nntpclient_result(client,"\007",0);
       sleep(1);
       Nntpclient_result(client,"\007",0);
     }
     client->short_article_counter++;
     sleep_scale= 3.0;
     if(client->short_article_maxtry>0) {
       if(client->short_article_counter>=client->short_article_maxtry) {
         sprintf(text,"--- TOO MANY FAILURES ON MESSAGE #%ld\n\n",
                      client->filter->msgno_idx);
         Nntpclient_result(client,text,0);
         sprintf(text,"--- message skipped ---\n\n",
                      client->filter->msgno_idx);
         Nntpclient_result(client,text,0);
         client->filter->msgno_idx++;
         client->short_article_counter= 0;
       }
       sleep_scale= 2.0+2.0*((double) client->short_article_counter)/
                            ((double) client->short_article_maxtry);
     }
wait_and_reconnect:;
     t_time= time(0);
     t_int= t_time/31.0;
     t_wait= ((t_time/31.0)-(double) t_int)*31.0*sleep_scale;
     t_wait+= 15.0;
     sprintf(text,"will try to reconnect in %.lf seconds\n",t_wait);
     Nntpclient_result(client,text,0);
     sleep((int) t_wait);
     goto try_reconnect;
   } else if(client->conn->errn==-38){ 

     Nntpconn_close(client->conn,1<<15);
     sprintf(text,"\n********** CONNECTION BROKEN **********\n");
     Nntpclient_result(client,text,0);
     sprintf(text,"   %s\n\n",client->conn->errmsg);
     Nntpclient_result(client,text,0);
     Nntpclient_result(client,"\007",0);
     sleep(1);
     Nntpclient_result(client,"\007",0);
     sleep_scale= 0.3;
     goto wait_and_reconnect;

   } else {
     goto ex;
   }
 }
 return(-1); /* should not be reachable */

ex:;
 if(argv!=NULL)
   free((char *) argv);
 if(argdat!=NULL)
   free((char *) argdat);
 return(ret);
}


int Nntpclient_rc(client,progname,flag)
struct NntpclienT *client;
char *progname;
int flag;
/*
 bit0= HOME
 bit1= current
 bit2= parent
*/
{
 char fileadr[SfileadrL],rawname[SfileadrL],*cpt;
 int ret;

 cpt= strrchr(progname,'/');
 if(cpt==NULL)
   cpt= progname;
 else
   cpt++;
 Nntpclient_set_progname(client,cpt,0);
 sprintf(rawname,".%src",cpt);
 if(flag&1){
   ret= Sfile_home_adr(rawname,fileadr,0);
   if(ret>0) {
     ret= Nntpclient_read_from(client,fileadr,1);
     if(ret<0) /* error occured after file was opened successfully */
       return(ret);
   }
 }
 if(flag&4){
   sprintf(fileadr,"../%s",rawname);
   ret= Nntpclient_read_from(client,fileadr,1);
   if(ret<0) /* error occured after file was opened successfully */
     return(ret);
 }
 if(flag&2){
   ret= Nntpclient_read_from(client,rawname,1);
   if(ret<0) /* error occured after file was opened successfully */
     return(ret);
 }
 return(1);

}

