/*
  cc -g -c nntpconn.c
*/


#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

/* Macro for allocation of object arrays */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));

#ifndef RID_WINNT351
    struct hostent *gethostbyname() ;
#define closesocket close
#endif


#include "../s_tools/sfile.h"

#include "nntpfiler.h"
#include "nntpconn.h"


/* -------------------------------- NntpconN ------------------------------ */


int Nntpconn_new(conn,servername,portnumber,flag)
struct NntpconN **conn;
char *servername;
int portnumber;
int flag;
/*
 bit0= dummy (read from file servername)
 bit1= do not open connection
*/
{
 struct NntpconN *c;
 *conn= c= TSOB_FELD(struct NntpconN,1);
 if(c==NULL) 
   return(-1); 
 strcpy(c->servername,servername);
 c->portnumber= portnumber;
 c->allows_posting= 0;
 c->dummy= (flag&1);
 c->login[0]= 0;
 c->password[0]= 0;
 c->sock= -1;
 c->active= 0;
 c->timeout= 86400;
 c->slow_server_limit= -1.0;
 c->line[0]= 0;
 c->line_valid= 0;
 c->endseq_pt= 0;
 c->zero_data_count= 0;
 c->zero_data_max= 10;
 c->show_io_stat= 1;
 Nntpconn_reset_t_count(c,0);
 c->baudlimitfile[0]= 0;
 c->r_baudlimit= 0.0;
 Nntpconn_reset_rw_count(c,0);
 c->sent_cmd[0]= 0;
 c->status_line[0]= 0;
 c->status_code= -1;
 c->buffer[0]= 0;
 c->buffrpt= c->buffer;
 c->buffer_status= 0;
 c->errn= 0;
 c->errmsg[0]= 0;
 c->outlet= NULL;
 c->outlet_name[0]= 0;
 c->filer= NULL;
 c->groupname[0]= 0;
 if(!(flag&2)){
   if(Nntpconn_open(*conn,0)<=0)
     return(0);
 }
 return(1);
}


int Nntpconn_destroy(conn,flag)
struct NntpconN **conn;
int flag;
{
 struct NntpconN *c;

 if((*conn)==NULL)
   return(0);
 c= *conn;
 Nntpconn_close(c,0);
 free((char *) c);
 *conn= NULL;  
 return(1);
}


int Nntpconn_open(conn,flag)
struct NntpconN *conn;
int flag;
{
 int sret,eins=1;
 struct sockaddr_in  server ;
 struct hostent *hp;
 char errmsg[Nntpconn_strL];
 int errn;

 conn->active= 1;
 conn->sent_cmd[0]= 0;
 if(conn->dummy) {
   conn->sock= open(conn->servername,O_RDONLY);
   if(conn->sock<0){
     conn->errn= errno;
     sprintf(conn->errmsg,
             "Error %d on opening input file (dummy socket): %s\n%s",
             errno,conn->servername,strerror(errno));
     return(0);
   }
   lseek(conn->sock,0,SEEK_SET);
   conn->active= 1;
   return(1);
 }

 conn->sock = socket(AF_INET, SOCK_STREAM, 0);
 if(conn->sock<0){
   conn->errn= errno;
   sprintf(conn->errmsg,"Error %d on opening inet stream socket:\n%s",
                        errno,strerror(errno));
   return(0); 
 }
#ifdef Nntpconn_keepalivE
 setsockopt(conn->sock,SOL_SOCKET,SO_KEEPALIVE,&eins,sizeof(eins));
#endif /* Nntpconn_keepalivE */
 server.sin_family = AF_INET ;
 hp = gethostbyname(conn->servername);
 if(hp == 0) {
   conn->errn= errno;
   sprintf(conn->errmsg, "unknown host : %s", conn->servername) ;
   closesocket(conn->sock);
   conn->sock= -1;
   return(0) ;
 }
 memcpy((char *)&server.sin_addr, (char *)hp->h_addr, hp->h_length) ;
 server.sin_port = htons(conn->portnumber);
 if(connect(conn->sock, (struct sockaddr *)&server, sizeof server) < 0) {
   conn->errn= errno;
   sprintf(conn->errmsg,"Error %d on connecting stream socket:\n%s",
                        errno,strerror(errno)) ;
   closesocket(conn->sock);
   conn->sock= -1;
   return(0) ;
 }
#ifdef Nntpconn_keepalivE
 setsockopt(conn->sock,SOL_SOCKET,SO_KEEPALIVE,&eins,sizeof(eins));
#endif /* Nntpconn_keepalivE */
 conn->active= 1;

 /* get status line from server and check it */
 sret= Nntpconn_recv_status(conn,0);
 if(sret<=0)
   goto close_again;
 Nntpconn_result(conn,conn->status_line,"",0,"",1);
 if(conn->status_code==200)
   conn->allows_posting= 1;
 else if(conn->status_code==201)
   conn->allows_posting= 0;
 else
   goto close_again;

 return(1);

close_again:;

 fprintf(stderr,"--- use of opened connection failed. closing.\n");

 strcpy(errmsg,conn->errmsg);
 errn= conn->errn;
 Nntpconn_close(conn,0);
 strcpy(conn->errmsg,errmsg);
 conn->errn= errn;
 return(0);
}


int Nntpconn_close(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit15= Emergency close : no QUIT, no waiting for answer
*/
{
 char *merk;
 int sret;

 if(conn->active==0) 
   return(2);

 if(!(flag&(1<<15))) {
   /* try to log out at server, save current error message */
   merk= (char *) malloc(strlen(conn->errmsg)+1);
   strcpy(merk,conn->errmsg);
   Nntpconn_result(conn,"\nQUIT\n","",0,"",1);
   if(conn->errn==-6) /* a timeout has occured */
     conn->timeout= 5;
   sret= Nntpconn_transaction(conn,"QUIT","",0,"",1);
   strcpy(conn->errmsg,merk);
 }

 /* close TCP/IP connection */
 if(conn->dummy)
   close(conn->sock);
 else
   closesocket(conn->sock);
 conn->sock= -1;
 conn->active= 0;

 return(1);
}


int Nntpconn_set_auth(conn,login,password,flag)
struct NntpconN *conn;
char *login,*password;
int flag;
{
 strcpy(conn->login,login);
 strcpy(conn->password,password);
 return(1);
}


int Nntpconn_set_outlet(conn,fp,name,flag)
struct NntpconN *conn;
FILE *fp;
char *name;
int flag;
{
 conn->outlet= fp;
 strcpy(conn->outlet_name,name);
 return(1);
}


int Nntpconn_set_timeout(conn,timeout,flag)
struct NntpconN *conn;
int timeout;
int flag;
{
 if(conn==NULL)
   return(0);
 conn->timeout= timeout;
 return(1);
}


int Nntpconn_set_slow_server(conn,limit,flag)
struct NntpconN *conn;
double limit;
int flag;
{
 if(conn==NULL)
   return(0);
 conn->slow_server_limit= limit;
 return(1);
}


int Nntpconn_set_r_baudlimit(conn,limit,flag)
struct NntpconN *conn;
double limit;
int flag;
{
 if(conn==NULL)
   return(0);
 conn->r_baudlimit= limit;
 Nntpconn_reset_rw_count(conn,0);
 return(1);
}


int Nntpconn_get_t_count(conn,t_start,t_bytecount,flag)
struct NntpconN *conn;
double *t_start;
double *t_bytecount;
int flag;
{
 *t_start= conn->t_start;
 *t_bytecount= conn->t_bytecount;
 return(1);
}


int Nntpconn_set_t_count(conn,t_start,t_bytecount,flag)
struct NntpconN *conn;
double t_start;
double t_bytecount;
int flag;
{
 conn->t_start= t_start;
 conn->t_bytecount= t_bytecount;
 return(1);
}


int Nntpconn_reset_t_count(conn,flag)
struct NntpconN *conn;
int flag;
{
 conn->t_start= Sfile_microtime(0);
 conn->t_bytecount= 0;
 return(1);
}


int Nntpconn_reset_rw_count(conn,flag)
struct NntpconN *conn;
int flag;
{
 conn->start_time= Sfile_microtime(0);
 conn->r_bytecount= 0;
 conn->w_bytecount= 0;
 return(1);
}


int Nntpconn_set_groupname(conn,groupname,flag)
struct NntpconN *conn;
char *groupname;
int flag;
{
 strcpy(conn->groupname,groupname);
 if(conn->filer==NULL)
   return(2);
 strcpy(conn->filer->groupname,groupname);
 return(1);
}


int Nntpconn_get_groupname(conn,groupname,flag)
struct NntpconN *conn;
char *groupname;
int flag;
{
 groupname[0]= 0;
 if(conn->filer==NULL)
   strcpy(groupname,conn->groupname);
 else
   strcpy(groupname,conn->filer->groupname);
 return(1);
}


int Nntpconn_set_filer(conn,filer,flag)
struct NntpconN *conn;
struct NntpfileR *filer;
int flag;
{
 conn->filer= filer;
 return(1);
}


int Nntpconn_err_not_active(conn,flag)
struct NntpconN *conn;
int flag;
{
 conn->errn= -2;
 sprintf(conn->errmsg,"NNTP connection not active (to %s %ld)",
                      conn->servername,conn->portnumber);
 return(-2);
}


int Nntpconn_result(conn,line,recv_text,recv_siz,save_name,flag)
struct NntpconN *conn;
char *line;
char *recv_text;
unsigned long recv_siz;
char *save_name;
int flag;
/*
 bit0= write reply to stdout
 bit1= append reply to recv_text
 bit2= append reply to file save_name
 bit3= write reply to conn->outlet
*/
{
 unsigned long l;
 int ret;
 FILE *fp;

 if(flag&1)
   fprintf(stdout,"%s",line);

 if(flag&2){
   if(line==recv_text){
     fprintf(stderr,"Nntpconn_result: Programmer's fault: line==recv_text\n");
     exit(1);
   }
   l= strlen(recv_text);
   if(l+strlen(line)>=recv_siz){
     conn->errn= -1;
     sprintf(conn->errmsg,"Result buffer overflow (%lu >= %lu)",
                          l+strlen(line),recv_siz);
     return(0);
   }
   strcpy(recv_text+l,line);
 }

 if(flag&4){
   fp= fopen(save_name,"a");
   if(fp==NULL){
     conn->errn= errno;
     sprintf(conn->errmsg,"Error %ld while opening result file %s\n%s",
                          errno,save_name,strerror(errno));
     return(-1);
   }
   fprintf(fp,"%s",line);
   fclose(fp);
 }

 if((flag&8) && conn->outlet!=NULL){
   if(fwrite(line,strlen(line),1,conn->outlet)<=0){
     conn->errn= errno;
     sprintf(conn->errmsg,"Error %ld on writing output file : %s\n",
                          errno,conn->outlet_name,strerror(errno));
     return(-1);
   }
 }

 return(1);
}


int Nntpconn_check_slow_server(conn,flag)
struct NntpconN *conn;
int flag;
/*
 return:
  -1= internal error
   0= slow service (synthetic timeout error)
   1= ok, not too slow
*/
{
 double t;
 if(conn->slow_server_limit<=0)
   return(1);
 t= Sfile_microtime(0)-conn->t_start;
 if(conn->timeout*2<t)
   if((conn->t_bytecount)/t < conn->slow_server_limit) {
     conn->errn= -6;
     sprintf(conn->errmsg,
             "Slow service: rate is only %ld =%.lfb/%.1lfs\n",
                           (int) (conn->t_bytecount/t),conn->t_bytecount,t);
     return(0);
   }
 return(1);
}


int Nntpconn_show_stat(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit0= clear full line and exit
 bit1= always show cmd
 bit3= display "speed" rather than sent_cmd
 bit4= display with newline
*/
{
 struct timeval tv;
 struct timezone tz;
 double t;
 int i;
 char cmd[81],line[321];

 if(conn->show_io_stat>2 || (flag&1)){
   fprintf(stderr,"\r%-79.79s\r"," ");
   conn->show_io_stat= 2-(conn->show_io_stat%2);
   if(flag&1)
     return(1);
 }
 t= Sfile_microtime(0)-conn->t_start;
 if(conn->sent_cmd[0]) {
   sscanf(conn->sent_cmd,"%s",cmd);
   strcat(cmd," | ");
 } else 
   strcpy(cmd,"connected, waiting for welcome message | ");
 if((conn->show_io_stat%2)==0 && !(flag&2)) 
   for(i=0;cmd[i]!=0 && cmd[i]!='|';i++) cmd[i]= ' ';
 conn->show_io_stat= 3-conn->show_io_stat;
 if(flag&8)
   strcpy(cmd,"  speed");
 if(t>0)
   sprintf(line,"%s %ld =%.lfb/%.1lfs ",cmd,(int) (conn->t_bytecount/t),
                                        conn->t_bytecount,t);
 else 
   sprintf(line,"%s %ld =%.lfb/%.1lfs ",cmd,0,conn->t_bytecount,t);
 fprintf(stderr,"\r%-40.40s",line);
 if(flag&16)
   fprintf(stderr,"\n");
 return(1);
}


int Nntpconn_get_stat(conn,t_time,t_bytes,flag)
struct NntpconN *conn;
double *t_time;
double *t_bytes;
int flag;
{
 *t_time= Sfile_microtime(0)-conn->t_start;
 *t_bytes= conn->t_bytecount;
 return(1);
}


int Nntpconn_select(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit0+1= mode:  0=read ; 1= write ; 2= exception
 bit2= timeout only 0.1 seconds on first try 
 bit3= only display statistics (bit0+1 have to be 0, conn->show_io_stat!=0)
 bit4= display with newline
 bit5= do not check for slow service
*/
{
 int mode,sret,i,show_stat=0,ret;
 fd_set rds,wts,exs;
 struct timeval tio;
 char text[81],*blink= NULL,cmd[81];
 static char modetext[4][16]={"Read","Write","Exception",""};

 if(conn->active==0)
   return(Nntpconn_err_not_active(conn,0));

 if(conn->dummy)
   return(1);

 mode= flag&3;
 if(conn->show_io_stat && (mode==0||mode==1)){
   Nntpconn_show_stat(conn,flag&(8|16));
   show_stat= 1;
   if(flag&8)
     return(1);
 }
 for(i=0;i<conn->timeout+1;i++){
   FD_ZERO(&rds);
   FD_ZERO(&wts);
   FD_ZERO(&exs);
   if     (mode==0) FD_SET(conn->sock,&rds);
   else if(mode==1) FD_SET(conn->sock,&wts);
   else if(mode==2) FD_SET(conn->sock,&exs);
   if(i==0 && (flag&4)){
     tio.tv_sec= 0;
     tio.tv_usec= 100000;
   } else {
     tio.tv_sec= 1;
     tio.tv_usec= 0;
   }


   /* >>> ASNNTP replace by call to Asagent_loop() */
   if((sret= select(conn->sock+1,&rds,&wts,&exs,&tio))!=0)
 break;



   fprintf(stderr,
"\r                                                                          \r"
          );
   if(show_stat) {
     Nntpconn_show_stat(conn,0);
     fprintf(stderr," | idle  %lds  (max %lds)",i+1,conn->timeout);
     conn->show_io_stat= 3;
   }
   if(!(flag&32)) {
     ret= Nntpconn_check_slow_server(conn,0);
     if(ret<0)
       return(ret);
     if(ret==0)
       return(-2);
   }
 }
 fprintf(stderr,
"\r                                                                          \r"
        );
 if(sret==-1){
   conn->errn= errno;
   sprintf(conn->errmsg,"Error %d on reading on stream socket:\n%s",
                        errno,strerror(errno));
   return(-1);
 }else if(sret==0){
   conn->errn= -6;
   sprintf(conn->errmsg,"%s-timeout after %ld sec\n",
                        modetext[mode],conn->timeout);
   return(-2);
 }
 return(1);
}


int Nntpconn_send(conn,send_text,flag)
struct NntpconN *conn;
char *send_text;
int flag;
/*
 bit0= raw: do not append crlf
*/
{
 int l,sret;
 char *line;
 line= conn->line;

 if(conn->active==0)
   return(Nntpconn_err_not_active(conn,0));

 sret= Nntpconn_select(conn,1);
 if(sret<=0)
   return(sret);
 strcpy(line,send_text);
 if(!(flag&1))
   Sfile_append_crlf(line,0);
 l= strlen(line);
 if(conn->dummy)
   sret= l;
 else
   sret= send(conn->sock,line,l,0);
 if(sret<0){
   conn->errn= errno;
   sprintf(conn->errmsg,"Error %d on writing on stream socket:\n%s",
                        errno,strerror(errno));
   return(-1);
 } else {
   conn->t_bytecount+= sret;
   conn->w_bytecount+= sret;
 }
 return(1);
}


int Nntpconn_recv(conn,line,size,flag)
struct NntpconN *conn;
char *line;
int size;
int flag;
{
 int ret,i;

 if(conn->dummy) {
   i= 0;
   while((ret= read(conn->sock,line+i,1))==1){
     i++;
     if(line[i-1]=='\n' || i>=size)
   break;
   }
   if(ret>=0)
     ret= i; 
 } else
   ret= recv(conn->sock,line,size,0);
 if(ret>0){
   conn->r_bytecount+= ret;
   conn->t_bytecount+= ret;
   conn->zero_data_count= 0;
 } else if(ret==0) {
   conn->zero_data_count++;
   if(conn->zero_data_count>conn->zero_data_max) {
     conn->errn= -29;
     sprintf(conn->errmsg,"Stream socket returns empty data\n");
     return(-2);
   }
 } else { /* ret<0 */
   conn->errn= -38;
   sprintf(conn->errmsg,"Connection broken. errno=%ld  '%s'",
                        errno,strerror(errno));
   return(-2);
 }
 return(ret);
}


int Nntpconn_recv_status(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit0= preliminarily show the incoming characters
*/
{
 int l,sret,cr_da,i;
 char *line;
 line= conn->status_line;

 if(conn->active==0)
   return(Nntpconn_err_not_active(conn,0));

 cr_da= 0;
 i= 0;
 sret= Nntpconn_select(conn,0);
 if(sret<=0)
   return(sret);
 while(1){
   sret= Nntpconn_recv(conn,line+i,1,0);
   if(sret<0) {
     if(sret!=-2){
       conn->errn= errno;
       sprintf(conn->errmsg,"Error %d on reading on stream socket:\n%s",
                            errno,strerror(errno));
     }
     return(-1);
   }
   if(line[i]=='\r')
     cr_da= 2;
   if(line[i]=='\n' && cr_da>0)
 break;
   cr_da--;
   i++;
   line[i]= 0;
   if(flag&1)
     fprintf(stderr,"%c",line[i-1]);
 }
 if(flag&1)
   fprintf(stderr,"\r");
 line[++i]= 0;
 conn->status_code= 0;
 sscanf(line,"%ld",&conn->status_code);
 return(1);
}


int Nntpconn_recv_text(conn,recv_text,recv_siz,save_name,flag)
struct NntpconN *conn;
char *recv_text;
unsigned long recv_siz;
char *save_name;
int flag;
/*
 bit0= write reply to stdout
 bit1= write reply to recv_text
 bit2= append reply to file save_name
 bit8= only check wether there are still data expected
*/
{
 int l,siz,sret,cr_da,i,status,rret,count;
 char *line;
 static char endseq[]= {13, 10, '.', 13, 10, 0, 0, 0, 0, 0, 0 };

 line= conn->line;
 siz= sizeof(conn->line)-1;
 if(siz>recv_siz-1) 
   siz= recv_siz-1;

 line[0]= 0;

 if(conn->active==0)
   return(Nntpconn_err_not_active(conn,0));
 if(endseq[conn->endseq_pt]==0)
   return(0);
 if(flag&256)
   return(1);

 count= 0;
 while(1){
   /* get a chunk of data */
   sret= Nntpconn_select(conn,0);
   if(sret<=0)
     return(sret);
   rret= Nntpconn_recv(conn,line,siz,0);
   if(rret<0) {
     if(rret!=-2){
       conn->errn= errno;
       sprintf(conn->errmsg,"Error %d on reading on stream socket:\n%s",
                            errno,strerror(errno));
     }
     return(-1);
   }
   line[rret]= 0;

   /* are we done ? */
   for(i=(rret<5?0:rret-5);i<rret;i++) {
     if(endseq[conn->endseq_pt]==line[i])
       conn->endseq_pt++;
     else {
       if(conn->endseq_pt>0)
         i--;
       conn->endseq_pt= 0;
     }
     if(endseq[conn->endseq_pt]==0) {
       Nntpconn_show_stat(conn,1); /* clear line */
   break;
     }
   }

   sret= Nntpconn_result(conn,line,recv_text,recv_siz,save_name,flag&7);
   if(sret<=0){
     return(-1);
   }

   if(endseq[conn->endseq_pt]==0)
     goto fertig;
 }

fertig:;
 return(1);
}


int Nntpconn_enforce_limits(conn,filename,flag)
struct NntpconN *conn;
char *filename;
int flag;
/*
 bit0= read configuration files
 bit1= do not wait (just read)
 bit2= do nothing if conn->baudlimit>0
*/
{ 
 double countlimit, num= -1.0;
 FILE *fp;
 int ret,count= 0;

 if((flag&4) && conn->r_baudlimit<=0) 
   return(2);

 if((flag&1)&&filename[0]!=0){
   if(conn->baudlimitfile!=filename)
     strcpy(conn->baudlimitfile,filename);

   ret= Sfile_file_to_double(filename,&num,0);
   if(ret<=0 || num<=0.0)
     num= 0.0;
   if(num!=conn->r_baudlimit)
     Nntpconn_set_r_baudlimit(conn,num,0);
 }
 if(flag&2)
   return(1);

 if(conn->r_baudlimit>0) {
   while(1){
     countlimit= conn->r_baudlimit
                 * ((double) (Sfile_microtime(0)-conn->start_time));
     if(conn->r_bytecount+conn->w_bytecount<=countlimit)
   break;
     if(count==0)
       Nntpconn_show_stat(conn,1); /* clear line */
     if(conn->show_io_stat && !(count%2))
       Nntpconn_show_stat(conn,2);
     Sfile_microsleep(100000,0);
     count++;
   }
   if(conn->show_io_stat && count)
     Nntpconn_show_stat(conn,0);
 }
 return(1);
}


int Nntpconn_transaction_reset(conn,flag)
struct NntpconN *conn;
int flag;
/*
 bit3= do not reset transaction count
*/
{
 conn->status_line[0]= 0;
 conn->status_code= -1;
 conn->line[0]= 0;
 conn->line_valid= 0;
 conn->endseq_pt= 2;
 conn->buffer[0]= 0;
 conn->buffrpt=  conn->buffer;
 conn->buffer_status= 0;
 if(conn->show_io_stat)
   conn->show_io_stat= 3; /* full line clear */
 Nntpconn_errmsg_reset(conn,0);
 if(!(flag&8))
   Nntpconn_reset_t_count(conn,0);
 return(1);
}


int Nntpconn_errmsg_reset(conn,flag)
struct NntpconN *conn;
int flag;
{
 conn->errmsg[0]= 0;
 conn->errn= 0;
 return(1);
}


int Nntpconn_transaction(conn,send_text,recv_text,recv_siz,save_name,flag)
struct NntpconN *conn;
char *send_text;
char *recv_text;
unsigned long recv_siz;
char *save_name;
int flag;
/*
 bit0= write reply to stdout
 bit1= write reply to recv_text
 bit2= append reply to file save_name
 bit3= do not reset transaction count
 bit4= do not try to authenticate (set error)

 bit8= do not send send_text
 bit9= do not read status reply 
bit10= do not read reply data after status line
*/
/*
 return:
   <0 fatal error
   =0 retryable error
    1 OK Transaction completed
    2 OK Transaction needs to be continued
*/
{
 int ret,l,sret,i,status;
 char *line;

 line= conn->line;

 if(conn->active==0)
   return(Nntpconn_err_not_active(conn,0));

 if(conn->line_valid) { /* delayed from last output buffer overflow */
   sret= Nntpconn_result(conn,conn->line,recv_text,recv_siz,save_name,flag&7);
   if(sret<=0)
     {ret= -1; goto ex;}
   conn->line_valid= 0;
   if(!Nntpconn_recv_text(conn,"",0,"",256))
     {ret= 1; goto ex;} /* the buffered line contains the end sign */
 }

 /* enforce bandwidth limit, if limited: re-read control file */
 Nntpconn_enforce_limits(conn,conn->baudlimitfile,(conn->r_baudlimit>0));

 /* write request */
 if(!(flag&256)){
   Nntpconn_transaction_reset(conn,(flag&8));
   strcpy(conn->sent_cmd,send_text);
redo_request:;
   sret= Nntpconn_send(conn,conn->sent_cmd,0);
   if(sret<=0)
     {ret= sret; goto ex;}
 }
 
 if(!(flag&512)) {
   /* read status reply */
   Nntpconn_transaction_reset(conn,8);
   sret= Nntpconn_recv_status(conn,0);
   if(sret<=0)
     {ret= sret; goto ex;}
   sret= Nntpconn_result(conn,conn->status_line,recv_text,recv_siz,save_name,
                         flag&7);
   if(sret<=0)
     {ret= -1; goto ex;}

   if(conn->status_code==480) { /* challenge to authenticate received */
     if(flag&16){
       sprintf(conn->errmsg,
               "Recursive login challenge caught. Authentication failed.");
       conn->errn= -35;
       {ret= 0; goto ex;}
     } 
     ret= Nntpconn_authenticate(conn,0);
     if(ret<=0)
       goto ex;
     goto redo_request;
   }

 }

 if(!(flag&1024)) {
   /* read reply text */
   /* Only these status codes indicate following reply texts */
   status= conn->status_code;
   if( status==111 ||
      !((status>=100&&status<200)           /* HELP or debug*/
        || ( status==211 && strncmp(conn->sent_cmd,"LISTGROUP",9)==0 )
        || status==215                      /* LIST */
        || status==218                      /* XINDEX */
        || (status>=220 && status<=222)     /* ARTICLE, HEAD, BODY */
        || status==224                      /* XOVER */
        || status==230                      /* NEWNEWS */
        || status==231                      /* NEWGROUPS */
        || status==240                      /* POST second stage */
        || status==282                      /* XGTITLE */
        || status==288                      /* XTHREAD */
     )) {ret= 1; goto ex;}

      
   sret= Nntpconn_recv_text(conn,recv_text,recv_siz,save_name,flag&7);
   conn->line_valid= 0;
   if(sret<=0){
     if(conn->errn==-1) { /* Buffer overflow */
       conn->line_valid= 1;
       Nntpconn_errmsg_reset(conn,0);
       ret= 2; goto ex;
     }
     ret= sret; goto ex;
   }
 }

 ret= 1;
ex:;
 if(ret!=2)
   Nntpconn_show_stat(conn,1); /* clear display line */
 return(ret);
}

 
/*
>telnet news.easynews.net 119
Trying 194.24.209.75...
Connected to news.easynews.net.
Escape character is '^]'.
200 Welcome to easynews -- http://www.easynews.net/ (Typhoon v1.2.2)
GROUP alt.binaries.pictures.fantasy-sci-fi
480 Authentication Required
AUTHINFO USER myname
381 More Authentication Required
AUTHINFO PASS xyzABC123
281 Authentication Accepted
QUIT
205 GoodBye
Connection closed by foreign host.
*/

int Nntpconn_authenticate(conn,flag)
struct NntpconN *conn;
int flag;
{
 char cmd[Nntpconn_strL],old_sent_cmd[Nntpconn_strL];
 int ret;

 strcpy(old_sent_cmd,conn->sent_cmd);

 if(conn->login[0]==0){
   sprintf(conn->errmsg,
         "Server requires authentication, but no login information is known."); 
   conn->errn= -35;
   ret= 0;
   goto ex;
 }
 sprintf(cmd,"AUTHINFO USER %s",conn->login);
 ret= Nntpconn_transaction(conn,cmd,"",0,"",16);
 if(ret<=0)
   return(ret);
 if(conn->status_code==281)
   goto ok;
 if(conn->status_code!=381) {
   sprintf(conn->errmsg,
           "Server does not reply 281 or 381. Unknown protocol as it seems:");
   sprintf(conn->errmsg+strlen(conn->errmsg),"\n  %s",conn->status_line);
   conn->errn= -35;
   ret= 0;
   goto ex;
 }
 sprintf(cmd,"AUTHINFO PASS %s",conn->password);
 ret= Nntpconn_transaction(conn,cmd,"",0,"",16);
 if(ret<=0)
   return(ret);
 if(conn->status_code!=281) {
   sprintf(conn->errmsg,
           "Server does not reply 281. Authentication failed:");
   sprintf(conn->errmsg+strlen(conn->errmsg),"\n  %s",conn->status_line);
   conn->errn= -35;
   ret= 0;
   goto ex;
 }
ok:;
 ret= 1;
ex:;
 strcpy(conn->sent_cmd,old_sent_cmd);
 return(ret);
}


int Nntpconn_getline(conn,line,linsiz,flag)
struct NntpconN *conn;
char *line;
int linsiz;
int flag;
{
 int ret,cr_da=0;
 char *wpt,*rpt;

 if(conn->buffer_status==0) { /* reply completely unread , reset buffer */
   conn->buffrpt= conn->buffer;
   conn->buffer[0]= 0;
 }

 rpt= conn->buffrpt;
 wpt= line;
 while(1){
   if((*rpt)==0){ /* end of buffer reached */
     if(conn->buffer_status==2 || conn->buffer_status==0) {
       /* read next buffer (reply data only) */
       conn->buffer[0]= 0;
       ret= Nntpconn_transaction(conn,"",
                                 conn->buffer,sizeof(conn->buffer),"",
                                 2|256|512);
       if(ret<=0)
         return(ret);
       conn->buffer_status= ret;
       rpt= conn->buffrpt= conn->buffer;
     } else {
       goto fertig;
     }
   }
   if(wpt-line>=linsiz-1)
     goto fertig;
   *(wpt++)= *(rpt++);

   if(*(rpt-1)=='\r')
     cr_da= 2;
   if(*(rpt-1)=='\n' && cr_da>0){
 break;
   }
   cr_da--;
 } 
 *wpt= 0;

fertig:;
 *wpt= 0;
 conn->buffrpt= rpt;
 if(line[0]==0)
   return(0);
 return(1);
}

